#include "glclient.h"
#include <morfbase.h>
#include <morfclasses.h>
#include <trace.h>
#include <time.h>

//#define FULLSCREEN

namespace APPLICATION
{
	CGFXApp::~CGFXApp()
	{
		fe->KillWindow();

		if (fe) delete fe;
		fe=NULL;
		if (GetNetwork()) delete GetNetwork();
		SetNetwork(NULL);
		if (GetDomain())	delete GetDomain();
		SetDomain(NULL);
	}

	void CGFXApp::ExecuteSlice()
	{
		fe->Render(GetDomain()->GetVisibleObjects());
	}

	FRONTEND::CFrontEnd *CGFXApp::GetFE()
	{
		return fe;
	}

		
	ATOM CGFXApp::RegisterWindowClass(HINSTANCE hInstance)
	{
		WNDCLASS	wc;						// Windows Class Structure

		wc.style			= CS_HREDRAW | CS_VREDRAW | CS_OWNDC;	// Redraw On Size, And Own DC For Window.
		wc.lpfnWndProc		=   (WNDPROC)staticWindowProcedure;
								//(WNDPROC) windowproc;
								//(WNDPROC)staticWindowProcedure;
								//(WNDPROC) WndProc;					// WndProc Handles Messages
		wc.cbClsExtra		= 0;									// No Extra Window Data
		wc.cbWndExtra		= 0;									// No Extra Window Data
		wc.hInstance		= hInstance;							// Set The Instance
		wc.hIcon			= LoadIcon(NULL, IDI_WINLOGO);			// Load The Default Icon
		wc.hCursor			= LoadCursor(NULL, IDC_ARROW);			// Load The Arrow Pointer
		wc.hbrBackground	= NULL;									// No Background Required For GL
		wc.lpszMenuName		= NULL;									// We Don't Want A Menu
		wc.lpszClassName	= szTitle;								// Set The Class Name

		if (!RegisterClass(&wc))									// Attempt To Register The Window Class
		{
			MessageBox(NULL,"Failed To Register The Window Class.","ERROR",MB_OK|MB_ICONEXCLAMATION);
			return FALSE;											// Return FALSE
		}

		return TRUE;
	}
		

	void CGFXApp::RegisterClasses()
	{
		CWinApplication::RegisterClasses();
		GENERAL::RegisterClasses();
		GENERAL::DumpClassHirachy();
	}

	bool CGFXApp::Init()
	{
		
		RegisterClasses();

		GetTicker().setTick(-100000);

		szTitle = "morf3gl";
		DOMAIN::CObjectDomain *objDomain;

		//SetObjectMan(new DOMAIN::CSimpleObjectManager());
		
		GENERAL::dprintf("Creating Objectdomain \n");
		
		objDomain = new DOMAIN::CObjectDomain();
		

		SetDomain(objDomain);

		GENERAL::dprintf("Creating Network \n");


		SetNetwork(new NETWORK::CTCPNetwork(GetDomain(), false));

		GENERAL::dprintf("Initializing Domain \n");


		GetDomain()->Init();

		RegisterWindowClass(hInstance);
			
		fe = new FRONTEND::CGLFrontEnd(this, hInstance,objDomain);
#ifdef FULLSCREEN
		if (!fe->CreateFeWindow(szTitle,1024,768,32,true))
#else
		if (!fe->CreateFeWindow(szTitle,800,600,32,false))
#endif
		{
			MessageBox(NULL,"Initialization Failed.","ERROR",MB_OK|MB_ICONEXCLAMATION);
			return FALSE;
		}
	
		if (!fe->Init())									
		{
			fe->KillWindow();							
			MessageBox(NULL,"Initialization Failed.","ERROR",MB_OK|MB_ICONEXCLAMATION);
			return FALSE;
		}

	//		FRONTEND::CShaderFile *sf = new FRONTEND::CShaderFile("scripts/base_wall.shader");
//		sf->FindShader("textures/base_wall/glass_frame");
        		
	//	DOMAIN::CGraphObject *po = ((DOMAIN::CGraphObject *) ((DOMAIN::CObjectDomain *) GetDomain())->GetObjMan()->GetRef(25));
	//	po->SetfeData(fe->CreateObject(po));


		CApplication::GetApplication()->GetNetwork()->Connect();

		return true;
	}


	LRESULT CALLBACK CGFXApp::staticWindowProcedure(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
	{
		DWORD tid = GetCurrentThreadId();
		static CGFXApp *app=NULL;
		if (uMsg==WM_CREATE)
		{
			LPCREATESTRUCT cs =(LPCREATESTRUCT)lParam;
			app = (CGFXApp*)cs->lpCreateParams;
		}

		return (app==NULL)?
			DefWindowProc(hWnd,uMsg,wParam,lParam):
			app->WindowProcedure(hWnd,uMsg,wParam,lParam);

			
		//WindowProcedure(hWnd,uMsg,wParam,lParam);

		//	(CGFXApp*)APPLICATION::CGFXApp::GetApplication()->WindowProcedure(hWnd,uMsg,wParam,lParam);
		
		
	}

	LRESULT CALLBACK CGFXApp::WindowProcedure(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
	{
		static FRONTEND::TKeyStates keyState;
		static FRONTEND::TMouseButtons buttons;
		typedef struct {
			WORD repeat:16;
			char scancode:8;
			bool extendedkey:1;
			char reserved:4;
			bool altdown:1;
			bool previousdown:1;
			bool released:1;
		} TKey;
		TKey *key;

		switch (uMsg)
		{
		case WM_SIZE:
			{
				int xSize = (short)LOWORD(lParam);
				int ySize = (short)HIWORD(lParam);
				fe->Resize(xSize,ySize);
			}
			break;
		case WM_DESTROY:
			GetDomain()->Terminate();
			PostQuitMessage(0);
			break;
		case WM_KEYUP:
			key = (TKey*)&lParam;
//			if (((int)wParam)==VK_SHIFT) keyState[1]=false;
//			if (((int)wParam)==VK_CONTROL) keyState[0]=false;
			keyState[wParam&0xff]=false;
			fe->KeyUp((int)wParam,keyState);
			break;

		case WM_KEYDOWN:
			key = (TKey*)&lParam;
/*			if (((int)wParam)==VK_SHIFT) keyState[1]=true;
			if (((int)wParam)==VK_CONTROL) keyState[0]=true;
			if (((int)wParam)==VK_F1) fe->DumpGLState();*/
					
			if (!keyState[wParam&0xff])
			{
				keyState[wParam&0xff]=true;
				fe->KeyDown((int)wParam,keyState);
			}
			break;

		case WM_CHAR:
			key = (TKey*)&lParam;
			fe->KeyPress((int)wParam,keyState);
			break;
		case WM_MOUSEMOVE:
			{
				int xPos = (short)LOWORD(lParam);
				int yPos = (short)HIWORD(lParam);

				fe->MouseMove(xPos,yPos,0,buttons,keyState);
			}
			break;
		case WM_LBUTTONDOWN:
			{
				int xPos = (short)LOWORD(lParam);
				int yPos = (short)HIWORD(lParam);
				buttons[FRONTEND::MOUSEBUTTON_LEFT]=true;

				fe->MouseDown(FRONTEND::MOUSEBUTTON_LEFT,xPos,yPos,0,buttons,keyState);
			}
			break;
		case WM_RBUTTONDOWN:
			{
				int xPos = (short)LOWORD(lParam);
				int yPos = (short)HIWORD(lParam);
				buttons[FRONTEND::MOUSEBUTTON_RIGHT]=true;

				fe->MouseDown(FRONTEND::MOUSEBUTTON_RIGHT,xPos,yPos,0,buttons,keyState);
			}
			break;
		case WM_LBUTTONUP:
			{
				int xPos = (short)LOWORD(lParam);
				int yPos = (short)HIWORD(lParam);
				buttons[FRONTEND::MOUSEBUTTON_LEFT]=false;

				fe->MouseUp(FRONTEND::MOUSEBUTTON_LEFT,xPos,yPos,0,buttons,keyState);
			}
			break;
		case WM_RBUTTONUP:
			{
				int xPos = (short)LOWORD(lParam);
				int yPos = (short)HIWORD(lParam);
				buttons[FRONTEND::MOUSEBUTTON_RIGHT]=false;

				fe->MouseUp(FRONTEND::MOUSEBUTTON_RIGHT,xPos,yPos,0,buttons,keyState);
			}
			break;

		default:
			return DefWindowProc(hWnd,uMsg,wParam,lParam);
		}

		return 1;
	}

	int CGFXApp::Run()
	{
		MSG msg;

		int lastframecount=0,frames=0;
		float lasttime = (float)clock()/(float)CLOCKS_PER_SEC;

		while (!GetDomain()->ShouldTerminate()) 
		{
			if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
			{
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			} else {
				try {
					GetNetwork()->Lock();
					GetDomain()->execute();
		
					fe->Render(GetDomain()->GetVisibleObjects());
					frames++;

					float time = ((float)clock()/(float)CLOCKS_PER_SEC);

					if ( time>lasttime+1.0f )
					{
	//					dprintf("fps %f\n",(frames-lastframecount)/(time-lasttime));
						lastframecount=frames;
						lasttime = time;
					}

					GetNetwork()->Unlock();
				}
				catch (exception & e)
				{
					MessageBox(NULL,e.what(),"exception",MB_OK);
					GetDomain()->Terminate();
				}
			}
		}

		return (int) msg.wParam;;
	}

//	};

} // namespace

