#include "CTL2Label.h"
#include "CTL2Container.h"

namespace FRONTEND
{
	CCTL2Label::CCTL2Label(CFrontEnd *_fe, DOMAIN::CGraphObject *_domainObj, HResGUIStyle &_guiStyle, void *_userdata, FCTL2CallBack _callback)
		: CCTL2Window(_fe, _domainObj, _guiStyle, _userdata, _callback),
		  text(NULL),
		  textFlags(TEXTFLAG_HLEFT|TEXTFLAG_VCENTER),
		  baseLine(0)
	{
	}

	CCTL2Label::~CCTL2Label()
	{
		SAFE_FREE(text);
	}

	void CCTL2Label::SetStyleElement(CResGUIStyle::TElement *_styleElement)
	{
		CCTL2Window::SetStyleElement(_styleElement);
	}

	void CCTL2Label::CalculateSurfaces()
	{
		CCTL2Window::CalculateSurfaces();

		TMorfPointi screenPos(pos.pos);
		if (GetParent()!=NULL)
			screenPos += GetParent()->GetScreenPos();

		surface.texts.clear();
		TLineInfoList::iterator it;

		surface.flags&=~TSurface::SF_TEXTONLY;
		for (it=lineInfos.begin();it!=lineInfos.end();it++)
		{
			TLineInfo &li=(*it);
			CVector3f lp = li.pos;
			lp.x+=(SCALAR)screenPos.x;
			lp.y+=(SCALAR)screenPos.y;
			surface.AddText(&surface.shader->fnt,lp,TColor4ub(0,0,0,255),text+li.cofs,li.ccnt,false,true);
		}
	}

	void CCTL2Label::SetText(char *_text)
	{
		SAFE_FREE(text);
		text = strdup(_text);
		CalculateTextMetrics();
	}

	void CCTL2Label::CalculateTextMetrics()
	{
		lineInfos.clear();
		if (text!=NULL)
		{
			EHAlign halign;
			EVAlign valign;

			switch (GetHAlign())
			{
			case TEXTFLAG_HLEFT:
				halign = HALIGN_LEFT;
				break;
			default:
			case TEXTFLAG_HCENTER:
				halign = HALIGN_CENTER;
				break;
			case TEXTFLAG_HRIGHT:
				halign = HALIGN_RIGHT;
				break;
			}
			switch (GetVAlign())
			{
			case TEXTFLAG_VTOP:
				valign = VALIGN_TOP;
				break;
			default:
			case TEXTFLAG_VCENTER:
				valign = VALIGN_CENTER;
				break;
			case TEXTFLAG_VBOTTOM:
				valign = VALIGN_BOTTOM;
				break;
			}
			TMorfPointf size((SCALAR)GetPos().w,(SCALAR)GetPos().h);
			surface.shader->fnt->CalculateMultilinedText(size,text,halign,valign,DoWordWrap(),IsMultiLined(),lineInfos);
		}

		CalculateSurfaces();
	}

	void CCTL2Label::_Resize(int w, int h)
	{
		CCTL2Window::_Resize(w,h);
		CalculateTextMetrics();
	}

}