#include "CTL2ScrollingContainer.h"

namespace FRONTEND 
{
	CCTL2ScrollingContainer::CCTL2ScrollingContainer(CFrontEnd *_fe, DOMAIN::CGraphObject *_domainObj, HResGUIStyle &_guiStyle, void *_userdata, FCTL2CallBack _callback)
		: CCTL2ClientWindow(_fe,_domainObj, _guiStyle, _userdata, _callback),
		  vScrollBar(new CCTL2ScrollBar(_fe,_domainObj, _guiStyle, CCTL2ScrollBar::SB_VERTICAL, NULL, StaticCallBack)),
		  hScrollBar(new CCTL2ScrollBar(_fe,_domainObj, _guiStyle, CCTL2ScrollBar::SB_HORIZONTAL, NULL, StaticCallBack)),
		  scrollAreaSize(300,300)
	{
	}

	CCTL2ScrollingContainer::~CCTL2ScrollingContainer()
	{
	}

	void CCTL2ScrollingContainer::Init(sptr<CRenderAble> &_self)
	{
		CCTL2ClientWindow::Init(_self);
		vScrollBar->Init((sptr<CRenderAble>)vScrollBar);
		vScrollBar->SetUserdata(this);
		CCTL2Container::InsertControl((sptr<CCTL2Window>)vScrollBar);

		hScrollBar->Init((sptr<CRenderAble>)hScrollBar);
		hScrollBar->SetUserdata(this);
		CCTL2Container::InsertControl((sptr<CCTL2Window>)hScrollBar);
	
		SetClientSize();
		CalculateSurfaces();
	}

	void CCTL2ScrollingContainer::SetClientSize()
	{
		CCTL2ClientWindow::SetClientSize();
		if (IsScrollAreaAutoSizeX())
			scrollAreaSize.x = clientArea->GetPos().w;
		if (IsScrollAreaAutoSizeY())
			scrollAreaSize.y = clientArea->GetPos().h;

/*		hScrollBar->SetPos(
			borderSize.l,
			borderSize.b,
			GetPos().w-(borderSize.l+borderSize.r+(vScrollBar->IsVisible()?vScrollBar->GetPos().w:0)),
			hScrollBar->GetPos().h);*/
		TMorfRecti &rect = clientArea->GetPos();
		hScrollBar->SetPos(
			rect.x,
			rect.y-hScrollBar->GetPos().h,
			rect.w,
			hScrollBar->GetPos().h);
		hScrollBar->SetMinVal(0);
		hScrollBar->SetMaxVal(scrollAreaSize.x);
		hScrollBar->SetPageSize(scrollAreaSize.x);
		hScrollBar->SetScrollPos(0);

/*		vScrollBar->SetPos(
			GetPos().w-(borderSize.r+vScrollBar->GetPos().w),
			borderSize.b+(hScrollBar->IsVisible()?hScrollBar->GetPos().h:0),
			vScrollBar->GetPos().w,
			GetPos().h-(borderSize.t+borderSize.b+(hScrollBar->IsVisible()?hScrollBar->GetPos().h:0)));*/
		vScrollBar->SetPos(
			rect.x+rect.w,
			rect.y,
			vScrollBar->GetPos().w,
			rect.h);
		vScrollBar->SetMinVal(0);
		vScrollBar->SetMaxVal(scrollAreaSize.y);
		vScrollBar->SetPageSize(scrollAreaSize.y);
		vScrollBar->SetScrollPos(0);	
	}

	void CCTL2ScrollingContainer::CalculateClientSize(TMorfRecti &rect)
	{
		CCTL2ClientWindow::CalculateClientSize(rect);
		if (hScrollBar->IsVisible())
		{
			rect.y+=hScrollBar->GetPos().h;
			rect.h-=hScrollBar->GetPos().h;
		}
		if (vScrollBar->IsVisible())
		{
			rect.w-=vScrollBar->GetPos().w;
		}
	}

	void CCTL2ScrollingContainer::CalculateWindowSize(int &_cx, int &_cy)
	{
		CCTL2ClientWindow::CalculateWindowSize(_cx,_cy);
		if (vScrollBar->IsVisible()) _cx+=vScrollBar->GetPos().w;
		if (hScrollBar->IsVisible()) _cy+=hScrollBar->GetPos().h;
	}

	void CCTL2ScrollingContainer::_Resize(int w, int h)
	{
		CCTL2ClientWindow::_Resize(w,h);
		hScrollBar->SetPageSize(GetClientArea()->GetPos().w);
		vScrollBar->SetPageSize(GetClientArea()->GetPos().h);
	}

	bool CCTL2ScrollingContainer::StaticCallBack(CCTL2Window* ctl,ECTL2CallBackAction action,unsigned long lparam,void *pparam, void *userdata)
	{
		if (userdata!=NULL)
			return ((CCTL2ScrollingContainer*)userdata)->CallBack(ctl,action,lparam,(TCTL2PParam*)pparam);
		else
			return DummyCallBack(ctl,action,lparam,pparam,userdata);
	}

	bool CCTL2ScrollingContainer::CallBack(CCTL2Window* ctl,ECTL2CallBackAction action,unsigned long lparam,TCTL2PParam *pparam)
	{
		if ( (action==CB2ACTION_SCROLLBARCHANGE) && ((ctl==&hScrollBar) || (ctl==&vScrollBar)) )
		{
			GetClientArea()->SetRenderOfs(-hScrollBar->GetScrollPos(),vScrollBar->GetScrollPos());
		}
		return true;
	}



	CCTL2Dialog::CCTL2Dialog(CFrontEnd *_fe, DOMAIN::CGraphObject *_domainObj, HResGUIStyle &_guiStyle, void *_userdata, FCTL2CallBack _callback)
		: CCTL2ScrollingContainer(_fe,_domainObj, _guiStyle, _userdata, _callback),
		  titleBar(new CCTL2ClientWindow(_fe,_domainObj, _guiStyle, NULL, StaticCallBack)),
		  titleLabel(new CCTL2Label(_fe,_domainObj, _guiStyle, NULL, StaticCallBack)),
		  closeButton(new CCTL2Button(_fe,_domainObj, _guiStyle, NULL, StaticCallBack))
	{
		ClearScrollAreaAutoSizeX();
		ClearScrollAreaAutoSizeY();
		
		titleBar->ClearAnchorBottom();
		titleBar->SetAnchorTop();
		titleBar->SetAnchorLeft();
		titleBar->SetAnchorRight();

		titleLabel->SetAnchorTop();
		titleLabel->SetAnchorLeft();
		titleLabel->SetAnchorRight();
		titleLabel->SetAnchorBottom();
		titleLabel->ClearReceiveMouseMoveEvents();
		titleLabel->ClearReceiveMouseButtonEvents();
		titleLabel->ClearReceiveKeyEvents();

		closeButton->SetAnchorTop();
		closeButton->ClearAnchorLeft();
		closeButton->SetAnchorRight();
		closeButton->SetAnchorBottom();
	}

	CCTL2Dialog::~CCTL2Dialog()
	{
	}

	void CCTL2Dialog::Init(sptr<CRenderAble> &_self)
	{
		CCTL2ScrollingContainer::Init(_self);

		titleBar->Init((sptr<CRenderAble>)titleBar);
		titleBar->SetStyleElement(guiStyle->titlebar);
		titleBar->SetUserdata(this);

		TMorfRecti &tp=titleBar->GetClientArea()->GetPos();
		titleLabel->Init((sptr<CRenderAble>)titleLabel);
		titleLabel->SetPos(tp.x,tp.y,tp.w,tp.h);
		titleBar->InsertControl((sptr<CCTL2Window>)titleLabel);
		titleLabel->SetText("titleLabel");

		closeButton->Init((sptr<CRenderAble>)closeButton);
		closeButton->SetStyleElement(guiStyle->btnclose);
		closeButton->SetUserdata(this);
		int dist = (titleBar->GetPos().h-closeButton->GetPos().h)/2;
		closeButton->Move(tp.w-(closeButton->GetPos().w+dist),dist);
		titleBar->InsertControl((sptr<CCTL2Window>)closeButton);

		CCTL2Container::InsertControl((sptr<CCTL2Window>)titleBar);
		SetClientSize();
		CalculateSurfaces();
	}

	void CCTL2Dialog::CalculateWindowSize(int &_cx, int &_cy)
	{
		CCTL2ScrollingContainer::CalculateWindowSize(_cx,_cy);
		if (titleBar->IsVisible()) _cy+=titleBar->GetPos().h;
	}

	void CCTL2Dialog::CalculateClientSize(TMorfRecti &rect)
	{
		if (titleBar->IsVisible())
		{
			rect.h-=titleBar->GetPos().h;
			titleBar->SetPos(
				borderSize.l,
				GetPos().h-(borderSize.t+titleBar->GetPos().h),
				GetPos().w-(borderSize.l+borderSize.r),
				titleBar->GetPos().h);
		}

		CCTL2ScrollingContainer::CalculateClientSize(rect);
	}

	bool CCTL2Dialog::CallBack(CCTL2Window* ctl,ECTL2CallBackAction action,unsigned long lparam,TCTL2PParam *pparam)
	{
		if ( (ctl==&titleBar) || (ctl==titleBar->GetClientArea()) )
		{
			if (action==CB2ACTION_MOUSEDOWN)
			{
				if (pparam->mouse.button==MOUSEBUTTON_LEFT)
				{
					TMorfPointi mp(pparam->mouse.x,pparam->mouse.y);
					titleBar->ClientToParent(this,mp);
					DragStart(mp.x,mp.y);
				}
			}
			return true;
		} 
		else if ((ctl==&closeButton) && (action==CB2ACTION_MOUSECLICK))
		{
			parent->RemoveControl(this);
			return true;
		}

		
		else return CCTL2ScrollingContainer::CallBack(ctl,action,lparam,pparam);
	}
}