#include "CTL2StringList.h"

namespace FRONTEND
{
	CCTL2StringList::CCTL2StringList(CFrontEnd *_fe, DOMAIN::CGraphObject *_domainObj, HResGUIStyle &_guiStyle, void *_userdata, FCTL2CallBack _callback)
		: CCTL2ClientWindow(_fe,_domainObj,_guiStyle,_userdata,_callback),
		  vScrollBar(new CCTL2ScrollBar(_fe,_domainObj, _guiStyle, CCTL2ScrollBar::SB_VERTICAL, NULL, StaticCallBack)),
		  visibleLines(0),
		  useLines(&lines),
		  updateTime(500),
		  curTime(0)
	{
	}

	CCTL2StringList::~CCTL2StringList()
	{
	}

	void CCTL2StringList::Init(sptr<CRenderAble> &_self)
	{
		CCTL2ClientWindow::Init(_self);

		vScrollBar->Init((sptr<CRenderAble>)vScrollBar);
		vScrollBar->SetUserdata(this);
		CCTL2Container::InsertControl((sptr<CCTL2Window>)vScrollBar);
		vScrollBar->SetScrollPos(0);
		vScrollBar->SetPageSize(10);
		vScrollBar->SetMinVal(0);
		vScrollBar->SetMaxVal((int)useLines->size());

		SetClientSize();
		CalculateSurfaces();
	}

	void CCTL2StringList::CalculateClientSize(TMorfRecti &rect)
	{
		CCTL2ClientWindow::CalculateClientSize(rect);
		rect.w-=vScrollBar->GetPos().w;
	}

	void CCTL2StringList::CalculateSurfaces()
	{
		CCTL2ClientWindow::CalculateSurfaces();

		TLineList::iterator it;
		TSurface &surf = clientArea->GetSurface();

		surf.texts.clear();

		if (visibleLines==0)
			return;

		TMorfPointi screenPos(pos.pos);
		if (GetParent()!=NULL)
			screenPos += GetParent()->GetScreenPos();
		clientArea->ClientToParent(this,screenPos);

		int n = vScrollBar->GetScrollPos()+(visibleLines-1);
		if (n<0) return;
		it = useLines->begin();
		if (it==useLines->end()) return;

		while (n--)
		{
			it++;
			if (it==useLines->end()) return;
		}


		if (it==useLines->end())
			return;

		CVector3f lp((SCALAR)screenPos.x,(SCALAR)screenPos.y,0);
		lp.y+=surf.shader->fnt->GetHeight();

		for (int i=0; i<visibleLines; i++, it--)
		{
			surf.AddText(&surf.shader->fnt,lp,TColor4ub(0,0,0,255),(char*)it->data(),(int)it->size(),false,true);
			lp.y+=surf.shader->fnt->GetHeight();
			if (it==useLines->begin())
				break;
		}
		

	}

	void CCTL2StringList::CalculateWindowSize(int &_cx, int &_cy)
	{
		CCTL2ClientWindow::CalculateWindowSize(_cx,_cy);
		if (vScrollBar->IsVisible()) _cx+=vScrollBar->GetPos().w;
	}

	void CCTL2StringList::_Resize(int w, int h)
	{
		CCTL2ClientWindow::_Resize(w,h);

		TMorfRecti &rect = clientArea->GetPos();
		visibleLines = rect.h/GetShader()->fnt->GetHeight();
		visibleLines += (rect.h%GetShader()->fnt->GetHeight()==0)?0:1;
		if (visibleLines<0) visibleLines=0;
		if (visibleLines>(int)useLines->size()) visibleLines=(int)useLines->size();
		vScrollBar->SetPageSize(visibleLines);
		vScrollBar->SetPos(
			rect.x+rect.w,
			rect.y,
			vScrollBar->GetPos().w,
			rect.h);
	}

	void CCTL2StringList::InsertLine(char *text, int at)
	{
		if (at==0)
		{
			useLines->push_front(text);
		} else if (at==-1)
		{
			useLines->push_back(text);
		} else {
			TLineList::iterator it=useLines->begin()+at;
			useLines->insert(it,text);
		}

		_Resize(GetPos().w, GetPos().h);
		vScrollBar->SetScrollPos(vScrollBar->GetScrollPos()+1);
		CalculateSurfaces();
	}

	bool CCTL2StringList::StaticCallBack(CCTL2Window* ctl,ECTL2CallBackAction action,unsigned long lparam,void *pparam, void *userdata)
	{
		if (userdata!=NULL)
			return ((CCTL2StringList*)userdata)->CallBack(ctl,action,lparam,(TCTL2PParam*)pparam);
		else
			return DummyCallBack(ctl,action,lparam,pparam,userdata);
	}

	bool CCTL2StringList::CallBack(CCTL2Window* ctl,ECTL2CallBackAction action,unsigned long lparam,TCTL2PParam *pparam)
	{
		if ( (action==CB2ACTION_SCROLLBARCHANGE) && (ctl==&vScrollBar) )
		{
			CalculateSurfaces();
			return true;
		} 
		else return DummyCallBack(ctl,action,lparam,pparam,this);
	}

	void CCTL2StringList::RenderScissor(TMorfRecti _scissor, int _dt)
	{
		curTime+=_dt;
		if (curTime>updateTime)
		{
			curTime-=updateTime;

			int lineNum = (int)useLines->size();
			int dSize = lineNum-vScrollBar->GetMaxVal();
			if (dSize!=0)
			{
				vScrollBar->SetMaxVal(lineNum);
				vScrollBar->SetScrollPos(vScrollBar->GetScrollPos()+dSize);
				CalculateSurfaces();
			}
		}
		CCTL2ClientWindow::RenderScissor(_scissor,_dt);
	}


}
