#ifndef CTL2WINDOW_H
#define CTL2WINDOW_H

#include <morfbase.h>
#include "../resources/guistyle.h"
#include "../implementation/renderer.h"

namespace FRONTEND
{
	#define CCTL2_NUM_MOUSEBUTTONS 3

	class CCTL2Window;
	class CCTL2Container;

	union TCTL2PParam
	{
		struct TMouse
		{
			EMouseButton button;
			INT x;
			INT y;
			INT z;
			TMouseButtons *buttons;
			TKeyStates *keys;
		} mouse;
		struct TKeys
		{
			TKeyStates *keystate;
		} keys;
		struct TResize
		{
			INT ox,oy;
			INT x,y;
		} resize;
		struct TMove
		{
			INT ox,oy;
			INT x,y;
		} move;
	};

	enum ECTL2CallBackAction {
		CB2ACTION_MOUSEMOVE = 0,
		CB2ACTION_MOUSEDOWN = 1,
		CB2ACTION_MOUSEUP = 2,
		CB2ACTION_MOUSECLICK = 3,
		CB2ACTION_MOUSEDBLCLICK = 4,
		CB2ACTION_MOUSEUPOFFWINDOW = 5,
		CB2ACTION_MOUSEENTER = 6,
		CB2ACTION_MOUSELEAVE = 7,
		CB2ACTION_KEYDOWN = 100,
		CB2ACTION_KEYPRESS = 101,
		CB2ACTION_KEYUP = 102,
		CB2ACTION_MOVED = 200,
		CB2ACTION_MOVING = 201,
		CB2ACTION_RESIZED = 202,
		CB2ACTION_RESIZING = 203,
		CB2ACTION_CONTROLINSERTED = 300,
		CB2ACTION_CONTROLREMOVED = 300,
		CB2ACTION_SCROLLBARCHANGE = 301
	};


	enum ECTL2ReceiveEvents {
		CTL2RECEVENT_MOUSEMOVE		= 0x00000001,
		CTL2RECEVENT_MOUSEBUTTONS	= 0x00000002,
		CTL2RECEVENT_KEYS			= 0x00000004,
		CTL2RECEVENT_ALL			= 0xFFFFFFFF,
	};

	enum ECTL2WindowFlags {
		CTL2WINFLAG_VISIBLE					= 0x00000001,
		CTL2WINFLAG_HASBORDER				= 0x00000002,
		CTL2WINFLAG_HASBACKGROUND			= 0x00000004,
		CTL2WINFLAG_MOVEABLE				= 0x00000008,
		CTL2WINFLAG_RESIZEABLE				= 0x00000010,
		CTL2WINFLAG_CANFOCUS				= 0x00000020,
		CTL2WINFLAG_ENABLED					= 0x00000040,

		CTL2WINFLAG_ANCHOR_LEFT				= 0x00000100,
		CTL2WINFLAG_ANCHOR_TOP				= 0x00000200,
		CTL2WINFLAG_ANCHOR_RIGHT			= 0x00000400,
		CTL2WINFLAG_ANCHOR_BOTTOM			= 0x00000800,

		CTL2WINFLAG_SCROLLAREA_AUTOSIZE_X	= 0x00001000,
		CTL2WINFLAG_SCROLLAREA_AUTOSIZE_Y	= 0x00002000,

		CTL2WINFLAG_ALL						= 0xFFFFFFFF,
	};
	enum ECTL2ResizeArea {
		CTL2RESIZEAREA_NONE			= 0,
		CTL2RESIZEAREA_TOPLEFT		= 1,
		CTL2RESIZEAREA_TOP			= 2,
		CTL2RESIZEAREA_TOPRIGHT		= 3,
		CTL2RESIZEAREA_RIGHT		= 4,
		CTL2RESIZEAREA_BOTTOMRIGHT	= 5,
		CTL2RESIZEAREA_BOTTOM		= 6,
		CTL2RESIZEAREA_BOTTOMLEFT	= 7,
		CTL2RESIZEAREA_LEFT			= 8};

	enum EWindowElements
	{
		WINDOW_ELEMENT_C = 0,
		WINDOW_ELEMENT_C_P,

		WINDOW_ELEMENT_MAX
	};

	enum EIconElements
	{
		ICON_ELEMENT = 0,
		ICON_ELEMENT_P,

		ICON_ELEMENT_MAX
	};

	enum EBorderElements
	{
		BORDER_ELEMENT_B_TL=0,
		BORDER_ELEMENT_B_T,
		BORDER_ELEMENT_B_TR,
		BORDER_ELEMENT_B_L,
		BORDER_ELEMENT_B_R,
		BORDER_ELEMENT_B_BL,
		BORDER_ELEMENT_B_B,
		BORDER_ELEMENT_B_BR,
		BORDER_ELEMENT_STD_BCKG,
		BORDER_ELEMENT_TITLE_BCKG,

		BORDER_ELEMENT_MAX
	};

	typedef bool (__cdecl *FCTL2CallBack)(CCTL2Window*,ECTL2CallBackAction,unsigned long,void *, void *);

	#define CCTL2_SETFLAG(funcName, var, flag) inline void funcName() {(var) |= (flag);}
	#define CCTL2_CLEARFLAG(funcName, var, flag) inline void funcName() {(var) &= ~(flag);}
	#define CCTL2_TOGGLEFLAG(funcName, var, flag) inline void funcName() {(var) ^= (flag);}
	#define CCTL2_CHECKFLAG(funcName, var, flag) inline bool funcName() { return ((var) & (flag)) != 0;}
	#define CCTL2_ACCESSFLAG(setFunc, clearFunc, toggleFunc, checkFunc, var, flag) \
				CCTL2_SETFLAG(setFunc,var,flag) \
				CCTL2_CLEARFLAG(clearFunc,var,flag) \
				CCTL2_TOGGLEFLAG(toggleFunc,var,flag) \
				CCTL2_CHECKFLAG(checkFunc,var,flag) 

	class CCTL2Window : public CRenderAble
	{
	public:
		static sptr<CCTL2Window> focusedControl;
		static sptr<CCTL2Window> mouseOverControl;
		static sptr<CCTL2Window> mouseCaptureControl;
		static sptr<CCTL2Window> mouseButtonDownControl[CCTL2_NUM_MOUSEBUTTONS];

		struct TDragInfo { int x; int y; int sx; int sy;};
		struct TResizeInfo { bool sizeTop; bool sizeLeft; bool sizeBottom; bool sizeRight; int x; int y; int px; int py; int sx; int sy; };
		static TDragInfo dragInfo;
		static TResizeInfo resizeInfo;
	protected:
		enum ECTL2WindowState {
			STATE_NONE = 0,
			STATE_MOVEING,
			STATE_RESIZING,
			STATE_MARKING,
			STATE_MAX
		};

		static bool DummyCallBack(CCTL2Window*,ECTL2CallBackAction,unsigned long,void *pparam, void *userdata);

		TSurface surface;

	protected:
		long					 ctlID;
		FCTL2CallBack			 callBack;
		TMorfRecti				 pos;
		CCTL2Container			*parent;
		CCTL2Window				*focusControl;
		void					*userdata;
		unsigned long			 flags;
		unsigned long			 reveiveEvents;
		HResGUIStyle			 guiStyle;
		CResGUIStyle::TElement	*styleElement;
		wptr<CCTL2Window>		 self;
		int						 state;

		inline SetFlag(bool _on, unsigned long _flag)
		{
			if (_on) 
				flags|=_flag;
			else
				flags&=~_flag;
		}

		virtual void _SetPos(int x, int y, int w, int h);
		virtual void _Move(int x, int y);
		virtual void _Resize(int w, int h);

	public:
		CCTL2Window(CFrontEnd *_fe, DOMAIN::CGraphObject *_domainObj, HResGUIStyle &_guiStyle, void *_userdata, FCTL2CallBack _callback = DummyCallBack);		
		virtual ~CCTL2Window();
		virtual void Init(sptr<CRenderAble> &_self);

		virtual void Render(int _dt=0);
		virtual void RenderScissor(TMorfRecti _scissor, int _dt);
		virtual void CalculateSurfaces();

		virtual void SetStyleElement(CResGUIStyle::TElement *_styleElement);
		virtual void SetSubStyle(int _num);
		virtual void SetPos(int x, int y, int w, int h);
		virtual void Move(int x, int y);
		virtual void Resize(int w, int h);
		virtual void SetFocus();
		virtual void SetMouseCapture();
		virtual void ReleaseMouseCapture();

		inline CResShader *GetShader() { return &styleElement->clientShader; }
		inline SetShader(sptr<CResShader> &_shader) { surface.shader=_shader; }
		inline CCTL2Container *GetParent() {return parent;}
		inline TMorfRecti &GetPos() {return pos;}
		inline int GetState() {return state;}
		inline void SetParent(CCTL2Container *_parent) {parent = _parent;}
		virtual void SetUserdata(void *_userdata);
		inline void SetCTLID(long _ctlID) {ctlID=_ctlID;} 
		inline long &GetCTLID() {return ctlID;} 

		CCTL2_ACCESSFLAG(Show,Hide,ToggleVisiblitiy,IsVisible,flags,CTL2WINFLAG_VISIBLE)
		CCTL2_ACCESSFLAG(ShowBorder,HideBorder,ToggleBorderVisiblitiy,HasBorder,flags,CTL2WINFLAG_HASBORDER)
		CCTL2_ACCESSFLAG(ShowBackground,HideBackground,ToggleBackgroundVisiblitiy,HasBackground,flags,CTL2WINFLAG_HASBACKGROUND)
		CCTL2_ACCESSFLAG(SetMoveable,SetNotMoveable,ToggleMoveable,IsMoveable,flags,CTL2WINFLAG_MOVEABLE)
		CCTL2_ACCESSFLAG(SetResizeable,SetNotResizeable,ToggleResizeable,IsResizeable,flags,CTL2WINFLAG_RESIZEABLE)
		CCTL2_ACCESSFLAG(SetCanFocus,ClearCanfocus,ToggleCanFocus,CanFocus,flags,CTL2WINFLAG_CANFOCUS)
		CCTL2_ACCESSFLAG(Enable,Disable,ToggleEnables,IsEnabled,flags,CTL2WINFLAG_CANFOCUS)
		CCTL2_ACCESSFLAG(SetAnchorLeft,ClearAnchorLeft,ToggleAnchorLeft,HasAnchorLeft,flags,CTL2WINFLAG_ANCHOR_LEFT)
		CCTL2_ACCESSFLAG(SetAnchorTop,ClearAnchorTop,ToggleAnchorTop,HasAnchorTop,flags,CTL2WINFLAG_ANCHOR_TOP)
		CCTL2_ACCESSFLAG(SetAnchorRight,ClearAnchorRight,ToggleAnchorRight,HasAnchorRight,flags,CTL2WINFLAG_ANCHOR_RIGHT)
		CCTL2_ACCESSFLAG(SetAnchorBottom,ClearAnchorBottom,ToggleAnchorBottom,HasAnchorBottom,flags,CTL2WINFLAG_ANCHOR_BOTTOM)
		
		CCTL2_ACCESSFLAG(SetReceiveMouseMoveEvents,ClearReceiveMouseMoveEvents,ToggleReceiveMouseMoveEvents,ReceivesMouseMoveEvents,reveiveEvents,CTL2RECEVENT_MOUSEMOVE)
		CCTL2_ACCESSFLAG(SetReceiveMouseButtonEvents,ClearReceiveMouseButtonEvents,ToggleReceiveMouseButtonEvents,ReceivesMouseButtonEvents,reveiveEvents,CTL2RECEVENT_MOUSEBUTTONS)
		CCTL2_ACCESSFLAG(SetReceiveKeyEvents,ClearReceiveKeyEvents,ToggleReceiveKeyEvents,ReceivesKeyEvents,reveiveEvents,CTL2RECEVENT_KEYS)

		CCTL2_ACCESSFLAG(SetScrollAreaAutoSizeX,ClearScrollAreaAutoSizeX,ToggleScrollAreaAutoSizeX,IsScrollAreaAutoSizeX,flags,CTL2WINFLAG_SCROLLAREA_AUTOSIZE_X)
		CCTL2_ACCESSFLAG(SetScrollAreaAutoSizeY,ClearScrollAreaAutoSizeY,ToggleScrollAreaAutoSizeY,IsScrollAreaAutoSizeY,flags,CTL2WINFLAG_SCROLLAREA_AUTOSIZE_Y)

		virtual void ClientToScreen(TMorfPointi &_windowPos);
		virtual void ScreenToClient(TMorfPointi &_screenPos);
		virtual void ClientToParent(CCTL2Container *_parent, TMorfPointi &_windowPos);
		virtual void ParentToClient(CCTL2Container *_parent, TMorfPointi &_parentPos);
		virtual bool HasParent(CCTL2Container *_parent);
		virtual bool IsOver(int x, int y);
		virtual ECTL2ResizeArea GetResizeArea(int x,int y);

		virtual bool LostFocus();
		virtual void GotFocus();
		virtual bool KeyDown(int key,TKeyStates &keystate);
		virtual bool KeyUp(int key,TKeyStates &keystate);
		virtual bool KeyPress(int key,TKeyStates &keystate);
		virtual bool MouseEnter();
		virtual bool MouseLeave();
		virtual bool MouseMove(INT x,INT y,INT z,TMouseButtons &buttons,TKeyStates &keys);
		virtual bool MouseDown(EMouseButton button, INT x,INT y,INT z,TMouseButtons &buttons,TKeyStates &keys);
		virtual bool MouseUp(EMouseButton button, INT x,INT y,INT z,TMouseButtons &buttons,TKeyStates &keys);
		virtual bool MouseClick(EMouseButton button, INT x,INT y,INT z,TMouseButtons &buttons,TKeyStates &keys);

		virtual void DragStart(int x, int y);
		virtual void DragMove(int x, int y);
		virtual void DragEnd();

		virtual void ResizeStart(ECTL2ResizeArea _ra, int x, int y);
		virtual void ResizeMove(int x, int y);
		virtual void ResizeEnd();

		inline TSurface &GetSurface() { return surface;}

		friend CCTL2Container;
	};
}

#endif