#ifndef RENDERER_H
#define RENDERER_H

#include "surface.h"
#include "../resources/shader.h"
//#include <morfbase.h>
#include <morfclasses.h>

#include <morftypes.h>
#include "../gl/glextensions.h"
#include <glheaders.h>

namespace FRONTEND
{
	#define MAXVERTICES		16284
	#define MAXPRIMITIVES	(MAXVERTICES>>2)
	#define MAXTEXTUREUNITS 4

	struct TGLTextureUnit
	{
		GLboolean	cur_active1D;
		GLint		cur_texture1D;
		GLboolean	cur_active2D;
		GLint		cur_texture2D;

		// ClientState
		// glTexCoordPointer
		GLboolean	cur_tcaEnabled;
		GLint		cur_tcaSize;
		GLint		cur_tcaTyp;
		GLint		cur_tcaStride;
		GLvoid*		cur_tcaPointer;

		int matrixstacksize;
	};

	struct TGLState
	{
		// Scissor
		GLboolean		cur_scissorEnabled;

		// PolygonMode			
		GLint			cur_polygonMode;

		// Culling				
		GLboolean		cur_cullEnabled;
		GLint			cur_cullFace;

		// Rendermode
		GLenum			cur_rendermode;

		// Render Select
		GLuint			cur_name;

		// Blend States
		GLboolean		cur_blendEnabled;
		GLint			cur_blendsrc;
		GLint			cur_blenddst;

		// AlphaTestStates
		GLboolean		cur_alphaTestEnabled;
		GLint			cur_alphafunc;
		GLfloat			cur_alpharef;

		// DepthBuffer
		GLboolean		cur_depthTestEnabled;
		GLboolean		cur_depthMask;
		GLint			cur_depthFunc;

		GLint			cur_textureUnit;
		TGLTextureUnit	textureUnits[MAXTEXTUREUNITS];

		// ClientState
		GLint			cur_clientTextureUnit;

		// Arrays
		GLboolean		cur_vaEnabled;
		GLint			cur_vaSize;
		GLint			cur_vaTyp;
		GLint			cur_vaStride;
		GLvoid*			cur_vaPointer;

		GLboolean		cur_naEnabled;
		GLint			cur_naSize;
		GLint			cur_naTyp;
		GLint			cur_naStride;
		GLvoid*			cur_naPointer;

		GLboolean		cur_iaEnabled;
		GLint			cur_iaSize;
		GLint			cur_iaTyp;
		GLint			cur_iaStride;
		GLvoid*			cur_iaPointer;

		GLboolean		cur_efaEnabled;
		GLint			cur_efaSize;
		GLint			cur_efaTyp;
		GLint			cur_efaStride;
		GLvoid*			cur_efaPointer;

		GLboolean		cur_caEnabled;
		GLint			cur_caSize;
		GLint			cur_caTyp;
		GLint			cur_caStride;
		GLvoid*			cur_caPointer;
	};

	class CGLFrontEnd;

	struct TLight
	{
		CVector4f pos;
		CVector3f diffuseColor;
		GENERAL::SCALAR constantAttenuation;
		GENERAL::SCALAR linearAttenuation;
		GENERAL::SCALAR quadraticAttenuation;

		inline CVector3f CalculateDiffuseCol(CVector3f vertex,CVector3f normal)
			{
//				CVector3f result;
				CVector3f dir = vertex-pos;
				GENERAL::SCALAR len2 = dir.abs2();
				GENERAL::SCALAR len = sqrtf(len2);
				dir/=len;
				
				GENERAL::SCALAR l=-dir.ScalarMultiply(normal);
				l = constantAttenuation*l+linearAttenuation*l/len+quadraticAttenuation*l/len2;
				if (l>1.0f) l=1.0f;
				if (l<0) l=0;
				return diffuseColor*l;
				/*
				if (result.x>1.0f) result.x=1.0f;
				if (result.y>1.0f) result.y=1.0f;
				if (result.z>1.0f) result.z=1.0f;*/
			}
		
		std::vector<int> litSurfaces;
		std::vector<TSurface*> shadowCasters;
	};

	struct TRendererQuad
	{
	public:
		inline TRendererQuad(const TRendererQuad &quad)
			: shader(quad.shader),
			  pos(quad.pos),
			  scissor(quad.scissor),
			  styleElement(quad.styleElement) {}
		inline TRendererQuad(CResShader *_shader, TMorfRectf &_pos, TMorfRecti &_scissor, int _styleElement)
			: shader(_shader),
			  pos(_pos),
			  scissor(_scissor),
			  styleElement(_styleElement) {}

		CResShader *shader;
		TMorfRectf pos;
		TMorfRecti scissor;
		int styleElement;
	};

	struct TShadowSource
	{
		CVector3f pos;
		GLuint map;

		TShadowSource();
		virtual ~TShadowSource();

	};

	class CRendererGL
	{
	public:

		struct TShaderInfo {
			inline TShaderInfo() {};
			inline TShaderInfo(const TShaderInfo &info)
			{
				surfaces = info.surfaces;
				shader = info.shader;
			};

			HResShader shader;
			std::vector<int> surfaces;
		};
		struct TGUIShaderInfo : public TShaderInfo
		{
			inline TGUIShaderInfo() : TShaderInfo() {}
			inline TGUIShaderInfo(const TGUIShaderInfo &info) 
				: TShaderInfo(info), 
				  scissor(info.scissor),
				  styleElement(info.styleElement) {}

			TMorfRecti scissor;
			int styleElement;
		};

		typedef std::map<CResShader*,TShaderInfo> TShaderMap;
		typedef std::vector<TGUIShaderInfo> TGUIShaderList;
		typedef std::map<TShadowSource*, TShadowSource*> TShadowSourceMap;

	protected:
		static TGLState glState;

		bool bRenderWireFrame;
		bool bRenderNormals;
		bool bRenderLights;

		bool bLightingEnabled;

		TShadowSourceMap shadowSources;

		int numTrianglesFlushed;
		int numVertices;
		int numNormals;
		int numColors;
		int numTexCoords[MAXTEXTUREUNITS];
		int numSurfaces;
		int numSilhouettes;
		int numShadowCaster;
		int numShadowReceiver;

		int maxVertices;
		int maxNormals;
		int maxColors;
		int maxTexCoords[MAXTEXTUREUNITS];
		int maxSurfaces;		
		int maxSilhouettes;
		int maxShadowCaster;
		int maxShadowReceiver;


		CVector3f *vertices;
		CVector3f *normals;
		CVector2f **texCoords;
		GENERAL::TColor4ub *vertexColors;
		int *silhoutteSurfaces;
		int *shadowCasterSurfaces;
		int *shadowReceiverSurfaces;
	
		TSurface **surfaces;

		TGUIShaderList guiShaderList;
		TShaderMap shaderMap;

		CGLFrontEnd *fe;
		CResManagerGL *resMan;
		HResTextureGL whiteTexture;

		void GrowVertices(int num);
		void GrowNormals(int num);
		void GrowColors(int num);
		void GrowTexCoords(int num, int texunit);
		void GrowSurfaces(int num);
		void GrowSilhouetteSurfaces(int num);
		void GrowShadowCasterSurfaces(int num);
		void GrowShadowReceiverSurfaces(int num);

		int MaxLightSources;

		inline int GetTriangleCount(GLint primitiveType, GLint numVertices)
		{
			switch (primitiveType)
			{
			case GL_POINTS: 
			case GL_LINE_LOOP: return numVertices;

			case GL_LINE_STRIP: return numVertices-1;

			case GL_LINES:
			case GL_QUADS: return numVertices/2;

			case GL_TRIANGLE_FAN:
			case GL_POLYGON: 
			case GL_TRIANGLE_STRIP: 
			case GL_QUAD_STRIP: return numVertices-2;

			case GL_TRIANGLES: return numVertices/3;
			default: return 0;
			}
		}

	public:
		CVector3f ambientLight;

		CRendererGL(CGLFrontEnd *_fe, CResManagerGL *_resMan);
		~CRendererGL();

		void InitStates();

		TShadowSource *CreateShadowSource();
		void RemoveShadowSource(TShadowSource* ss);

		void RenderPlane(TColor4ub color, CVector3f normal, SCALAR d);
		void PushSilhouette(TSurface *_surface);
		void PushSurface(TSurface *_surface, bool shadowCaster, bool shadowReceiver);
		void PushGUISurface(TSurface *_surface, TMorfRecti &_scissor, int _styleElement=0);
		void Flush();

		int SetRenderMode(int mode); 
		inline bool GetRenderWireFrame() { return bRenderWireFrame; };
		inline void SetRenderWireFrame(bool _bRenderWireFrame){	bRenderWireFrame=_bRenderWireFrame;};
		inline bool GetRenderNormals() { return bRenderNormals; };
		inline void SetRenderNormals(bool _bRenderNormals)	{bRenderNormals=_bRenderNormals;};
		inline bool GetRenderLights() { return bRenderLights; };
		inline void SetRenderLights(bool _bRenderLights)	{bRenderLights=_bRenderLights;};

		inline bool GetLightingEnabled() { return bLightingEnabled; };
		inline void SetLightingEnabled(bool _bLightingEnabled)	{bLightingEnabled=_bLightingEnabled;};


		inline int GetTriangleCount() {return numTrianglesFlushed;};

	protected:


		void RenderShadowMaps(TShadowSource *ss); 
		void RenderShadows(TShadowSource *ss); 
		void RenderShadowMaps(); 
		void RenderShadows(); 

		void RenderScene();
		void RenderGUI();
		void RenderLight(TLight *light);
		void RenderNormals();
		void RenderSilhouettes();

		void InitShader(TShaderInfo &s);
//		void SetCullMode(CResShader *s);

		void FinishShader(TShaderInfo &s);
		GLint InitLayer(TShaderInfo &s, int &layer, TLight *light);
		void DoneLayer(TShaderInfo &s, int layer);

//		void RenderLight(TLight &l);

//		void SetShaderState();
		void CalculateVertices(TShaderInfo &s);
		void CalculateNormals(TShaderInfo &s);
		void SetScissor(TMorfRecti *_scissorRect);
		void SetVertexColors(TShaderInfo &s, CResShader::Layer *l, TLight *light);
		void SetTextureCoordinates(int texunit, TShaderInfo &s, CResShader::Layer *l);
		void BindTexture(int texunit, TShaderInfo &s, CResShader::Layer *l);
		void BindTexture(int texunit, GLint texNum);
		void DisableTextureUnit(int texunit);
		void SetBlendState(CResShader::Layer *l);
		void SetAlphaFunc(CResShader::Layer *l);
		void SetDepthFunc(CResShader::Layer *l);
		void SetSurfaceLights(TSurface *s);
		void RenderSurface(TSurface *s);
		void RenderSurfaceNormals(TSurface *s);

		inline void ExtrudeVerts(CVector3f *dest, CVector3f *verts, CVector3f *directions, int &num, float fac)
		{
			CVector3f *dst, *v, *d;
			int i;
			for (i=0, dst=dest, v=verts, d=directions; i<num; i++,dst++,v++,d++)
			{
				(*dst)=(*v) + ((*d) * fac);
			}
		}

		inline void SetVertexPointer(void *p)
		{
			if (p!=NULL)
			{
				if (glState.cur_vaPointer!=p)
				{
					glVertexPointer(3,GL_FLOAT,sizeof(CVector3f),p);
					glState.cur_vaPointer=p;
				}

				if (!glState.cur_vaEnabled)
				{
					glEnableClientState(GL_VERTEX_ARRAY);
					glState.cur_vaEnabled=true;
				}
			} else {
				if (glState.cur_vaEnabled)
				{
					glDisableClientState(GL_VERTEX_ARRAY);
					glState.cur_vaEnabled=false;
				}
			}
		}

		inline void SetNormalPointer(void *p)
		{
			if (p!=NULL)
			{
				if (glState.cur_naPointer!=p)
				{
					glNormalPointer(GL_FLOAT,sizeof(CVector3f),p);
					glState.cur_naPointer=p;
				}

				if (!glState.cur_naEnabled)
				{
					glEnableClientState(GL_NORMAL_ARRAY);
					glState.cur_naEnabled=true;
				}
			} else {
				if (glState.cur_naEnabled)
				{
					glDisableClientState(GL_NORMAL_ARRAY);
					glState.cur_naEnabled=false;
				}
			}
		}

		inline void SetColorPointer(void *p)
		{
			if (p!=NULL)
			{
				if (glState.cur_caPointer!=p)
				{
					glColorPointer(4,GL_UNSIGNED_BYTE,sizeof(TColor4ub),p);
					glState.cur_caPointer=p;
				}

				if (!glState.cur_caEnabled)
				{
					glEnableClientState(GL_COLOR_ARRAY);
					glState.cur_caEnabled=true;
				}
			} else {
				if (glState.cur_caEnabled)
				{
					glDisableClientState(GL_COLOR_ARRAY);
					glState.cur_caEnabled=false;

					glColor4ub(255,255,255,255);
				}
			}
		}

		inline void SetTextureCoordPointer(void *p,int textureunit)
		{
			textureunit-=GL_TEXTURE0;
			if (p!=NULL)
			{
				if (p!=glState.textureUnits[textureunit].cur_tcaPointer)
				{
					if (glState.cur_clientTextureUnit!=textureunit)
					{
						glClientActiveTexture(textureunit+GL_TEXTURE0);
						glState.cur_clientTextureUnit=textureunit;
					}

					glTexCoordPointer(2,GL_FLOAT,sizeof(CVector2f),p);
					glState.textureUnits[textureunit].cur_tcaPointer=p;
				}

				if (!glState.textureUnits[textureunit].cur_tcaEnabled)
				{
					if (glState.cur_clientTextureUnit!=textureunit)
					{
						glClientActiveTexture(textureunit+GL_TEXTURE0);
						glState.cur_clientTextureUnit=textureunit;
					}

					glEnableClientState(GL_TEXTURE_COORD_ARRAY);
					glState.textureUnits[textureunit].cur_tcaEnabled=true;
				}
			} else {
				if (glState.textureUnits[textureunit].cur_tcaEnabled)
				{
					if (glState.cur_clientTextureUnit!=textureunit)
					{
						glClientActiveTexture(textureunit+GL_TEXTURE0);
						glState.cur_clientTextureUnit=textureunit;
					}

					glDisableClientState(GL_TEXTURE_COORD_ARRAY);
					glState.textureUnits[textureunit].cur_tcaEnabled=false;
				}
			}
		}

		float Eval(float *p);
	};
}

#endif