#include "surface.h"

namespace FRONTEND
{
	inline int GetPolyCount(GLint primitiveType, GLint numVertices)
	{
		switch (primitiveType)
		{
		case GL_POINTS: 
		case GL_LINE_LOOP: 
		case GL_LINE_STRIP:
		case GL_LINES:			return 0;
		case GL_POLYGON:		return 1;
		case GL_QUADS:			return numVertices/4;
		case GL_QUAD_STRIP:		return (numVertices/2 - 1);
		case GL_TRIANGLE_FAN:
		case GL_TRIANGLE_STRIP:	return numVertices-2;
		case GL_TRIANGLES:		return numVertices/3;
		default: return 0;
		}
	}

	int GetEdgesPerPoly(GLint primitiveType, GLint numVertices)
	{
		switch (primitiveType)
		{
		case GL_POINTS: 
		case GL_LINE_LOOP: 
		case GL_LINE_STRIP:
		case GL_LINES:			return 0;
		case GL_POLYGON:		return numVertices;
		case GL_QUADS: 
		case GL_QUAD_STRIP:		return 4;
		case GL_TRIANGLE_FAN:
		case GL_TRIANGLE_STRIP:
		case GL_TRIANGLES:		return 3;
		default: return 0;
		}
	}

	int TSurface::GetPolygon(int num, int *indices)
	{
		int firstIndex=-1;
		int i,edges=0;
		switch (primitiveType)
		{
		default:
		case GL_POINTS: 
		case GL_LINE_LOOP: 
		case GL_LINE_STRIP:
		case GL_LINES:			return 0;
		case GL_POLYGON:		firstIndex=0;
								edges=(type==SURFACETYPE_POLYGON)?numVertices:mesh.numIndices; 
								break;
		case GL_QUADS: 			firstIndex=num*4;
								edges=4;
								break;
		case GL_QUAD_STRIP:		firstIndex=num*2;
								edges=4;
								break;
		case GL_TRIANGLE_FAN:
		case GL_TRIANGLE_STRIP:
								firstIndex=num;
								edges=3;
								break;
		case GL_TRIANGLES:		firstIndex=num*3;
								edges=3;
								break;
		}
			
		switch (type)
		{
		case SURFACETYPE_POLYGON:
			for (i=0; i<edges; i++)
			{
				indices[i]=firstIndex+i;
			}
			break;
		case SURFACETYPE_MESH:
			for (i=0; i<edges; i++)
			{
				indices[i]=mesh.indices[firstIndex+i];
			}
			break;
		}

		return edges;
	}

	void TSurface::AnalyseMeshStructure(TSurfaceStructure &out)
	{
		return;  // due to memory leak creation

		int numPolys;		
		int edgesPerPoly;

		out.surface=this;

		switch (type)
		{
		case SURFACETYPE_POLYGON:
			numPolys = GetPolyCount(primitiveType,numVertices);
			edgesPerPoly = GetEdgesPerPoly(primitiveType,numVertices);
			break;
		case SURFACETYPE_MESH:
			numPolys = GetPolyCount(primitiveType,mesh.numIndices);
			edgesPerPoly = GetEdgesPerPoly(primitiveType,mesh.numIndices);
			break;
		}

		// find equal verts
		std::map<int,int> indexTranslation;
		for (int i=0; i<numVertices; i++)
		{
			indexTranslation[i]=i;
			for (int n=i+1; n<numVertices; n++)
			{
				if (vertices[i]==vertices[n])
				{
					indexTranslation[i]=n;
				}
			}
		}

		// save polys
		for (int i=0; i<numPolys; i++)
		{
			TSurfaceStructure::TPoly poly;
			poly.indices	  = (int*) malloc(sizeof(int)*edgesPerPoly);
			poly.neigbourPoly = (int*) malloc(sizeof(int)*edgesPerPoly);
			for (int n=0; n<edgesPerPoly; n++)
				poly.neigbourPoly[n]=-1;

			GetPolygon(i,poly.indices);
			for (int n=0; n<edgesPerPoly; n++)
			{
				poly.indices[n]=indexTranslation[poly.indices[n]];
			}

			out.polys.push_back(poly);
		}


		bool meshClosed=true;
		// calculate edges
		for (int i=0; i<numPolys; i++)
		{
			TSurfaceStructure::TPoly &poly=out.polys[i];
			for (int e=0; e<edgesPerPoly; e++)
			{
				// this poly has already a neigbour
				if (poly.neigbourPoly[e]!=-1)
					continue;
				TSurfaceStructure::TEdge edge;
				edge.index1=poly.indices[e];
				edge.index2=poly.indices[(e+1)%edgesPerPoly];
				edge.polyIndex1=i;
				edge.polyIndex2=-1;
				edge.nextEdges.clear();
				edge.prevEdges.clear();

				for (int n=i+1; n<numPolys; n++)
				{
					TSurfaceStructure::TPoly &npoly=out.polys[n];
					for (int f=0; f<edgesPerPoly; f++)
					{
						int f1=npoly.indices[f];
						int f2=npoly.indices[(f+1)%edgesPerPoly];

						if ( ((edge.index1==f1) && (edge.index2==f2)) ||
							 ((edge.index2==f1) && (edge.index1==f2)) )
						{
							poly.neigbourPoly[e]=n;
							npoly.neigbourPoly[f]=i;
							edge.polyIndex2=n;
						}
					}
				}

				out.edges.push_back(edge);
				if (edge.polyIndex2==-1)
					meshClosed=false;
			}
		}


	}
}