#ifndef SURFACE_H
#define SURFACE_H

#include <morfbase.h>
#include "../resources/shader.h"
#include "../resources/font.h"

namespace FRONTEND
{
	struct TSurface;
	struct TSurfaceStructure
	{
		struct TPoly
		{
			int *indices;
			int *neigbourPoly;
		};

		struct TEdge
		{
			int index1;
			int index2;

			int polyIndex1;
			int polyIndex2;

			std::vector<int> nextEdges;
			std::vector<int> prevEdges;
		};

		std::vector<TPoly> polys;
		std::vector<TEdge> edges;

		TSurface *surface;
	};

	struct TSurfaceText
	{
		inline TSurfaceText(CResFont *_fnt,
							GENERAL::CVector3f &_pos,
							GENERAL::TColor4ub _color,
							char *_text,
							int _numChars,
							bool _marked,
							bool _setPos) 
			: fnt(_fnt),
			  pos(_pos),
			  color(_color),
			  text(_text),
			  numChars(_numChars),
			  marked(_marked),
			  setPos(_setPos) {}
		inline TSurfaceText(const TSurfaceText &st)
			: fnt(st.fnt),
			  pos(st.pos),
			  color(st.color),
			  text(st.text),
			  numChars(st.numChars),
			  marked(st.marked),
			  setPos(st.setPos){}
		inline TSurfaceText()
			: fnt(NULL),
			  pos(0.0f,0.0f,0.0f),
			  color(255,255,255,255),
			  text(NULL),
			  numChars(0),
			  marked(false),
			  setPos(true){}


		CResFont *fnt;
		GENERAL::CVector3f pos;
		GENERAL::TColor4ub color;
		char *text;
		int numChars;
		bool marked;
		bool setPos;
	};

	struct TSurface
	{
		HResShader shader;
		CResTexture *lightMap;

		int numVertices;
		int primitiveType;
		GENERAL::CVector3f *vertices;
		GENERAL::CVector3f *normals;
		GENERAL::CVector2f *texCoords;
		GENERAL::CVector2f *lightMapTexCoords;
		GENERAL::TColor4ub *colors;
		GENERAL::CMatrix4f *modelView;

		int name;

		/// @todo remove, or use !!
		unsigned long flags;
		enum TSurfaceFlags {
			SF_DYNAMIC_LIGHT = 0x0001,
			SF_TEXTONLY = 0x0002
		};

		enum TSurfaceType {
			SURFACETYPE_UNKNOWN,
			SURFACETYPE_POLYGON,
			SURFACETYPE_MESH
		} type;

		union {
			struct TSurfaceTypeMesh
			{
				int numIndices;
				int *indices;
			} mesh;
		} ;

		DOMAIN::CGraphObject::tLightsType *visibleLights;
		std::vector<TSurfaceText> texts;

		int GetPolygon(int num, int *indices);
		void AnalyseMeshStructure(TSurfaceStructure &out);
		inline int AddText(CResFont *_font, GENERAL::CVector3f &_pos, GENERAL::TColor4ub _color, char *_text, int _numChars, bool _marked, bool _setPos)
		{
			texts.push_back(TSurfaceText(_font,_pos,_color,_text,_numChars,_marked, _setPos));
			return (int)texts.size()-1;
		}
		inline TSurfaceText &GetText(int num)
		{
			return texts[num];
		}
		inline void RemoveText(int num)
		{
			texts.erase(texts.begin()+num);
		}



// dynamic data, this data is used by the renderer class. do not touch this
		int useVertices;
		int useNormals;
		int useTexCoords[16];
		int useColors;
	};


}

#endif