#include "md3Object.h"
#include "../implementation/glFE.h"

namespace FRONTEND
{
	//void LoadModelTextures(GENERAL::tMD3Model *pModel, char *strPath);
  
	CMD3Object::CMD3Object(void *_fe, DOMAIN::CGraphObject *_domainObj, char *_filename) 
		: CRenderAble(_fe, _domainObj), 
		  domainLinkMap(((DOMAIN::CMD3Object*)GetDomainObj())->GetLinks())
	{
		filename = strdup(_filename);
		pSurfaces=0;
		pSurfaceStructures=0;
		soundChannel=-1;
		sound=NULL;
		oldLink=0;
	
//		while (APPLICATION::CApplication::GetApplication()->GetLoadingthread().IsLoading())
//			Sleep(10);
//		while (link->GetStatus()==RESSTATUS_NOTLOADED)
//			Sleep(10);


		animt=0;
		framet =0;

		currentFrame=0;
		nextFrame=0;
		currentAnim=-1;

		shadowSource=NULL;

		//nextAnim=currentAnim;

	}

	void CMD3Object::Init(sptr<CRenderAble> &_self)
	{
		CRenderAble::Init(_self);
		link = ((CGLFrontEnd*)fe)->GetResMan()->GetPtr(_self,filename, RESTYP_MD3LINK);
		recFlag=0;

		/// @todo ResourceLoaded may be called two times
		if ( (!pSurfaces) && (link->GetStatus()>0) )
			ResourceLoaded(&link);
	}


	void CMD3Object::ResourceLoaded(CFEResource *_res)
	{
		if (_res!=&link)
			return;

//		pSurfaces = (TSurface*)malloc(link->GetLink().numOfMeshes*sizeof(TSurface));
		pSurfaces = new TSurface[link->GetLink().numOfMeshes];//*sizeof(TSurface));
		pSurfaceStructures = new TSurfaceStructure[link->GetLink().numOfMeshes];
		TSurface *curSurface=pSurfaces;
		TSurfaceStructure *curSurfaceStructure=pSurfaceStructures;
		
		std::vector<MD3MODEL::tMD3Mesh>::iterator it;
		for (it=link->GetLink().pMeshes.begin();it!=link->GetLink().pMeshes.end();it++)
		{
			curSurface->shader = ((CGLFrontEnd*)fe)->GetResMan()->GetPtr(sptr<CRenderAble>(),it->pShaders[0].strName, RESTYP_SHADER);
			if (curSurface->shader->GetShaderStatus()==CResShader::EShaderStatus::SS_NOTLOADED)
			{
				GENERAL::dprintf("[ INFO ] shader %s not found! Using default shader\n",it->pShaders[0].strName);
				curSurface->shader->LoadDefaultShader(it->pShaders[0].strName);
				curSurface->shader->layer[0]->rgbGen = RGBGEN_LIGHTING_DIFFUSE;
			}


			curSurface->numVertices = it->numOfVerts;
			curSurface->vertices = (GENERAL::CVector3f*) malloc( sizeof(GENERAL::CVector3f) * curSurface->numVertices);
			curSurface->normals  = (GENERAL::CVector3f*) malloc( sizeof(GENERAL::CVector3f) * curSurface->numVertices);
			curSurface->texCoords= (GENERAL::CVector2f*) malloc( sizeof(GENERAL::CVector2f) * curSurface->numVertices);
			for (int i=0; i<curSurface->numVertices; i++)
			{
				curSurface->texCoords[i].x = it->pTexCoords[i].x;
				curSurface->texCoords[i].y = it->pTexCoords[i].y;
				curSurface->vertices[i]=it->pVerts[i];
			}
			curSurface->mesh.numIndices = it->numOfTriangles*3;
			curSurface->mesh.indices = (int*)malloc(sizeof(int)*curSurface->mesh.numIndices);
			for (int i=0; i<it->numOfTriangles; i++)
			{
				curSurface->mesh.indices[i*3  ] = it->pTriangles[i].vertexIndices[0];
				curSurface->mesh.indices[i*3+1] = it->pTriangles[i].vertexIndices[1];
				curSurface->mesh.indices[i*3+2] = it->pTriangles[i].vertexIndices[2];
			}
			curSurface->lightMapTexCoords = NULL;
			curSurface->lightMap = NULL;
			curSurface->colors = NULL;
			curSurface->type = TSurface::SURFACETYPE_MESH;
			curSurface->primitiveType = GL_TRIANGLES;
			curSurface->name=0;
			curSurface->flags=0;
			curSurface->modelView=&modelView;

			curSurfaceStructure->polys.clear();
			curSurfaceStructure->edges.clear();

			curSurface->AnalyseMeshStructure(*curSurfaceStructure);

			curSurfaceStructure++;
			curSurface++;
		}
	}

	CMD3Object::~CMD3Object()
	{
		if (shadowSource)
			((CGLFrontEnd*)fe)->GetRenderer().RemoveShadowSource(shadowSource);
		if (sound!=NULL) ((CGLFrontEnd*)fe)->FreeSound(sound);
		SAFE_FREE(filename);
		if (pSurfaces)
		{
			for (int i=0; i<link->GetLink().numOfMeshes;i++)
			{
				SAFE_FREE(pSurfaces[i].vertices);
				SAFE_FREE(pSurfaces[i].normals);
				SAFE_FREE(pSurfaces[i].texCoords);
				SAFE_FREE(pSurfaces[i].mesh.indices);
			}
		}
		SAFE_DELETE_AR(pSurfaceStructures);
		SAFE_DELETE_AR(pSurfaces);
	}

	void CMD3Object::Update(int _flags)
	{
		CRenderAble::Update(_flags);

/*		ClearRecFlags();
		Propagate(domainObj->GetFSMState());
//		ExecuteCommand(domainObj->GetFSMState());*/
	}

	void CMD3Object::ClearRecFlags(DOMAIN::CMD3Object::tLinkMap &_linkMap)
	{
		if (recFlag==0) return;

		recFlag=0;

		for (int i=0; i<link->GetLink().numOfTags; i++)
		{
			int tagID = link->GetLink().pTags[i].tagid;
			
			DOMAIN::CMD3Object::tLinkMap::iterator it = _linkMap.find(tagID);
			if (it!=_linkMap.end())
			{
				std::vector<DOMAIN::CMD3Object::tLink>::iterator lit;

				for (lit=it->second.begin(); lit!=it->second.end(); lit++)
				{
					((CMD3Object*)&lit->link)->ClearRecFlags(_linkMap);
				}
			}
		}
	}

	void CMD3Object::UpdateAnimation(float _dt)
	{
		if ((link->GetLink().numOfAnimations==0) || (currentAnim==-1)) return;
		MD3MODEL::tMD3AnimationInfo &ai = link->GetLink().pAnimations[currentAnim];
		
		float cf;
		int icf;

		animt += _dt;
		framet = modff(animt*((float)ai.framesPerSecond/1.0f), &cf);
		icf=(int)cf;
        
		int nloop = ai.numFrames - ai.loopingFrames;

		if (icf<ai.numFrames)
		{
			currentFrame=icf+ai.startFrame;
		} else {
//			ClearRecFlags();
			ExecuteCommand(0);
			currentFrame = (((int)cf-nloop)%ai.loopingFrames)+nloop+ai.startFrame;
		}

		icf++;
		if (icf<ai.numFrames)
		{
			nextFrame=icf+ai.startFrame;
		} else {
			nextFrame = (((int)cf+1-nloop)%ai.loopingFrames)+nloop+ai.startFrame;
		}
	}

	long CMD3Object::ExecuteCommand(long cmd)
	{
		if (cmd<0)
		{
			animt=0.0f;
			framet=0.0f;
			currentAnimName=-cmd;
			currentAnim=TranslateAnimName(-cmd);

			return cmd;

		}
		long nextAnim=link->GetFSM()->GetNextState(currentAnimName,cmd);
		if (nextAnim>0)
		{
			animt=0.0f;
			framet=0.0f;
			currentAnimName=nextAnim;
			currentAnim=TranslateAnimName(nextAnim);
			return cmd;
		}
		if (nextAnim<0)
		{
			animt=0.0f;
			framet=0.0f;
			currentAnimName=-nextAnim;
			currentAnim=TranslateAnimName(-nextAnim);

			return nextAnim;
		}
		
		return cmd;
	}

	void CMD3Object::PropagateStateCommand(int _cmd, DOMAIN::CMD3Object::tLinkMap &_linkMap)
	{
		if (recFlag & 0x2) return;
		
		

		recFlag |= 0x2;

		long prop = ExecuteCommand(_cmd);

		for (int i=0; i<link->GetLink().numOfTags; i++)
		{
			int tagID = link->GetLink().pTags[i].tagid;
			
			DOMAIN::CMD3Object::tLinkMap::iterator it = _linkMap.find(tagID);
			if (it!=_linkMap.end())
			{
				std::vector<DOMAIN::CMD3Object::tLink>::iterator lit;

				for (lit=it->second.begin(); lit!=it->second.end(); lit++)
				{
					((CMD3Object*)&lit->link)->PropagateStateCommand(prop,_linkMap);
				}
			}
		}
	}

	void CMD3Object::PropagateFECommand(EMD3FECommand _cmd, DOMAIN::CMD3Object::tLinkMap &_linkMap)
	{
		if (recFlag & 0x4) return;
		recFlag |= 0x4;

		ExecuteFECommand(_cmd);

		for (int i=0; i<link->GetLink().numOfTags; i++)
		{
			int tagID = link->GetLink().pTags[i].tagid;
			
			DOMAIN::CMD3Object::tLinkMap::iterator it = _linkMap.find(tagID);
			if (it!=_linkMap.end())
			{
				std::vector<DOMAIN::CMD3Object::tLink>::iterator lit;

				for (lit=it->second.begin(); lit!=it->second.end(); lit++)
				{
					((CMD3Object*)&lit->link)->PropagateFECommand(_cmd,_linkMap);
				}
			}
		}
	}

	void CMD3Object::ExecuteFECommand(EMD3FECommand _cmd)
	{
		switch (_cmd)
		{
		case MD3SELECT:
			isSelected = true;
			break;
		case MD3DESELECT:
			isSelected = false;
			break;
		} 
	}

	void CMD3Object::Render(int _dt)
	{
		if (shadowSource)
		{
			shadowSource->pos = ((DOMAIN::CMD3Object*)domainObj)->GetPos();
		}

		if (sound)
		{
			CVector3f pos = ((DOMAIN::CMD3Object*)domainObj)->GetPos();
			CVector3f vel = ((DOMAIN::CMD3Object*)domainObj)->GetVelocity();
			pos*=1000;

			signed char res = FSOUND_3D_SetAttributes(soundChannel, &pos.x, &vel.x);
			pos*=1000;
		}

		if (!pSurfaces) return; 

		((CMD3Object*)&domainLinkMap[1].front().link)->ClearRecFlags(domainLinkMap);

		std::queue<long> &feCommands = ((DOMAIN::CMD3Object*)GetDomainObj())->feCommands;
		while (!feCommands.empty())
		{
			long cmd = feCommands.front();
			feCommands.pop();

			((CMD3Object*)&domainLinkMap[1].front().link)->PropagateStateCommand(cmd,domainLinkMap);
		}
		

		GENERAL::CVector3f &pos = ((DOMAIN::CMD3Object*)GetDomainObj())->GetPos();
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glTranslatef(pos.x,pos.y,pos.z);

		glTranslatef(0,-((DOMAIN::CMD3Object*)GetDomainObj())->GetBoundingBox().minV.y ,0);

		GENERAL::CVector3f &rotation=((DOMAIN::CWorldObject*) domainObj)->GetRotation();
		GENERAL::CVector3f axis =  rotation.VectorProduct(GENERAL::CVector3f(1.0f,0.0f,0.0f));
		GENERAL::CVector3f v(rotation.x,0,rotation.z);
		v.Normal();
		GENERAL::SCALAR angle = -v.Angle(GENERAL::CVector3f(1.0f,0.0f,0.0f));
		glRotatef((GLfloat)angle * (180 / PI),0.0f,axis.y>0.0f?1.0f:-1.0f,0.0f);


		

		glRotatef(-90, 1, 0, 0);

		GENERAL::CVector3f &scale = ((DOMAIN::CMD3Object*)GetDomainObj())->GetScale();
		glScalef(scale.x, scale.y, scale.z);

		if ((domainLinkMap[1].front().id!=oldLink) && (domainLinkMap[1].front().id==-250))
		{
			oldLink=-250;
			if (sound!=NULL) ((CGLFrontEnd*)fe)->FreeSound(sound);
			sound = ((CGLFrontEnd*)fe)->CreateSound("data/audio/plasma/shot.wav",false);
			soundChannel = ((CGLFrontEnd*)fe)->PlaySound(sound);
		} 
		else if ((domainLinkMap[1].front().id!=oldLink) && (domainLinkMap[1].front().id==-251))
		{
			oldLink=-251;
			if (sound!=NULL) ((CGLFrontEnd*)fe)->FreeSound(sound);
			sound = ((CGLFrontEnd*)fe)->CreateSound("data/audio/plasma/impact.wav",false);
			soundChannel = ((CGLFrontEnd*)fe)->PlaySound(sound,soundChannel);
		} 
		else if ((domainLinkMap[1].front().id!=oldLink) && (domainLinkMap[1].front().id==-252))
		{
			oldLink=-252;
			if (sound!=NULL) ((CGLFrontEnd*)fe)->FreeSound(sound);
			sound = ((CGLFrontEnd*)fe)->CreateSound("data/audio/rocket/shot.wav",false);
			soundChannel = ((CGLFrontEnd*)fe)->PlaySound(sound);
		} 
		else if ((domainLinkMap[1].front().id!=oldLink) && (domainLinkMap[1].front().id==-253))
		{
			oldLink=-253;
			if (sound!=NULL) ((CGLFrontEnd*)fe)->FreeSound(sound);
			sound = ((CGLFrontEnd*)fe)->CreateSound("data/audio/rocket/impact.wav",false);
			soundChannel = ((CGLFrontEnd*)fe)->PlaySound(sound,soundChannel);
		} 
		else if ((domainLinkMap[1].front().id!=oldLink) && (domainLinkMap[1].front().id==-254))
		{
			oldLink=-254;
			if (sound!=NULL) ((CGLFrontEnd*)fe)->FreeSound(sound);
			sound = ((CGLFrontEnd*)fe)->CreateSound("data/audio/fire/sshotf1b.wav",false);
			soundChannel = ((CGLFrontEnd*)fe)->PlaySound(sound);
		} 
		else if ((domainLinkMap[1].front().id!=oldLink) && (domainLinkMap[1].front().id==-255))
		{
			oldLink=-255;
			if (sound!=NULL) ((CGLFrontEnd*)fe)->FreeSound(sound);
			sound = ((CGLFrontEnd*)fe)->CreateSound("data/audio/fire/bfg_fire.wav",false);
			soundChannel = ((CGLFrontEnd*)fe)->PlaySound(sound,soundChannel);
		} 
		((CMD3Object*)&domainLinkMap[1].front().link)->RenderRecursive(_dt,domainLinkMap);
//		RenderRecursive(_dt, domainLinkMap);

		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
	}

	GENERAL::CVector3f cPos; // unschn!!!

	bool CompareLights(DOMAIN::CLight *l1, DOMAIN::CLight *l2)
	{
		// nicht schn!! verwendet nur LinAtt
		GENERAL::CVector3f lv1 = l1->GetPos()-cPos;
		GENERAL::CVector3f lv2 = l2->GetPos()-cPos;

		float lf1 = (l1->GetLinearAtt())*lv1.abs();
		float lf2 = (l2->GetLinearAtt())*lv2.abs();

		return (lf1 < lf2);
	}




	void CMD3Object::RenderRecursive(int _dt, DOMAIN::CMD3Object::tLinkMap &_linkMap)
	{
		/// todo unbsafe cast to DOMAIN::CQuakeBsp*

		DOMAIN::CGraphObject::tLightsType *lights;

		DOMAIN::CQuakeBsp *bsp = ((DOMAIN::CQuakeBsp*)((DOMAIN::CLevelObject*)domainObj)->GetLvl());

		int cluster = bsp->m_pLeafs[bsp->FindLeaf(((DOMAIN::CLevelObject*)domainObj)->GetPos())].cluster;
		lights  = bsp->m_pClusterVisibleLights+cluster;

		cPos = ((DOMAIN::CWorldObject*)domainObj)->GetPos();
//		std::sort(lights->begin(), lights->end(), CompareLights); 





		UpdateAnimation(((float)_dt)/1000.0f);

		glMatrixMode(GL_MODELVIEW);
		if (recFlag &0x1)
			return;

		glMatrixMode(GL_MODELVIEW);
		glGetFloatv(GL_MODELVIEW_MATRIX,modelView.m);

		for (int n=0; n<link->GetLink().numOfMeshes; n++)
		{
			MD3MODEL::tMD3Mesh &mesh = link->GetLink().pMeshes[n];
			int currentIndex = currentFrame*mesh.numOfVerts;
			int nextIndex = nextFrame*mesh.numOfVerts;
			for (int i=0; i<pSurfaces[n].numVertices; i++)
			{
				CVector3f &cpos = mesh.pVerts[currentIndex];
				CVector3f deltaPos = (mesh.pVerts[nextIndex]-cpos)*framet;
				pSurfaces[n].vertices[i]=deltaPos+cpos;

				CVector3f &cnormal = mesh.pNormals[currentIndex];
				CVector3f deltaNormal = (mesh.pNormals[nextIndex]-cnormal)*framet;
				pSurfaces[n].normals[i]=deltaNormal+cnormal;
				pSurfaces[n].normals[i].Normal();

				currentIndex++;
				nextIndex++;
			}

			if (selectMode==RENDERSELECT_NORMAL)
			{
				pSurfaces[n].name=domainObj->GetID();
			} else if (selectMode==RENDERSELECT_DETAIL)
			{
				pSurfaces[n].name=(int)this;
			}

			pSurfaces[n].visibleLights = lights;

			((CGLFrontEnd*)fe)->GetRenderer().PushSurface(pSurfaces+n, shadowCaster, shadowReceiver);

			if (isSelected)
				((CGLFrontEnd*)fe)->GetRenderer().PushSilhouette(pSurfaces+n);

			/*glActiveTexture(GL_TEXTURE0);
			glDisable(GL_TEXTURE_2D);
			glActiveTexture(GL_TEXTURE1);
			glDisable(GL_TEXTURE_2D);
			glActiveTexture(GL_TEXTURE2);
			glDisable(GL_TEXTURE_2D);
			glActiveTexture(GL_TEXTURE3);
			glDisable(GL_TEXTURE_2D);
			glDisable(GL_BLEND);
			glBlendFunc(GL_ONE,GL_ONE);

			glBegin(GL_LINES);
			for (int i=0; i<pSurfaceStructures[n].edges.size(); i++)
			{
				if ( (pSurfaceStructures[n].edges[i].polyIndex1==-1) ||
					 (pSurfaceStructures[n].edges[i].polyIndex2==-1))
					glColor4ub(255,0,0,255);
				else 
					glColor4ub(0,255,0,255);
				int i1 = pSurfaceStructures[n].edges[i].index1;
				int i2 = pSurfaceStructures[n].edges[i].index2;
				glVertex3fv(&pSurfaces[n].vertices[pSurfaceStructures[n].edges[i].index1].x);
				glVertex3fv(&pSurfaces[n].vertices[pSurfaceStructures[n].edges[i].index2].x);
			}
			glEnd();*/
		}

		//isRendered=true;
		recFlag |= 0x1;

		for (int i=0; i<link->GetLink().numOfTags; i++)
		{
			int tagID = link->GetLink().pTags[i].tagid;
		
			DOMAIN::CMD3Object::tLinkMap::iterator it = _linkMap.find(tagID);
			if (it!=_linkMap.end())
			{
				std::vector<DOMAIN::CMD3Object::tLink>::iterator lit;

				if (!it->second.empty())
				{


					glMatrixMode(GL_MODELVIEW);
					glPushMatrix();

					float *pMatrix, *pNextMatrix;
					float finalMatrix[16] = {0};
					CQuaternion qQuat, qNextQuat, qInterpolatedQuat;

					GENERAL::CVector3f vPosition = link->GetLink().pTags[currentFrame * link->GetLink().numOfTags + i].vPosition;
					GENERAL::CVector3f vNextPosition = link->GetLink().pTags[nextFrame * link->GetLink().numOfTags + i].vPosition;
					
					vPosition = vPosition + (vNextPosition - vPosition)*framet;
					
					pMatrix = link->GetLink().pTags[currentFrame * link->GetLink().numOfTags + i].rotation.m;
					pNextMatrix = link->GetLink().pTags[nextFrame * link->GetLink().numOfTags + i].rotation.m;

					qQuat.CreateFromMatrix( pMatrix, 3);
					qNextQuat.CreateFromMatrix( pNextMatrix, 3 );

					qInterpolatedQuat = qQuat.Slerp(qQuat, qNextQuat, framet);

					qInterpolatedQuat.CreateMatrix( finalMatrix );

					//memset(finalMatrix,0,16*sizeof(float));
					//finalMatrix[0] = finalMatrix[5] = finalMatrix[10] = finalMatrix[15] = 1.0f;
					
					finalMatrix[12] = vPosition.x;
					finalMatrix[13] = vPosition.y;
					finalMatrix[14] = vPosition.z;

					glMatrixMode(GL_MODELVIEW);
					//glGetFloatv(GL_MODELVIEW_MATRIX,finalMatrix);
					glMultMatrixf(finalMatrix);

					// do translations
					for (lit=it->second.begin(); lit!=it->second.end(); lit++)
					{
						((CMD3Object*)&lit->link)->SetSelectMode(selectMode,selectName);
						((CMD3Object*)&lit->link)->RenderRecursive(_dt, _linkMap);
					}
					glMatrixMode(GL_MODELVIEW);
					glPopMatrix();
				}
			}
		}
	}

	bool CMD3Object::MouseDown(EMouseButton button, int x,int y,int z,TMouseButtons &buttons,TKeyStates &keys)
	{
		if (button==MOUSEBUTTON_RIGHT)
		{
			DOMAIN::CGameObject::tActionList al;
			GetDomainObj()->GetActionList(al,0);
			((CGLFrontEnd*)fe)->PopUp(this,al);
			return true;
		} 
		else if (button==MOUSEBUTTON_LEFT)
		{
			Select();
			return true;
		} else
			return false;

//		((CMD3Object*)domainObj->GetfeData())->framet=0;
//		((CMD3Object*)domainObj->GetfeData())->animt=0;
	}

	void CMD3Object::Select()
	{
		CRenderAble::Select();
		ClearRecFlags(domainLinkMap);
		PropagateFECommand(MD3SELECT,domainLinkMap);
	}

	void CMD3Object::Deselect()
	{
		CRenderAble::Deselect();
		ClearRecFlags(domainLinkMap);
		PropagateFECommand(MD3DESELECT,domainLinkMap);
	}

}