#include "quakebsp.h"
#include "../gl/glextensions.h"
#include "../implementation/glFE.h"

namespace FRONTEND
{

	CObjQuakeBsp::CObjQuakeBsp(void *_fe, CResourceManager *_resMan, DOMAIN::CQuakeBsp *_bsp) : CRenderAble(_fe,_bsp), resMan(_resMan)
	{
		m_numOfFaces		= 0;
		m_numOfVerts		= 0;
		m_numOfPatches		= 0;
		m_numOfLightmaps	= 0;		
		m_numOfMeshIndices	= 0;
		m_numOfShaders		= 0;

		m_pFaces		= NULL;
		m_pVerts		= NULL;	
		m_pPatches		= NULL;
		m_pLightmaps	= NULL;
		m_pMeshIndices	= NULL;
		m_pShaders		= NULL;

		LoadDataFromFile();

		Tesselate(CONFIG::g_config.ConfInt["GeometricDetail"]);

		shaders = new HResShader[GetBsp()->m_numOfTextures];
		
		for (int i=0; i<GetBsp()->m_numOfTextures; i++)
		{
			char name[128];
			sprintf(name,"%s",GetBsp()->m_pTextures[i].strName);
			shaders[i]=resMan->GetPtr(sptr<CRenderAble>(),name, RESTYP_SHADER);
		}

		for (int i=0; i<m_numOfFaces; i++)
		{
			GENERAL::tBSPFace &f = m_pFaces[i];
			if (shaders[f.textureID]->GetShaderStatus()==CResShader::EShaderStatus::SS_NOTLOADED)
			{
				GENERAL::dprintf("[ INFO ] shader %s not found! Using default shader\n",GetBsp()->m_pTextures[f.textureID].strName);
				if (f.lightmapID>=0)
				{
					shaders[f.textureID]->LoadDefaultShaderLightMap(GetBsp()->m_pTextures[f.textureID].strName);
				} else {
					shaders[f.textureID]->LoadDefaultShader(GetBsp()->m_pTextures[f.textureID].strName);
				}
			}
		}
	}
	
	CObjQuakeBsp::~CObjQuakeBsp()
	{
		for (int i=0; i<m_numOfPatches; i++)
		{
			SAFE_FREE(m_pPatches[i].vertices);
			SAFE_FREE(m_pPatches[i].indices);
		}

		SAFE_DELETE_AR(shaders)
		SAFE_DELETE_AR(m_pShaders);
		SAFE_DELETE_AR(m_pPatches);
		SAFE_FREE(m_pVerts);	
		SAFE_FREE(m_pFaces);	
		SAFE_FREE(m_pLightmaps);
		SAFE_FREE(m_pMeshIndices);

	}

	void CObjQuakeBsp::LoadDataFromFile()
	{
		GENERAL::CQuakeBSPFile qbsp;
		if (!qbsp.Open(GetBsp()->m_szFname))
			return;

		m_pVerts = qbsp.GetVertices(m_numOfVerts);
		m_pFaces = qbsp.GetFaces(m_numOfFaces);
		m_pLightmaps = qbsp.GetLightMaps(m_numOfLightmaps);
		m_pMeshIndices = qbsp.GetMeshIndices(m_numOfMeshIndices);
		m_pShaders = qbsp.GetShaders(m_numOfShaders);

/*		GENERAL::tBSPHeader header = {0};
		GENERAL::tBSPLump lumps[DOMAIN::CQuakeBsp::kMaxLumps] = {0};

		fread(&header, 1, sizeof(tBSPHeader), fp);
		fread(&lumps, kMaxLumps, sizeof(tBSPLump), fp);*/
	}

	void CObjQuakeBsp::Tesselate(int tesselFactor)
	{
		int i;
		// generate patches
		m_numOfPatches=0;
		for (i = 0; i<m_numOfFaces; i++)
		{
			if (m_pFaces[i].type == FACE_PATCH) 
				m_numOfPatches++;
		}

		m_pPatches = new TPatch[m_numOfPatches];
		TPatch *cp = m_pPatches;

		for (i = 0; i<m_numOfFaces; i++)
		{
			if (m_pFaces[i].type != FACE_PATCH)
				continue; // skip face if not a patch

			int num_u = (m_pFaces[i].size[0]-1)/2;
			int num_v = (m_pFaces[i].size[1]-1)/2;

			TPatch::TBiQaudricBezier *p = (TPatch::TBiQaudricBezier*)
				malloc(num_u*num_v*sizeof(TPatch::TBiQaudricBezier));
			
			int u,v;

			for (v=0; v<num_v; v++)
			{
				for (u=0; u<num_u; u++)
				{
					int cp=m_pFaces[i].startVertIndex+
						(v*2+1)*m_pFaces[i].size[0]+2*u+1;
					p[v*num_u+u].upatch[0].controlPoints[0] = m_pVerts[cp-1-(v*2+1)*m_pFaces[i].size[0]];
					p[v*num_u+u].upatch[0].controlPoints[1] = m_pVerts[cp-0-(v*2+1)*m_pFaces[i].size[0]];
					p[v*num_u+u].upatch[0].controlPoints[2] = m_pVerts[cp+1-(v*2+1)*m_pFaces[i].size[0]];
					p[v*num_u+u].upatch[1].controlPoints[0] = m_pVerts[cp-1];
					p[v*num_u+u].upatch[1].controlPoints[1] = m_pVerts[cp-0];
					p[v*num_u+u].upatch[1].controlPoints[2] = m_pVerts[cp+1];
					p[v*num_u+u].upatch[2].controlPoints[0] = m_pVerts[cp-1+(v*2+1)*m_pFaces[i].size[0]];
					p[v*num_u+u].upatch[2].controlPoints[1] = m_pVerts[cp-0+(v*2+1)*m_pFaces[i].size[0]];
					p[v*num_u+u].upatch[2].controlPoints[2] = m_pVerts[cp+1+(v*2+1)*m_pFaces[i].size[0]];					}
			}

			for (v=0; v<num_v; v++)
			{
				for (u=0; u<num_u; u++)
				{
					int ui,vi;

					int index=m_pFaces[i].startVertIndex+v*2*m_pFaces[i].size[0]+2*u;

					for (vi=0;vi<3;vi++)
					{
						for (ui=0;ui<3;ui++)
						{
							p[v*num_u+u].upatch[vi].controlPoints[ui]=
								m_pVerts[index+vi*m_pFaces[i].size[0]+ui];
						}
					}
				}
			}
			cp->sizeu=num_u*tesselFactor+1;
			cp->sizev=num_v*tesselFactor+1;

			cp->numOfVerts = cp->sizeu*cp->sizev;
			cp->vertices = (GENERAL::tBSPVertex*)malloc(sizeof(GENERAL::tBSPVertex)*cp->numOfVerts);

			int vi=0;

			for (v=0; v<cp->sizev; v++)
			{
				for (u=0; u<cp->sizeu; u++)
				{
					float up=((float)(u%tesselFactor))/(float)tesselFactor;
					float vp=((float)(v%tesselFactor))/(float)tesselFactor;

					int pnv = v/tesselFactor;
					int pnu = u/tesselFactor;
					if (pnv==num_v) 
					{
						pnv--;
						vp=1.0f;
					}
					if (pnu==num_u) 
					{
						pnu--;
						up=1.0f;
					}

					int pn = pnv*num_u+pnu;

					p[pn].Tesselate(cp->vertices[vi],up,vp);
					vi++;
				}
				
			}

			free(p);

			cp->indices = (int*)malloc(sizeof(int)*(cp->sizev-1)*(cp->sizeu-1)*2*3);
			cp->numOfIndices=0;
			for (int v=0;v<cp->sizev-1; v++)
			{
				for (int u=0;u<cp->sizeu-1; u++)
				{
					cp->indices[cp->numOfIndices+0]=v*cp->sizeu+u;
					cp->indices[cp->numOfIndices+1]=(v+1)*cp->sizeu+u+1;
					cp->indices[cp->numOfIndices+2]=v*cp->sizeu+u+1;
					cp->indices[cp->numOfIndices+3]=v*cp->sizeu+u;
					cp->indices[cp->numOfIndices+4]=(v+1)*cp->sizeu+u;
					cp->indices[cp->numOfIndices+5]=(v+1)*cp->sizeu+u+1;
					cp->numOfIndices+=6;
				}
			}
				
			m_pFaces[i].startVertIndex = (int)(cp-m_pPatches);
			cp++;
		}
	}

	////////////////////////////////////
	CObjQuakeBspGL::CObjQuakeBspGL(CGLFrontEnd *_fe, CResManagerGL *_resMan, DOMAIN::CQuakeBsp *_bsp) : CObjQuakeBsp(_fe, _resMan, _bsp) 
	{
		pFaceRendered = (bool*)malloc(sizeof(bool)*m_numOfFaces);
		pClusterRendered = (bool*)malloc(sizeof(bool)*(GetBsp()->m_clusters.numOfClusters));
//		pLightMaps = (GLuint*)malloc(sizeof(int*)*m_numOfLightmaps);
		pLightMaps = (CResTextureGL**)malloc(sizeof(CResTextureGL*)*m_numOfLightmaps);


		for (int i=0; i<m_numOfLightmaps; i++)
		{
			ChangeGamma(4,(GLubyte*)m_pLightmaps[i].imageBits);
			void *data = malloc(128*128*3);
			memcpy(data,m_pLightmaps[i].imageBits,128*128*3);
			pLightMaps[i]=new CResTextureGL(_resMan,NULL,3,GL_UNSIGNED_BYTE,128,128,data);


/*			glGenTextures(1,&pLightMaps[i]);
			glPixelStorei(GL_UNPACK_ALIGNMENT,1);
			glBindTexture(GL_TEXTURE_2D,pLightMaps[i]);
			gluBuild2DMipmaps(GL_TEXTURE_2D,3,128,128,GL_RGB,GL_UNSIGNED_BYTE,m_pLightmaps[i].imageBits);
			glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE,GL_MODULATE);*/
		}


//		surfaces = (TSurface*)malloc(m_numOfFaces*sizeof(TSurface));
		surfaces = new TSurface[m_numOfFaces];
//		memset(surfaces,0,sizeof(TSurface)*m_numOfFaces);

		for (int i=0; i<m_numOfFaces; i++)
		{
			GENERAL::tBSPFace &f = m_pFaces[i];
			surfaces[i].modelView=NULL;

			surfaces[i].shader = shaders[f.textureID];
			surfaces[i].lightMap = f.lightmapID>=0?pLightMaps[f.lightmapID]:NULL;
			surfaces[i].name=0;
			surfaces[i].flags=TSurface::SF_DYNAMIC_LIGHT;

			if ( (f.type==FACE_POLYGON) || (f.type==FACE_MESH) )
			{
				surfaces[i].numVertices = f.numOfVerts;
				surfaces[i].vertices = (CVector3f*)malloc(f.numOfVerts*sizeof(CVector3f));
				surfaces[i].normals = (CVector3f*)malloc(f.numOfVerts*sizeof(CVector3f));
				surfaces[i].texCoords = (CVector2f*)malloc(f.numOfVerts*sizeof(CVector2f));
				surfaces[i].lightMapTexCoords = (CVector2f*)malloc(f.numOfVerts*sizeof(CVector2f));
				surfaces[i].colors = (TColor4ub*)malloc(f.numOfVerts*sizeof(TColor4ub));

				GENERAL::tBSPVertex *v = m_pVerts+f.startVertIndex;
				for (int n=0; n<f.numOfVerts; n++)
				{
					surfaces[i].vertices[n]=v->vPosition;
					surfaces[i].normals[n]=v->vNormal;
					surfaces[i].texCoords[n].x=v->vTextureCoord.x;
					surfaces[i].texCoords[n].y=v->vTextureCoord.y;
					surfaces[i].lightMapTexCoords[n].x=v->vLightmapCoord.x;
					surfaces[i].lightMapTexCoords[n].y=v->vLightmapCoord.y;
					surfaces[i].colors[n]=v->color;
					v++;
				}

				if (f.type==FACE_POLYGON)
				{
					surfaces[i].type=TSurface::SURFACETYPE_POLYGON;
					surfaces[i].primitiveType = GL_TRIANGLE_FAN;
				} else { 
					surfaces[i].type=TSurface::SURFACETYPE_MESH;
					surfaces[i].primitiveType = GL_TRIANGLES;
					surfaces[i].mesh.numIndices=f.numMeshVerts;
					surfaces[i].mesh.indices=(int*)malloc(f.numMeshVerts*sizeof(int));
					for (int n=0; n<f.numMeshVerts; n++)
					{
						surfaces[i].mesh.indices[n]=m_pMeshIndices[f.meshVertIndex+n];
					}
				}
			} else if (f.type==FACE_PATCH)
			{
				TPatch &p = m_pPatches[f.startVertIndex];

				surfaces[i].type=TSurface::SURFACETYPE_MESH;
				surfaces[i].primitiveType = GL_TRIANGLES;

				surfaces[i].numVertices = p.numOfVerts;
				surfaces[i].vertices = (CVector3f*)malloc(p.numOfVerts*sizeof(CVector3f));
				surfaces[i].normals = (CVector3f*)malloc(p.numOfVerts*sizeof(CVector3f));
				surfaces[i].texCoords = (CVector2f*)malloc(p.numOfVerts*sizeof(CVector2f));
				surfaces[i].lightMapTexCoords = (CVector2f*)malloc(p.numOfVerts*sizeof(CVector2f));
				surfaces[i].colors = (TColor4ub*)malloc(p.numOfVerts*sizeof(TColor4ub));

				GENERAL::tBSPVertex *v = p.vertices;
				for (int n=0; n<p.numOfVerts; n++)
				{
					surfaces[i].vertices[n]=v->vPosition;
					surfaces[i].normals[n]=v->vNormal;
					surfaces[i].texCoords[n].x=v->vTextureCoord.x;
					surfaces[i].texCoords[n].y=v->vTextureCoord.y;
					surfaces[i].lightMapTexCoords[n].x=v->vLightmapCoord.x;
					surfaces[i].lightMapTexCoords[n].y=v->vLightmapCoord.y;
					surfaces[i].colors[n]=v->color;
					v++;
				}		

				surfaces[i].mesh.numIndices=p.numOfIndices;
				surfaces[i].mesh.indices=(int*)malloc(p.numOfIndices*sizeof(int));
				memcpy(surfaces[i].mesh.indices,p.indices,p.numOfIndices*sizeof(int));
			} else {
//				DebugBreak();
				surfaces[i].type=TSurface::SURFACETYPE_UNKNOWN;
				surfaces[i].primitiveType = GL_TRIANGLES;
				surfaces[i].numVertices=0;
			}
		}
	};

	
	CObjQuakeBspGL::~CObjQuakeBspGL() 
	{
//		glDeleteTextures(m_numOfLightmaps,pLightMaps);

		for (int i=0; i<m_numOfFaces; i++)
		{
			GENERAL::tBSPFace &f = m_pFaces[i];
			if (f.type!=FACE_BILLBORD)
			{
				GENERAL::tBSPFace &f = m_pFaces[i];
				free(surfaces[i].vertices);
				free(surfaces[i].normals);
				free(surfaces[i].texCoords);
				free(surfaces[i].lightMapTexCoords);
				free(surfaces[i].colors);
				if (f.type!=FACE_POLYGON)
				{
					free(surfaces[i].mesh.indices);
				}
			}
		}

		SAFE_DELETE_AR(surfaces);

		for (int i=0; i<m_numOfLightmaps; i++)
			delete pLightMaps[i];
		free(pLightMaps);
		free(pFaceRendered);
		free(pClusterRendered);
	}

	void CObjQuakeBspGL::ChangeGamma(float factor, GLubyte *pImage)
	{
		for(int i = 0; i < m_numOfVerts; i++)
		{
			float scale = 1.0f, temp = 0.0f;
			float r = 0, g = 0, b = 0;

			// extract the current RGB values
			r = (float)m_pVerts[i].color.r;
			g = (float)m_pVerts[i].color.g;
			b = (float)m_pVerts[i].color.b;

			// Multiply the factor by the RGB values, while keeping it to a 255 ratio
			r = r * factor / 255.0f;
			g = g * factor / 255.0f;
			b = b * factor / 255.0f;
			
			// Check if the the values went past the highest value
			if(r > 1.0f && (temp = (1.0f/r)) < scale) scale=temp;
			if(g > 1.0f && (temp = (1.0f/g)) < scale) scale=temp;
			if(b > 1.0f && (temp = (1.0f/b)) < scale) scale=temp;

			// Get the scale for this pixel and multiply it by our pixel values
			scale*=255.0f;		
			r*=scale;	g*=scale;	b*=scale;

			// Assign the new gamma'nized RGB values to our image
			m_pVerts[i].color.r = (byte)r;
			m_pVerts[i].color.g = (byte)g;
			m_pVerts[i].color.b = (byte)b;
		}
		for(int i = 0; i < 128*128; i++, pImage += 3) 
		{
			float scale = 1.0f, temp = 0.0f;
			float r = 0, g = 0, b = 0;

			// extract the current RGB values
			r = (float)pImage[0];
			g = (float)pImage[1];
			b = (float)pImage[2];

			// Multiply the factor by the RGB values, while keeping it to a 255 ratio
			r = r * factor / 255.0f;
			g = g * factor / 255.0f;
			b = b * factor / 255.0f;
			
			// Check if the the values went past the highest value
			if(r > 1.0f && (temp = (1.0f/r)) < scale) scale=temp;
			if(g > 1.0f && (temp = (1.0f/g)) < scale) scale=temp;
			if(b > 1.0f && (temp = (1.0f/b)) < scale) scale=temp;

			// Get the scale for this pixel and multiply it by our pixel values
			scale*=255.0f;		
			r*=scale;	g*=scale;	b*=scale;

			// Assign the new gamma'nized RGB values to our image
			pImage[0] = (byte)r;
			pImage[1] = (byte)g;
			pImage[2] = (byte)b;
		}
	}

	int CObjQuakeBspGL::FindLeaf(GENERAL::CVector3f &pos)
	{
		GENERAL::tBSPNode *nodes=GetBsp()->m_pNodes;

		int current=0;
		float distance;

		while (current>=0)
		{
			GENERAL::tBSPPlane &plane=GetBsp()->m_pPlanes[nodes[current].plane];
			distance = pos.ScalarMultiply(plane.vNormal)-plane.d;

			if (distance>=0)
			{
				// i am in front of the plane
				current=nodes[current].front;
			} else {
				// i am in back of the plane
				current=nodes[current].back;
			}
		}

		return ~current;
	}

	void CObjQuakeBspGL::RenderLeaf(GENERAL::tBSPLeaf &leaf, int _dt)
	{
		int faceCount = leaf.numOfLeafFaces;

		while (faceCount--)
		{
			int faceIndex = GetBsp()->m_pLeafFaces[leaf.leafface+faceCount];

			if (!pFaceRendered[faceIndex])
			{
				if (selectMode==RENDERSELECT_NORMAL)
					selectName=domainObj->GetID();
				if (selectMode==RENDERSELECT_DETAIL)
					selectName=faceIndex;

				surfaces[faceIndex].name=selectName;

				if (m_pFaces[faceIndex].type!=FACE_BILLBORD)
					((CGLFrontEnd*)fe)->GetRenderer().PushSurface(surfaces+faceIndex,false, true);
				pFaceRendered[faceIndex]=true;
			}
		}

		if (leaf.cluster!=-1)
		{
			if (!pClusterRendered[leaf.cluster]) 
			{
				if (((DOMAIN::CQuakeBsp*) GetDomainObj())->m_pClusterChilds[leaf.cluster]->GetChilds().size()>0)
				{
					int blu=1;
				}
				((DOMAIN::CQuakeBsp*) GetDomainObj())->m_pClusterChilds[leaf.cluster]->GetfeData()->Render(_dt);
				pClusterRendered[leaf.cluster]=true;
			}
		}

	}

	void CObjQuakeBspGL::Render(int _dt)
	{



		faces_renderd = 0;

		glPushAttrib(GL_ALL_ATTRIB_BITS);
		glPushMatrix();
		memset(pFaceRendered,false,m_numOfFaces);
		memset(pClusterRendered,false,GetBsp()->m_clusters.numOfClusters);

		int cluster = GetBsp()->m_pLeafs[FindLeaf(((CGLFrontEnd*)fe)->GetCamera().GetPos())].cluster;

//		char strr[50];
//		sprintf(strr,"cluster: %i\n", cluster); 
//		GENERAL::dprintf(strr);

/*		int i = m_numOfFaces;
		while (i--)
		{
			RenderFace(m_pFaces[i]);
		}*/

		int i = GetBsp()->m_numOfLeafs;

		while (i--)
		{
			GENERAL::tBSPLeaf *leaf = &GetBsp()->m_pLeafs[i];

			if (!GetBsp()->IsClusterVisible(cluster,leaf->cluster))
				continue;

			if (!((CGLFrontEnd*)fe)->GetCamera().GetFrustum().IsBoxInside(
					GENERAL::CVector3f((GENERAL::SCALAR)leaf->min.x,(GENERAL::SCALAR)leaf->min.y,(GENERAL::SCALAR)leaf->min.z),
					GENERAL::CVector3f((GENERAL::SCALAR)leaf->max.x,(GENERAL::SCALAR)leaf->max.y,(GENERAL::SCALAR)leaf->max.z)))
				continue;

			RenderLeaf(GetBsp()->m_pLeafs[i], _dt);
		}

		//CRenderAble::Render(_dt);


		//((CGLFrontEnd*)fe)->GetRenderer().Flush();

		glPopMatrix();
		glPopAttrib();

	}

	void CObjQuakeBspGL::RenderFace(GENERAL::tBSPFace &face)
	{
		//faces_renderd++;
		//does nothing not needed anymore
	}

	bool CObjQuakeBspGL::MouseDown(EMouseButton button, int x,int y,int z,TMouseButtons &buttons,TKeyStates &keys)
	{
		int faceIndex=((CGLFrontEnd*)fe)->RenderSelectDetail(x,y,this);
		CResShader *shad = &shaders[m_pFaces[faceIndex].textureID];
		GENERAL::dprintf("[ INFO ] CObjQuakeBspGL MouseDown button(%ld), objectID(%ld), \n",button,domainObj->GetID());
		GENERAL::dprintf("	       Detail face(%d), shader(%s), sort(%d)\n",faceIndex,shad->shadername,shad->sortvalue);
		return false;
	}

	void CObjQuakeBspGL::RenderLeafBrushes(CVector3f &pos)
	{
		GENERAL::tBSPLeaf &leaf=GetBsp()->m_pLeafs[GetBsp()->FindLeaf(pos)];
		for (int i=0; i<leaf.numOfLeafBrushes; i++)
		{
			GENERAL::tBSPBrush &brush=GetBsp()->m_pBrushes[GetBsp()->m_pLeafBrushes[leaf.leafBrush+i]];
			for (int n=0; n<brush.numOfBrushSides; n++)
			{
				GENERAL::tBSPBrushSide &brushside=GetBsp()->m_pBrushSides[brush.brushSide+n];
				((CGLFrontEnd*)fe)->GetRenderer().RenderPlane(
					TColor4ub(255,255,0,20),
					GetBsp()->m_pPlanes[brushside.plane].vNormal,
					GetBsp()->m_pPlanes[brushside.plane].d);
			}
		}
	}


}
