#include "shader.h"
#include <stdio.h>
#include <sys/stat.h>

#include <glheaders.h>
#include "../implementation/resman.h"

namespace FRONTEND
{
	#define SHADERPATH "scripts/"

/*******************************************************************************************************************/
/* CResShader                                                                                                      */
/*******************************************************************************************************************/
	CResShader::~CResShader()
	{
		for (int i=0; i<num_layers; i++)
		{
			while (layer[i]->tcMod)
			{
				funclist *f=layer[i]->tcMod->next;
				delete layer[i]->tcMod;
				layer[i]->tcMod=f;
			}
			while (layer[i]->aGen)
			{
				funclist *f=layer[i]->aGen->next;
				delete layer[i]->aGen;
				layer[i]->aGen=f;
			}

			delete layer[i];
		}

		for (i=0; i<sky_params.num_layers; i++)
		{
			delete sky_params.layer[i][0];
			delete sky_params.layer[i][1];
			delete sky_params.layer[i][2];
			delete sky_params.layer[i][3];
			delete sky_params.layer[i][4];
			delete sky_params.layer[i][5];
		}

		if (fog_params.layer)
		{
			delete fog_params.layer;
		}
	}

	CResShader::Layer *CResShader::CreateLayer()
	{
		Layer *layer = new Layer();
//		if (!layer) ThrowException(ALLOCATION_ERROR, "LayerManager::CreateLayer.layer");

		layer->flags = 0;
		layer->num_maps = 0;
		layer->anim_speed = 0;
		
//SPTR		for (int i = 0; i < MAX_ANIMFRAMES; ++i) layer->map[i] = NULL;

		layer->tcMod = NULL;
		layer->aGen = NULL;

		layer->alphaGen = ALPHAGEN_NONE;
		layer->alphaGenParams[0] = 0;

		layer->tcGen = TCGEN_BASE;
		layer->rgbGen = RGBGEN_IDENTITY;

		layer->alphafunc = GL_ALWAYS;
		layer->alphafuncref = 1.0f;

		layer->blendsrc = GL_ONE;
		layer->blenddst = GL_ZERO;

		layer->depthFunc = GL_LEQUAL;
		layer->depthWrite = true;
		
		layer->frame=0;

		return layer;
	}

/*******************************************************************************************************************/
/* CResShaderGL                                                                                                    */
/*******************************************************************************************************************/
	CResShaderGL::CResShaderGL(CResManagerGL *_resMan,char *_file, int _targetstatus)
		: CResShader((CResourceManager*)_resMan,_file,_targetstatus) 
	{
		sortvalue = SORT_NORMAL;
		shaderStatus=SS_NOTLOADED;
		if (_targetstatus!=RESSTATUS_NOTLOADED) LoadFromFile(_file, _targetstatus);
	}

	CResShaderGL::~CResShaderGL()
	{
	}

	bool CResShaderGL::LoadFromFile(char *_file, int _targetstatus)
	{
		if (shaderStatus!=SS_NOTLOADED)
			return true;

//		bool shaderLoaded=false;
		char *fn=strdup(_file);
		char *token;
		std::vector<char *> tokens;
		tokens.clear();

		int i=0;
		token = strtok(fn,"/\\");
		while (token)
		{
			i++;
			tokens.push_back(token);
			token=strtok(NULL,"/\\");
		}

		if (i>1)
		{
			char *shaderfile=(char*)malloc(strlen(tokens[i-2])+strlen(".shader")+strlen(SHADERPATH)+1);
			sprintf(shaderfile,"%s%s.shader",SHADERPATH,tokens[i-2]);

			CShaderFile sf(shaderfile);
			if (sf.LoadShader(this,_file))
			{
				shaderStatus=SS_LOADED;
			}

			free(shaderfile);
		}

		free(fn);
		return true;
	}

	bool CResShaderGL::LoadDefaultShader(char *_file)
	{
		if (shaderStatus!=SS_NOTLOADED)
			return true;

		flags = 0;
		sortvalue=SORT_OPAQUE;

		num_layers = 1;
		SAFE_DELETE(layer[0]);
		layer[0] = CreateLayer();
		layer[0]->depthWrite = 1;
		layer[0]->blendsrc = CShaderParser::IntForConst("GL_ONE");
		layer[0]->blenddst = CShaderParser::IntForConst("GL_ZERO");
		layer[0]->flags = 0;
		layer[0]->rgbGen = RGBGEN_VERTEX;
		char fname[512];
		sprintf(fname,"%s.tga",_file);
		layer[0]->map[0] = resMan->GetPtr(sptr<CRenderAble>(),fname,RESTYP_TEXTURE);

		shaderStatus = SS_DEFAULT;
		return true;
	}

	bool CResShaderGL::LoadDefaultShaderGUI(char *_file)
	{
		if (shaderStatus!=SS_NOTLOADED)
			return true;

		flags = 0;
		sortvalue=SORT_OPAQUE;

		num_layers = 1;
		SAFE_DELETE(layer[0]);
		layer[0] = CreateLayer();
		layer[0]->depthWrite = 1;
		layer[0]->blendsrc = CShaderParser::IntForConst("GL_SRC_ALPHA");
		layer[0]->blenddst = CShaderParser::IntForConst("GL_ONE_MINUS_SRC_ALPHA");
		layer[0]->flags = LF_ELEMENT;
		layer[0]->tcGen = TCGEN_ELEMENT;
		layer[0]->rgbGen = RGBGEN_IDENTITY;
		char fname[512];
		sprintf(fname,"%s.tga",_file);
		
		TElementInfo ei;
		
		ei.tex = resMan->GetPtr(sptr<CRenderAble>(),fname,RESTYP_TEXTURE);
		ei.texCoords[0] = TMorfPointf(0.0f,1.0f);
		ei.texCoords[1] = TMorfPointf(0.0f,0.0f);
		ei.texCoords[2] = TMorfPointf(1.0f,0.0f);
		ei.texCoords[3] = TMorfPointf(1.0f,1.0f);
		elements.push_back(ei);

		shaderStatus = SS_DEFAULT_GUI;
		return true;
	}

	bool CResShaderGL::LoadDefaultShaderLightMap(char *_file)
	{
		if (shaderStatus!=SS_NOTLOADED)
			return true;

		sortvalue=SORT_OPAQUE;
		flags = SF_HASLIGHTMAP | SF_MULTITEX;
		
		num_layers = 2;
		SAFE_DELETE(layer[0]);
		layer[0] = CreateLayer();
		layer[0]->depthWrite = 1;
		layer[0]->blendsrc = CShaderParser::IntForConst("GL_ONE");
		layer[0]->blenddst = CShaderParser::IntForConst("GL_ZERO");
		layer[0]->flags = LF_LIGHTMAP | LF_NOMIPMAP | LF_NOPICMIP | LF_CLAMP | LF_MULTITEX;
		layer[0]->tcGen = TCGEN_LIGHTMAP;
		layer[0]->rgbGen = RGBGEN_IDENTITY;
		
		SAFE_DELETE(layer[1]);
		layer[1] = CreateLayer();
		layer[1]->depthWrite = 1;
		layer[1]->blendsrc = CShaderParser::IntForConst("GL_DST_COLOR");
		layer[1]->blenddst = CShaderParser::IntForConst("GL_ZERO");
		layer[1]->flags = 0;
		layer[1]->rgbGen = RGBGEN_IDENTITY;


		sky_params.num_layers = 0;

		// load texture into memory
		// original	r->shader->layer[1]->map[0] = textures.AddTexinfo(r->name, r->shader->layer[1]->flags);
		char fname[512];
		sprintf(fname,"%s.tga",_file);
		layer[1]->map[0] = resMan->GetPtr(sptr<CRenderAble>(),fname,RESTYP_TEXTURE);
		
		shaderStatus = SS_DEFAULT_LIGHTMAP;
		return true;
	}

	bool CResShaderGL::LoadIntoGraphicAdapter(int _targetsatus)
	{
		return true;
	}

	bool CResShaderGL::UnloadFromGraphicAdapter()
	{
		return true;
	}

/*******************************************************************************************************************/
/* CShaderFile                                                                                                     */
/*******************************************************************************************************************/
	CShaderFile::CShaderFile(char *_fileName)
	{
		fileName = strdup(_fileName);

		// find shaderfile
		char *fn = GENERAL::morfFindFile(fileName);

		if (fn!=NULL)
		{
			struct _stat fileStat;
			_stat(fn,&fileStat);
			fileSize = fileStat.st_size;

			FILE *f = GENERAL::morfOpenFile(fileName,"rb");
			fileContent = (char*)malloc(fileSize);
			fread(fileContent,1,fileSize,f);
			fclose(f);

			CShaderParser::StartParseBuffer(fileContent, (const int)fileSize);
		} else {
			fileContent=NULL;
			fileSize=NULL;
		}
	}

	CShaderFile::~CShaderFile()
	{
		if (fileName!=NULL) free(fileName);
		if (fileContent!=NULL) free(fileContent);
	}

	struct constants
	{
		char *cad;
		int value;
	};

	#define NUM_CONSTANTS	68
	const struct constants clist[128] =
	{
		{ "FALSE",									0		},
		{ "GL_ALWAYS",								0x0207	},
		{ "GL_BACK",								0x0405	},
		{ "GL_CLAMP",								0x2900	},
		{ "GL_CURRENT_FOG_COORDINATE_EXT",			0x8453	},
		{ "GL_DECAL",								0x2101	},
		{ "GL_DECR",								0x1E03	},
		{ "GL_DST_ALPHA",							0x0304	},
		{ "GL_DST_COLOR",							0x0306	},
		{ "GL_EQUAL",								0x0202	},
		{ "GL_EYE_LINEAR",							0x2400	},
		{ "GL_EYE_PLANE",							0x2502	},
		{ "GL_FALSE",								0		},
		{ "GL_FLAT",								0x1D00	},
		{ "GL_FOG_COORDINATE_ARRAY_EXT",			0x8457	},
		{ "GL_FOG_COORDINATE_ARRAY_POINTER_EXT",	0x8456	},
		{ "GL_FOG_COORDINATE_ARRAY_STRIDE_EXT",		0x8455	},
		{ "GL_FOG_COORDINATE_ARRAY_TYPE_EXT",		0x8454	},
		{ "GL_FOG_COORDINATE_EXT",					0x8451	},
		{ "GL_FOG_COORDINATE_SOURCE_EXT",			0x8450	},
		{ "GL_FRAGMENT_DEPTH_EXT",					0x8452	},
		{ "GL_FRONT",								0x0404	},
		{ "GL_FRONT_AND_BACK",						0x0408	},
		{ "GL_GEQUAL",								0x0206	},
		{ "GL_GREATER",								0x0204	},
		{ "GL_INCR",								0x1E02	},
		{ "GL_KEEP",								0x1E00	},
		{ "GL_LEQUAL",								0x0203	},
		{ "GL_LESS",								0x0201	},
		{ "GL_LINEAR",								0x2601	},
		{ "GL_LINEAR_MIPMAP_LINEAR",				0x2703	},
		{ "GL_LINEAR_MIPMAP_NEAREST",				0x2701	},
		{ "GL_MODULATE",							0x2100	},
		{ "GL_NEAREST",								0x2600	},
		{ "GL_NEAREST_MIPMAP_LINEAR",				0x2702	},
		{ "GL_NEAREST_MIPMAP_NEAREST",				0x2700	},
		{ "GL_NEVER",								0x0200	},
		{ "GL_NOTEQUAL",							0x0205	},
		{ "GL_OBJECT_LINEAR",						0x2401	},
		{ "GL_OBJECT_PLANE",						0x2501	},
		{ "GL_ONE",									1		},
		{ "GL_ONE_MINUS_DST_ALPHA",					0x0305	},
		{ "GL_ONE_MINUS_DST_COLOR",					0x0307	},
		{ "GL_ONE_MINUS_SRC_ALPHA",					0x0303	},
		{ "GL_ONE_MINUS_SRC_COLOR",					0x0301	},
		{ "GL_REPEAT",								0x2901	},
		{ "GL_REPLACE",								0x1E01	},
		{ "GL_SMOOTH",								0x1D01	},
		{ "GL_SPHERE_MAP",							0x2402	},
		{ "GL_SRC_ALPHA",							0x0302	},
		{ "GL_SRC_ALPHA_SATURATE",					0x0308	},
		{ "GL_SRC_COLOR",							0x0300	},
		{ "GL_TEXTURE_ENV",							0x2300	},
		{ "GL_TEXTURE_ENV_COLOR",					0x2201	},
		{ "GL_TEXTURE_ENV_MODE",					0x2200	},
		{ "GL_TEXTURE_GEN_MODE",					0x2500	},
		{ "GL_TEXTURE_MAG_FILTER",					0x2800	},
		{ "GL_TEXTURE_MIN_FILTER",					0x2801	},
		{ "GL_TEXTURE_WRAP_S",						0x2802	},
		{ "GL_TEXTURE_WRAP_T",						0x2803	},
		{ "GL_TRUE",								1		},
		{ "GL_ZERO",								0		},
		{ "NV_EYE_PLANE_ABSOLUTE_EXT",				0x855C	},
		{ "NV_EYE_RADIAL_EXT",						0x855B	},
		{ "NV_FOG_DISTANCE_MODE_EXT",				0x855A	},
		{ "TRUE",									1		},
		{ "ZZZZ",									0		},
		{ "ZZZZ",									0		}
	};

	int CShaderParser::IntForConst(const char *name)
	{
		int prin = 0;
		int fin = NUM_CONSTANTS;
		int medio;
		int dir;

		while (prin <= fin)
		{
			medio = (prin + fin) >> 1;
			
			if ( medio >= NUM_CONSTANTS || medio < 0 ) return 0;

			dir = stricmp(clist[medio].cad, name);
			if (dir < 0) prin = medio+1;
			else if (dir > 0) fin = medio-1;
			else return clist[medio].value;
		}

		GENERAL::dprintf("^5WARNING: Unknown constant \"%s\". Returning 0.", name);
		return 0;
	}

	char *CShaderFile::GetShaderName()
	{
		return CShaderParser::GetToken(true)?CShaderParser::token:0;
	}


	bool CShaderFile::FindShader(char *_shaderName)
	{
		char *shaderName = GetShaderName();
		while (shaderName!=NULL)
		{
			if (stricmp(shaderName,_shaderName)==0)
				return true;
			if (!SkipShader()) return false;
			char *shaderName = GetShaderName();
		}
		return false;
	}

	bool CShaderFile::LoadShader(CResShader *shad,char *_shaderName)
	{

		sprintf(shad->shadername,"%s",_shaderName);		
		
		if (fileContent==NULL)
			return false; 

		if (!FindShader(_shaderName))
			return false;


		shad->num_layers = 0;
		shad->sky_params.num_layers = 0;

		if (CShaderParser::GetToken(true))
		{
			if (strcmp(CShaderParser::token, "{"))
			{
				GENERAL::dprintf("^5WARNING: '{' expected in line %d (found \"%s\")", CShaderParser::scriptline, CShaderParser::token);
				return false;
			}
		}

		while (CShaderParser::GetToken(true))	// while not EOF
		{
			if (!strcmp(CShaderParser::token, "{"))
			{
				shad->layer[shad->num_layers] = shad->CreateLayer();
				ParseLayer(shad, shad->layer[shad->num_layers]);
				++shad->num_layers;

				continue;
			}

			else if (!strcmp(CShaderParser::token, "}")) break;	// ends shader
			
			else if (!stricmp(CShaderParser::token, "cull"))
			{
				if (!CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
					continue;
				}

				if (!stricmp(CShaderParser::token, "front")) shad->cull = CULL_FRONT;
				else if (!stricmp(CShaderParser::token, "back")) shad->cull = CULL_BACK;
				else if (!stricmp(CShaderParser::token, "none") ||
						!stricmp(CShaderParser::token, "disable") ||
						!stricmp(CShaderParser::token, "twosided")) shad->cull = CULL_NONE;
				else
				{
					// unknown parameter !!
					GENERAL::dprintf("^5WARNING: unknown parameter \"%s\" in line %d", CShaderParser::token, CShaderParser::scriptline);
				}
			}
			else if (!stricmp(CShaderParser::token, "font"))
			{
				if (!CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
					continue;
				}
				shad->fnt= shad->GetResourceManager()->GetPtr(sptr<CRenderAble>(),CShaderParser::token,RESTYP_BMPFONT); 
			}
			else if (!stricmp(CShaderParser::token, "fontcolor"))
			{
				char *eptr;
				if (!CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
					continue;
				}
				shad->fntColor.r = (GENERAL::UBYTE)strtol(CShaderParser::token,&eptr,0);
				if (!CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
					continue;
				}
				shad->fntColor.g = (GENERAL::UBYTE)strtol(CShaderParser::token,&eptr,0);
				if (!CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
					continue;
				}
				shad->fntColor.b = (GENERAL::UBYTE)strtol(CShaderParser::token,&eptr,0);
				if (!CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
					continue;
				}
				shad->fntColor.a = (GENERAL::UBYTE)strtol(CShaderParser::token,&eptr,0);
			}
			else if (!stricmp(CShaderParser::token, "nopicmip"))
			{
				shad->flags |= SF_NOPICMIP;
			}
			
			else if (!stricmp(CShaderParser::token, "nomipmaps"))
			{
				shad->flags |= SF_NOMIPMAP;
			}
			
			else if (!stricmp(CShaderParser::token, "polygonOffset"))
			{
				shad->flags |= SF_POLYGONOFFSET;
			}
			
			else if (!stricmp(CShaderParser::token, "portal"))
			{
				shad->flags |= SF_PORTAL;
				shad->sortvalue = SORT_PORTAL;
			}
			
			else if (!stricmp(CShaderParser::token, "sort"))
			{
				if (!CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
					continue;
				}
				
				if (!stricmp(CShaderParser::token, "portal"))
				{
					shad->flags |= SF_PORTAL;
					shad->sortvalue = SORT_PORTAL;
				}
				else if (!stricmp(CShaderParser::token, "sky")) shad->sortvalue = SORT_SKY;
				else if (!stricmp(CShaderParser::token, "opaque")) shad->sortvalue = SORT_OPAQUE;
				else if (!stricmp(CShaderParser::token, "banner")) shad->sortvalue = SORT_BANNER;
				else if (!stricmp(CShaderParser::token, "underwater")) shad->sortvalue = SORT_UNDERWATER;
				else if (!stricmp(CShaderParser::token, "additive")) shad->sortvalue = SORT_ADDITIVE;
				else if (!stricmp(CShaderParser::token, "nearest")) shad->sortvalue = SORT_NEAREST;
				else // integer value
				{
					shad->sortvalue = atoi(CShaderParser::token);
					if (shad->sortvalue == 0)
					{
						// FIXME: does atoi return 0 when no isdigit?
						shad->sortvalue = SORT_OPAQUE;
						// unknown parameter !!
						GENERAL::dprintf("^5WARNING: invalid parameter \"%s\" in line %d", CShaderParser::token, CShaderParser::scriptline);
					}
				}
			}
			else if (!stricmp(CShaderParser::token, "deformvertexes"))
			{
				if (!CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
					continue;
				}
				
				if (!stricmp(CShaderParser::token, "wave"))
				{
					// Search for a free indice
					int idx;
					for (idx = 0; idx < MAX_DEFORMS; ++idx)
						if (shad->deformVertexes[idx].type == DEFORMVERTEXES_NONE) break;
					if (idx >= MAX_DEFORMS)
					{
						GENERAL::dprintf("Unable to allocate more than %d deformvertexes", MAX_DEFORMS);
						continue;
					}

					shad->deformVertexes[idx].type = DEFORMVERTEXES_WAVE;
					++shad->num_deforms;
					if (!CShaderParser::GetToken(false))
					{
						GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
						continue;
					}
					shad->deformVertexes[idx].params[1] = (float) atof(CShaderParser::token);
					shad->deformVertexes[idx].params[0] = ParseFunc();
					for (int i = 2; i < 6; ++i)
					{
						if (!CShaderParser::GetToken(false))
						{
							GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
							break;
						}
						shad->deformVertexes[idx].params[i] = (float) atof(CShaderParser::token);
					}
				}
				else if (!stricmp(CShaderParser::token, "normal"))
				{
					// Search for a free indice
					int idx;
					for (idx = 0; idx < MAX_DEFORMS; ++idx)
						if (shad->deformVertexes[idx].type == DEFORMVERTEXES_NONE) break;
					if (idx >= MAX_DEFORMS)
					{
						GENERAL::dprintf("Unable to allocate more than %d deformvertexes", MAX_DEFORMS);
						continue;
					}

					shad->deformVertexes[idx].type = DEFORMVERTEXES_NORMAL;
					++shad->num_deforms;

					// Reading amplitude and frequency
					for (int i = 0; i < 2; ++i)
					{
						if (!CShaderParser::GetToken(false))
						{
							GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
							break;
						}
						shad->deformVertexes[idx].params[i] = (float) atof(CShaderParser::token);
					}				
				}
				else if (!stricmp(CShaderParser::token, "bulge"))
				{
					// Search for a free indice
					int idx;
					for (idx = 0; idx < MAX_DEFORMS; ++idx)
						if (shad->deformVertexes[idx].type == DEFORMVERTEXES_NONE) break;
					if (idx >= MAX_DEFORMS)
					{
						GENERAL::dprintf("Unable to allocate more than %d deformvertexes", MAX_DEFORMS);
						continue;
					}

					shad->deformVertexes[idx].type = DEFORMVERTEXES_BULGE;
					++shad->num_deforms;
					for (int i = 0; i < 3; ++i)
					{
						if (!CShaderParser::GetToken(false)) break;
						shad->deformVertexes[idx].params[i] = (float) atof(CShaderParser::token);
					}
				}
				else if (!stricmp(CShaderParser::token, "move"))
				{
					// Search for a free indice
					int idx, i;
					for (idx = 0; idx < MAX_DEFORMS; ++idx)
						if (shad->deformVertexes[idx].type == DEFORMVERTEXES_NONE) break;
					if (idx >= MAX_DEFORMS)
					{
						GENERAL::dprintf("Unable to allocate more than %d deformvertexes", MAX_DEFORMS);
						continue;
					}

					// FIXME: This was based on .shader content, but it doesn't respect Q3A Shader Manual !!
					shad->deformVertexes[idx].type = DEFORMVERTEXES_MOVE;
					++shad->num_deforms;
					for (i = 1; i < 4; ++i)
					{
						if (!CShaderParser::GetToken(false)) continue;
						shad->deformVertexes[idx].params[i] = (float) atof(CShaderParser::token);
					}
					shad->deformVertexes[idx].params[0] = ParseFunc();
					for (i = 4; i < 8; ++i)
					{
						if (!CShaderParser::GetToken(false)) continue;
						shad->deformVertexes[idx].params[i] = (float) atof(CShaderParser::token);
					}
				}
				else if (!stricmp(CShaderParser::token, "autosprite"))
				{
					// Search for a free indice
					int idx;
					for (idx = 0; idx < MAX_DEFORMS; ++idx)
						if (shad->deformVertexes[idx].type == DEFORMVERTEXES_NONE) break;
					if (idx >= MAX_DEFORMS)
					{
						GENERAL::dprintf("Unable to allocate more than %d deformvertexes", MAX_DEFORMS);
						continue;
					}

					shad->deformVertexes[idx].type = DEFORMVERTEXES_AUTOSPRITE;
					++shad->num_deforms;
				}
				else if (!stricmp(CShaderParser::token, "autosprite2"))
				{
					// Search for a free indice
					int idx;
					for (idx = 0; idx < MAX_DEFORMS; ++idx)
						if (shad->deformVertexes[idx].type == DEFORMVERTEXES_NONE) break;
					if (idx >= MAX_DEFORMS)
					{
						GENERAL::dprintf("Unable to allocate more than %d deformvertexes", MAX_DEFORMS);
						continue;
					}

					shad->deformVertexes[idx].type = DEFORMVERTEXES_AUTOSPRITE2;
					++shad->num_deforms;
				}
				else
				{
					// unknown parameter !!
					GENERAL::dprintf("^5WARNING: unknown parameter \"%s\" in line %d", CShaderParser::token, CShaderParser::scriptline);
				}
			}
			else if (!stricmp(CShaderParser::token, "texture"))
			{
				long texId = 0;
				long texx  = 0;
				long texy  = 0;


				if (!CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
					continue;
				}
				else
				{
					char *eptr = 0;
					texId = strtol(CShaderParser::token, &eptr, 0);
				}

				if (!CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
					continue;
				}
				else
				{
					char *eptr = 0;
					texx = strtol(CShaderParser::token, &eptr, 0);
				}

				if (!CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
					continue;
				}
				else
				{
					char *eptr = 0;
					texy = strtol(CShaderParser::token, &eptr, 0);
				}

				if (!CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
					continue;
				}
				else
				{
					//GENERAL::dprintf(CShaderParser::token);
					shad->textures[texId].tex=shad->GetResourceManager()->GetPtr(sptr<CRenderAble>(),CShaderParser::token,RESTYP_TEXTURE);
					shad->textures[texId].scale[0]=texx;
					shad->textures[texId].scale[1]=texy;

				}
			}
			else if (!stricmp(CShaderParser::token, "element"))
			{
				long texId;
				float coord[4];
				CResShader::TElementInfo t;

				if (!CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
					continue;
				}
				else
				{
					char *eptr = 0;
					texId = strtol(CShaderParser::token, &eptr, 0);
				};

				for (int i=0; i<4; i++)
				{
					if (!CShaderParser::GetToken(false))
					{
						GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
						continue;
					}
					else
					{
						char *eptr = 0;
						coord[i] = (float) strtod(CShaderParser::token, &eptr);
					};
				}

				t.tex = shad->textures[texId].tex;

				t.texCoords[2].x=(coord[0]+coord[2]-0.0f)/shad->textures[texId].scale[0];
				t.texCoords[2].y=(coord[1]+0.0f)/shad->textures[texId].scale[1];

				t.texCoords[3].x=(coord[0]+coord[2]-0.0f)/shad->textures[texId].scale[0];
				t.texCoords[3].y=(coord[1]+coord[3]-0.0f)/shad->textures[texId].scale[1];

				t.texCoords[0].x=(coord[0]+0.0f)/shad->textures[texId].scale[0];
				t.texCoords[0].y=(coord[1]+coord[3]-0.0f)/shad->textures[texId].scale[1];

				t.texCoords[1].x=(coord[0]+0.0f)/shad->textures[texId].scale[0];
				t.texCoords[1].y=(coord[1]+0.0f)/shad->textures[texId].scale[1];

				shad->elements.push_back(t);
			}
			else if (!stricmp(CShaderParser::token, "surfaceparm"))
			{
				if (!CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
					continue;
				}
				if (!stricmp(CShaderParser::token, "alphashadow"))
				{
					shad->surface_flags |= SURF_ALPHASHADOW;
				}
				else if (!stricmp(CShaderParser::token, "areaportal"))
				{
					shad->content_flags |= CONTENTS_AREAPORTAL;
				}
				else if (!stricmp(CShaderParser::token, "clusterportal"))
				{
					shad->content_flags |= CONTENTS_CLUSTERPORTAL;
				}
				else if (!stricmp(CShaderParser::token, "donotenter"))
				{
					shad->content_flags |= CONTENTS_DONOTENTER;
				}
				else if (!stricmp(CShaderParser::token, "flesh"))
				{
					shad->surface_flags |= SURF_FLESH;
				}
				else if (!stricmp(CShaderParser::token, "fog"))
				{
					shad->content_flags |= CONTENTS_FOG;
				}
				else if (!stricmp(CShaderParser::token, "lava"))
				{
					shad->content_flags |= CONTENTS_LAVA;
				}
				else if (!stricmp(CShaderParser::token, "metalsteps"))
				{
					shad->surface_flags |= SURF_METALSTEPS;
				}
				else if (!stricmp(CShaderParser::token, "nodamage"))
				{
					shad->surface_flags |= SURF_NODAMAGE;
				}
				else if (!stricmp(CShaderParser::token, "nodlight"))
				{
					shad->surface_flags |= SURF_NODLIGHT;
				}
				else if (!stricmp(CShaderParser::token, "nodraw"))
				{
					shad->surface_flags |= SURF_NODRAW;
				}
				else if (!stricmp(CShaderParser::token, "nodrop"))
				{
					shad->content_flags |= CONTENTS_NODROP;
				}
				else if (!stricmp(CShaderParser::token, "noimpact"))
				{
					shad->surface_flags |= SURF_NOIMPACT;
				}
				else if (!stricmp(CShaderParser::token, "nomarks"))
				{
					shad->surface_flags |= SURF_NOMARKS;
				}
				else if (!stricmp(CShaderParser::token, "nolightmap"))
				{
					shad->surface_flags |= SURF_NOLIGHTMAP;
				}
				else if (!stricmp(CShaderParser::token, "nosteps"))
				{
					shad->surface_flags |= SURF_NOSTEPS;
				}
				else if (!stricmp(CShaderParser::token, "nonsolid"))
				{
					shad->surface_flags |= SURF_NONSOLID;
				}
				else if (!stricmp(CShaderParser::token, "origin"))
				{
					shad->content_flags |= CONTENTS_ORIGIN;
				}
				else if (!stricmp(CShaderParser::token, "playerclip"))
				{
					shad->content_flags |= CONTENTS_PLAYERCLIP;
				}
				else if (!stricmp(CShaderParser::token, "slick"))
				{
					shad->surface_flags |= SURF_SLICK;
				}
				else if (!stricmp(CShaderParser::token, "slime"))
				{
					shad->content_flags |= CONTENTS_SLIME;
				}
				else if (!stricmp(CShaderParser::token, "structural"))
				{
					shad->content_flags |= CONTENTS_STRUCTURAL;
				}
				else if (!stricmp(CShaderParser::token, "trans"))
				{
					shad->content_flags |= CONTENTS_TRANSLUCENT;
				}
				else if (!stricmp(CShaderParser::token, "water"))
				{
					shad->content_flags |= CONTENTS_WATER;
				}
				else if (!stricmp(CShaderParser::token, "nomipmaps"))
				{
				//	GENERAL::dprintf("^5WARNING: unexpected \"nomipmaps\" parameter found in line %d", CShaderParser::scriptline);
					shad->flags |= LF_NOMIPMAP;
				}
				else if (!stricmp(CShaderParser::token, "sky"))
				{
				//	GENERAL::dprintf("^5WARNING: unexpected \"sky\" parameter found in line %d", CShaderParser::scriptline);
					shad->surface_flags |= SURF_SKY;
				}
				else
				{
					GENERAL::dprintf("^5WARNING: unknown parameter \"%s\" in line %d", CShaderParser::token, CShaderParser::scriptline);
				}
			}
			else if (!stricmp(CShaderParser::token, "fogparms"))
			{	// syntax: fogparms ( <red> <green> <blue> ) <distance to Opaque>
				shad->content_flags |= CONTENTS_FOG;

				if (!CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
					continue;
				}

				if (strcmp(CShaderParser::token, "("))
				{
					GENERAL::dprintf("^5WARNING: expected \"(\" character in line %d", CShaderParser::scriptline);
					continue;
				}

				for (int i = 0; i < 3; ++i)
				{
					if (!CShaderParser::GetToken(false)) break;
					shad->fog_params.params[i] = (float) atof(CShaderParser::token);
				}

				shad->fog_params.params[3] = 1.f;

				if (!CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
					continue;
				}

				if (strcmp(CShaderParser::token, ")"))
				{			
					GENERAL::dprintf("^5WARNING: ) expected in line %d", CShaderParser::scriptline);
					continue;
				}

				if (!CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
					continue;
				}

				// This parameter represents 1/density
				shad->fog_params.params[4] = (float) atof(CShaderParser::token);

				// Create the new fog layer
				shad->fog_params.layer = shad->CreateLayer();
				shad->fog_params.layer->blendsrc = CShaderParser::IntForConst("GL_SRC_ALPHA");
				shad->fog_params.layer->blenddst = CShaderParser::IntForConst("GL_ONE_MINUS_SRC_ALPHA");
				shad->fog_params.layer->depthFunc = CShaderParser::IntForConst("GL_LEQUAL");
				shad->fog_params.layer->num_maps = 1;
//FOG				shad->fog_params.layer->map[0] = &gRender->fogtexture;			
				shad->fog_params.layer->alphafunc = CShaderParser::IntForConst("GL_GREATER");
				shad->fog_params.layer->alphafuncref = 0.05f;
				shad->fog_params.layer->rgbGen = RGBGEN_FOG;
				shad->fog_params.layer->tcGen = TCGEN_FOG;
	//			shad->fog_params.layer->depthWrite = false;
				shad->sortvalue = SORT_FOG;
//				shad->num_layers++;
			}
			else if (!stricmp(CShaderParser::token, "skyparms"))
			{
				bool hasFarBox = false;

				// farbox parameter
				if (!CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: farbox parameter expected in line %d", CShaderParser::scriptline);
					continue;
				}
				if (strcmp(CShaderParser::token, "-"))
				{
					hasFarBox = true;

					// a farbox is present - we need to load the 6 env textures
					for (int i = 0; i < 6; ++i)
					{
						shad->sky_params.layer[shad->sky_params.num_layers][i] = shad->CreateLayer();
						shad->sky_params.layer[shad->sky_params.num_layers][i]->blendsrc = CShaderParser::IntForConst("GL_ONE");
						shad->sky_params.layer[shad->sky_params.num_layers][i]->blenddst = CShaderParser::IntForConst("GL_ZERO");
						shad->sky_params.layer[shad->sky_params.num_layers][i]->depthWrite = false;
						shad->sky_params.layer[shad->sky_params.num_layers][i]->alphafunc = CShaderParser::IntForConst("GL_GEQUAL");
						shad->sky_params.layer[shad->sky_params.num_layers][i]->alphafuncref = 0.5f;
//						shad->num_layers++;
					}

					char texname[64] = { '\0' };
/* original			sprintf(texname, "%s_rt", CShaderParser::token); shad->sky_params.layer[shad->sky_params.num_layers][0]->map[0] = textures->AddTexinfo(texname, LF_CLAMP);
					sprintf(texname, "%s_bk", CShaderParser::token); shad->sky_params.layer[shad->sky_params.num_layers][1]->map[0] = textures->AddTexinfo(texname, LF_CLAMP);
					sprintf(texname, "%s_up", CShaderParser::token); shad->sky_params.layer[shad->sky_params.num_layers][2]->map[0] = textures->AddTexinfo(texname, LF_CLAMP);
					sprintf(texname, "%s_lf", CShaderParser::token); shad->sky_params.layer[shad->sky_params.num_layers][3]->map[0] = textures->AddTexinfo(texname, LF_CLAMP);
					sprintf(texname, "%s_ft", CShaderParser::token); shad->sky_params.layer[shad->sky_params.num_layers][4]->map[0] = textures->AddTexinfo(texname, LF_CLAMP);
					sprintf(texname, "%s_dn", CShaderParser::token); shad->sky_params.layer[shad->sky_params.num_layers][5]->map[0] = textures->AddTexinfo(texname, LF_CLAMP);*/

					sprintf(texname, "%s_rt", CShaderParser::token); shad->sky_params.layer[shad->sky_params.num_layers][0]->map[0] = shad->GetResourceManager()->GetPtr(sptr<CRenderAble>(),texname,RESTYP_TEXTURE); shad->sky_params.layer[shad->sky_params.num_layers][0]->map[0]->SetWrapMode(GL_CLAMP);
					sprintf(texname, "%s_bk", CShaderParser::token); shad->sky_params.layer[shad->sky_params.num_layers][1]->map[0] = shad->GetResourceManager()->GetPtr(sptr<CRenderAble>(),texname,RESTYP_TEXTURE); shad->sky_params.layer[shad->sky_params.num_layers][1]->map[0]->SetWrapMode(GL_CLAMP);
					sprintf(texname, "%s_up", CShaderParser::token); shad->sky_params.layer[shad->sky_params.num_layers][2]->map[0] = shad->GetResourceManager()->GetPtr(sptr<CRenderAble>(),texname,RESTYP_TEXTURE); shad->sky_params.layer[shad->sky_params.num_layers][2]->map[0]->SetWrapMode(GL_CLAMP);
					sprintf(texname, "%s_lf", CShaderParser::token); shad->sky_params.layer[shad->sky_params.num_layers][3]->map[0] = shad->GetResourceManager()->GetPtr(sptr<CRenderAble>(),texname,RESTYP_TEXTURE); shad->sky_params.layer[shad->sky_params.num_layers][3]->map[0]->SetWrapMode(GL_CLAMP);
					sprintf(texname, "%s_ft", CShaderParser::token); shad->sky_params.layer[shad->sky_params.num_layers][4]->map[0] = shad->GetResourceManager()->GetPtr(sptr<CRenderAble>(),texname,RESTYP_TEXTURE); shad->sky_params.layer[shad->sky_params.num_layers][4]->map[0]->SetWrapMode(GL_CLAMP);
					sprintf(texname, "%s_dn", CShaderParser::token); shad->sky_params.layer[shad->sky_params.num_layers][5]->map[0] = shad->GetResourceManager()->GetPtr(sptr<CRenderAble>(),texname,RESTYP_TEXTURE); shad->sky_params.layer[shad->sky_params.num_layers][5]->map[0]->SetWrapMode(GL_CLAMP);


					++shad->sky_params.num_layers;
				}

				// cloudheight parameter
				if (!CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: cloudheight parameter expected in line %d", CShaderParser::scriptline);
					continue;
				}

				if (strcmp(CShaderParser::token, "-"))
					shad->sky_params.cloudheight = (int) atoi(CShaderParser::token);
				else
					shad->sky_params.cloudheight = 128;	// default value

				// nearbox parameter
				if (!CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: nearbox parameter expected in line %d", CShaderParser::scriptline);
					continue;
				}
		
				if (strcmp(CShaderParser::token, "-"))
				{
					// a nearbox is present - we need to load the 6 env textures
					for (int i = 0; i < 6; ++i)
					{
						shad->sky_params.layer[shad->sky_params.num_layers][i] = shad->CreateLayer();
						shad->sky_params.layer[shad->sky_params.num_layers][i]->blendsrc = CShaderParser::IntForConst("GL_ONE");
						if (hasFarBox)
							shad->sky_params.layer[shad->sky_params.num_layers][i]->blenddst = CShaderParser::IntForConst("GL_ONE");
						else	// no blending if no farbox
							shad->sky_params.layer[shad->sky_params.num_layers][i]->blenddst = CShaderParser::IntForConst("GL_ZERO");
						shad->sky_params.layer[shad->sky_params.num_layers][i]->depthWrite = false;
						shad->sky_params.layer[shad->sky_params.num_layers][i]->alphafunc = CShaderParser::IntForConst("GL_GEQUAL");
						shad->sky_params.layer[shad->sky_params.num_layers][i]->alphafuncref = 0.5f;
//						shad->num_layers++;
					}

					char texname[64] = { '\0' };
/* original			sprintf(texname, "%s_rt", CShaderParser::token); shad->sky_params.layer[shad->sky_params.num_layers][0]->map[0] = textures->AddTexinfo(texname, LF_CLAMP);
					sprintf(texname, "%s_bk", CShaderParser::token); shad->sky_params.layer[shad->sky_params.num_layers][1]->map[0] = textures->AddTexinfo(texname, LF_CLAMP);
					sprintf(texname, "%s_up", CShaderParser::token); shad->sky_params.layer[shad->sky_params.num_layers][2]->map[0] = textures->AddTexinfo(texname, LF_CLAMP);
					sprintf(texname, "%s_lf", CShaderParser::token); shad->sky_params.layer[shad->sky_params.num_layers][3]->map[0] = textures->AddTexinfo(texname, LF_CLAMP);
					sprintf(texname, "%s_ft", CShaderParser::token); shad->sky_params.layer[shad->sky_params.num_layers][4]->map[0] = textures->AddTexinfo(texname, LF_CLAMP);
					sprintf(texname, "%s_dn", CShaderParser::token); shad->sky_params.layer[shad->sky_params.num_layers][5]->map[0] = textures->AddTexinfo(texname, LF_CLAMP);*/

					sprintf(texname, "%s_rt", CShaderParser::token); shad->sky_params.layer[shad->sky_params.num_layers][0]->map[0] = shad->GetResourceManager()->GetPtr(sptr<CRenderAble>(),texname,RESTYP_TEXTURE); shad->sky_params.layer[shad->sky_params.num_layers][0]->map[0]->SetWrapMode(GL_CLAMP);
					sprintf(texname, "%s_bk", CShaderParser::token); shad->sky_params.layer[shad->sky_params.num_layers][1]->map[0] = shad->GetResourceManager()->GetPtr(sptr<CRenderAble>(),texname,RESTYP_TEXTURE); shad->sky_params.layer[shad->sky_params.num_layers][1]->map[0]->SetWrapMode(GL_CLAMP);
					sprintf(texname, "%s_up", CShaderParser::token); shad->sky_params.layer[shad->sky_params.num_layers][2]->map[0] = shad->GetResourceManager()->GetPtr(sptr<CRenderAble>(),texname,RESTYP_TEXTURE); shad->sky_params.layer[shad->sky_params.num_layers][2]->map[0]->SetWrapMode(GL_CLAMP);
					sprintf(texname, "%s_lf", CShaderParser::token); shad->sky_params.layer[shad->sky_params.num_layers][3]->map[0] = shad->GetResourceManager()->GetPtr(sptr<CRenderAble>(),texname,RESTYP_TEXTURE); shad->sky_params.layer[shad->sky_params.num_layers][3]->map[0]->SetWrapMode(GL_CLAMP);
					sprintf(texname, "%s_ft", CShaderParser::token); shad->sky_params.layer[shad->sky_params.num_layers][4]->map[0] = shad->GetResourceManager()->GetPtr(sptr<CRenderAble>(),texname,RESTYP_TEXTURE); shad->sky_params.layer[shad->sky_params.num_layers][4]->map[0]->SetWrapMode(GL_CLAMP);
					sprintf(texname, "%s_dn", CShaderParser::token); shad->sky_params.layer[shad->sky_params.num_layers][5]->map[0] = shad->GetResourceManager()->GetPtr(sptr<CRenderAble>(),texname,RESTYP_TEXTURE); shad->sky_params.layer[shad->sky_params.num_layers][5]->map[0]->SetWrapMode(GL_CLAMP);
					++shad->sky_params.num_layers;
				}
			}

			// skip non-used commands
			else if (!strnicmp(CShaderParser::token, "qer_", 4))
			{
				// ignore
				while (CShaderParser::GetToken(false)){};
				continue;
			}
			else if (!strnicmp(CShaderParser::token, "q3map_", 6))
			{
				// ignore
				while (CShaderParser::GetToken(false)){};
				continue;
			}
			else if (!stricmp(CShaderParser::token, "light"))
			{
				// ignore
				while (CShaderParser::GetToken(false)){};
				continue;
			}
			else if (!stricmp(CShaderParser::token, "tessSize"))
			{
				// ignore
				while (CShaderParser::GetToken(false)){};
				continue;
			}
			
			// unknown command
			else
			{
				GENERAL::dprintf("^5WARNING: unknown command \"%s\" in line %d", CShaderParser::token, CShaderParser::scriptline);
			}

			// unknown keyword skip line
			while (CShaderParser::GetToken(false))	// while not EOL of EOF
			{
				GENERAL::dprintf("^5WARNING: skipping \"%s\" in line %d", CShaderParser::token, CShaderParser::scriptline);
			}
		}

		return true;
	}

	void CShaderFile::ParseLayer(CResShader *shad, CResShader::Layer *layer)
	{
		if (shad->flags & SF_NOMIPMAP) layer->flags |= LF_NOMIPMAP;
		if (shad->flags & SF_NOPICMIP) layer->flags |= LF_NOPICMIP;

		while (CShaderParser::GetToken(true))
		{
			if (!strcmp(CShaderParser::token, "{"))
			{
				// display error, token unexpected!
				break;
			}
			
			else if (!strcmp(CShaderParser::token, "}")) return;	// ends layer!
			
			else if (!stricmp(CShaderParser::token, "map"))
			{
				if (!CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
					continue;
				}

				if (!stricmp(CShaderParser::token, "$lightmap"))
				{
					shad->flags |= SF_HASLIGHTMAP;
					layer->flags = LF_LIGHTMAP;
					layer->tcGen = TCGEN_LIGHTMAP;
//SPTR					layer->map[0] = NULL;
					layer->rgbGen = RGBGEN_IDENTITY;
				}
				else if (!stricmp(CShaderParser::token, "$element"))
				{
					layer->flags = LF_ELEMENT;
					layer->tcGen = TCGEN_ELEMENT;
//SPTR					layer->map[0] = NULL;
					layer->rgbGen = RGBGEN_IDENTITY;
				}
				else
				{
// original			layer->map[0] = textures->AddTexinfo(CShaderParser::token, layer->flags);
					layer->map[0] = shad->GetResourceManager()->GetPtr(sptr<CRenderAble>(),CShaderParser::token,RESTYP_TEXTURE);
					if (layer->flags & LF_CLAMP)
						 layer->map[0]->SetWrapMode(GL_CLAMP);
				}
				continue;
			}
			
			else if (!stricmp(CShaderParser::token, "clampmap"))
			{
				if (!CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
					continue;
				}
				layer->flags |= LF_CLAMP;
// original		layer->map[0] = textures->AddTexinfo(CShaderParser::token, layer->flags);
				layer->map[0] = shad->GetResourceManager()->GetPtr(sptr<CRenderAble>(),CShaderParser::token,RESTYP_TEXTURE);
				layer->map[0]->SetWrapMode(GL_CLAMP);
				continue;
			}
			
			else if (!stricmp(CShaderParser::token, "animmap"))
			{
				int num_maps = 0;
				if (!CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
					continue;
				}
				layer->anim_speed = (float) atof(CShaderParser::token);
				
				while (CShaderParser::GetToken(false))
				{
// original			layer->map[num_maps] = textures->AddTexinfo(CShaderParser::token, layer->flags);
					layer->map[num_maps] = shad->GetResourceManager()->GetPtr(sptr<CRenderAble>(),CShaderParser::token,RESTYP_TEXTURE);
					if (layer->flags & LF_CLAMP)
						layer->map[num_maps]->SetWrapMode(GL_CLAMP);
					++num_maps;
				}
				layer->num_maps = num_maps;
				continue;
			}
			else if (!stricmp(CShaderParser::token, "blendfunc"))
			{
				if (!CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
					continue;
				}

				if (!stricmp(CShaderParser::token, "add"))
				{
					layer->blendsrc = layer->blenddst = CShaderParser::IntForConst("GL_ONE");
					layer->depthWrite = false;
				}
				else if (!stricmp(CShaderParser::token, "filter"))
				{
					layer->blendsrc = CShaderParser::IntForConst("GL_ZERO");
					layer->blenddst = CShaderParser::IntForConst("GL_SRC_COLOR");
					layer->depthWrite = false;
				}
				else if (!stricmp(CShaderParser::token, "blend"))
				{
					layer->blendsrc = CShaderParser::IntForConst("GL_SRC_ALPHA");
					layer->blenddst = CShaderParser::IntForConst("GL_ONE_MINUS_SRC_ALPHA");
					layer->depthWrite = false;
				}
				else
				{
					// TODO: analize better the blendfunc!!
					layer->blendsrc = CShaderParser::IntForConst(CShaderParser::token);
					if (!CShaderParser::GetToken(false))
					{
						GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
						continue;
					}	
					layer->blenddst = CShaderParser::IntForConst(CShaderParser::token);
					layer->depthWrite = (layer->blendsrc == 1 && layer->blenddst == 0 && shad->num_layers == 0);
				}

				if (shad->num_layers == 0 && shad->sortvalue == SORT_OPAQUE/*default*/)
				{
					// FIXME: thats a hack to work around texture sorting problems, id docs say
					//		  that any shader with blendfunc has SORT_ADDITIVE by default.
					if (layer->blendsrc == 1 && layer->blenddst == 1 && shad->num_layers == 0) shad->sortvalue = SORT_ADDITIVE;
					else shad->sortvalue = SORT_SEETHROUGH;
				}
				//if (layer->blendsrc == CShaderParser::IntForConst("GL_SRC_ALPHA") && layer->blenddst == CShaderParser::IntForConst("GL_ONE_MINUS_SRC_ALPHA")) layer->rgbGen = RGBGEN_IDENTITY_LIGHTING;
				//else if (layer->blendsrc == CShaderParser::IntForConst("GL_ONE") && layer->blenddst == CShaderParser::IntForConst("GL_ONE")) layer->rgbGen = RGBGEN_IDENTITY_LIGHTING;
				if (layer->blendsrc == CShaderParser::IntForConst("GL_ZERO") && layer->blenddst == CShaderParser::IntForConst("GL_SRC_COLOR")) layer->rgbGen = RGBGEN_IDENTITY;
				else if (layer->blendsrc == CShaderParser::IntForConst("GL_DST_COLOR") && layer->blenddst == CShaderParser::IntForConst("GL_ZERO")) layer->rgbGen = RGBGEN_IDENTITY;
			}
			
			else if (!stricmp(CShaderParser::token, "depthfunc"))
			{
				if (!CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
					continue;
				}
				if (!stricmp(CShaderParser::token, "lequal")) layer->depthFunc = CShaderParser::IntForConst("GL_LEQUAL");
				else if (!stricmp(CShaderParser::token, "equal")) layer->depthFunc = CShaderParser::IntForConst("GL_EQUAL");
				else if (!stricmp(CShaderParser::token, "always")) layer->depthFunc = CShaderParser::IntForConst("GL_ALWAYS");
				else
				{
					// warning!!
					GENERAL::dprintf("^5WARNING: unknown parameter \"%s\" in line %d", CShaderParser::token, CShaderParser::scriptline);
					continue;
				}
			}
			
			else if (!stricmp(CShaderParser::token, "depthwrite"))
			{
				layer->depthWrite = true;
			}
			
			else if (!stricmp(CShaderParser::token, "detail"))
			{
				layer->flags |= LF_DETAIL;
			}
			
			else if (!stricmp(CShaderParser::token, "alphafunc"))
			{
				if (!CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
					continue;
				}
				if (!stricmp(CShaderParser::token, "gt0"))
				{
					layer->alphafunc = CShaderParser::IntForConst("GL_GREATER");
					layer->alphafuncref = 0.0f;
					//layer->depthWrite = false;
				}
				else if (!stricmp(CShaderParser::token, "lt128"))
				{
					layer->alphafunc = CShaderParser::IntForConst("GL_LESS");
					layer->alphafuncref = 0.5f;
					//layer->depthWrite = false;
				}
				else if (!stricmp(CShaderParser::token, "ge128"))
				{
					layer->alphafunc = CShaderParser::IntForConst("GL_GEQUAL");
					layer->alphafuncref = 0.5f;
					//layer->depthWrite = false;
				}
				else
				{
					GENERAL::dprintf("^5WARNING: unknown parameter \"%s\" in line %d", CShaderParser::token, CShaderParser::scriptline);
					continue;
				}
			}

			else if (!stricmp(CShaderParser::token, "rgbgen"))
			{
				if (!CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
					continue;
				}
				if (!stricmp(CShaderParser::token, "vertex"))
				{
					shad->flags |= SF_HASCOLORS;
					layer->rgbGen = RGBGEN_VERTEX;
				}
				else if (!stricmp(CShaderParser::token, "lightingdiffuse"))
				{
					shad->flags |= SF_HASCOLORS;
					layer->rgbGen = RGBGEN_LIGHTING_DIFFUSE;
				}
				else if (!stricmp(CShaderParser::token, "oneminusvertex"))
				{
					shad->flags |= SF_HASCOLORS;
					layer->rgbGen = RGBGEN_ONE_MINUS_VERTEX;
				}
				else if (!stricmp(CShaderParser::token, "identity"))
				{
					layer->rgbGen = RGBGEN_IDENTITY;
				}
				else if (!stricmp(CShaderParser::token, "identitylighting"))
				{
					layer->rgbGen = RGBGEN_IDENTITY_LIGHTING;
				}
				else if (!stricmp(CShaderParser::token, "wave"))
				{
					layer->rgbGen = RGBGEN_WAVE;
					layer->rgbGenParams[0] = ParseFunc();
					for (int i = 1; i < 5; ++i)
					{
						if (!CShaderParser::GetToken(false))
						{
							GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
							break;
						}
						layer->rgbGenParams[i] = (float) atof(CShaderParser::token);
					}
				}

				while (CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: skipping \"%s\" in line %d", CShaderParser::token, CShaderParser::scriptline);
				}
				continue;
			}
			else if (!stricmp(CShaderParser::token, "alphagen"))
			{
				// TODO: add distance atenuation !!
				if (!CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
					continue;
				}
				if (!stricmp(CShaderParser::token, "portal"))
				{
					layer->alphaGen = ALPHAGEN_PORTAL;

					if (!CShaderParser::GetToken(false))
					{
						GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
						continue;
					}
					layer->alphaGenParams[0] = (float) atof(CShaderParser::token);
				}
				else if (!stricmp(CShaderParser::token, "vertex"))
				{
					//shad->flags |= SF_HASCOLORS;
					layer->alphaGen = ALPHAGEN_VERTEX;
				}
				else if (!stricmp(CShaderParser::token, "identity"))
				{
					layer->alphaGen = ALPHAGEN_IDENTITY;
				}
				else if (!stricmp(CShaderParser::token, "entity"))
				{
					layer->alphaGen = ALPHAGEN_ENTITY;
				}
				else if (!stricmp(CShaderParser::token, "lightingspecular"))
				{
					layer->alphaGen = ALPHAGEN_SPECULAR;
				}
				else if (!stricmp(CShaderParser::token, "wave"))
				{
					layer->alphaGen = ALPHAGEN_WAVE;
					layer->alphaGenParams[0] = ParseFunc();
					for (int i = 1; i < 5; ++i)
					{
						if (!CShaderParser::GetToken(false))
						{
							GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
							break;
						}
						layer->alphaGenParams[i] = (float) atof(CShaderParser::token);
					}
				}
				else if (!stricmp(CShaderParser::token, "const") || !stricmp(CShaderParser::token, "constant"))
				{
					GENERAL::dprintf("^5WARNING: Constant alphagen not supported");
					layer->alphaGen = ALPHAGEN_CONST;
					layer->aGen[0].func = FUNC_CONST;
					layer->aGen[0].p[0] = 0;
				}

				while (CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: skipping \"%s\" in line %d", CShaderParser::token, CShaderParser::scriptline);
				}
				continue;
			}

			else if (!stricmp(CShaderParser::token, "tcgen"))
			{
				if (!CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
					continue;
				}
				if (!stricmp(CShaderParser::token, "environment"))
				{
					layer->tcGen = TCGEN_ENVIRONMENT;
					layer->flags |= LF_USENORMALS;
					shad->flags |= SF_USENORMALS;
				}

				while (CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: skipping \"%s\" in line %d", CShaderParser::token, CShaderParser::scriptline);
				}
				continue;
			}
			
			else if (!stricmp(CShaderParser::token, "tcmod"))
			{
				if (!CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
					continue;
				}
				if (!stricmp(CShaderParser::token, "rotate"))
				{
					CResShader::funclist *f;
					layer->flags |= LF_HASTCMOD;
					shad->flags |= SF_HASTCMOD;
					if (!layer->tcMod) f = layer->tcMod = new CResShader::funclist;
					else
					{
						for (f = layer->tcMod; f->next; f = f->next){}
						f = f->next = new CResShader::funclist;
					}
					f->next = NULL;
					f->func = TCMOD_ROTATE;
					f->parameters = 1;
					for (int i = 0; i < MAX_PARAMS; ++i)
					{
						if (!CShaderParser::GetToken(false)) break;
						f->p[i] = (float) atof(CShaderParser::token);
					}
				}
				else if (!stricmp(CShaderParser::token, "scale"))
				{
					CResShader::funclist *f;
					layer->flags |= LF_HASTCMOD;
					shad->flags |= SF_HASTCMOD;
					if (!layer->tcMod) f = layer->tcMod = new CResShader::funclist;
					else
					{
						for (f = layer->tcMod; f->next; f = f->next){}
						f = f->next = new CResShader::funclist;
					}
					f->next = NULL;
					f->func = TCMOD_SCALE;
					f->parameters = 2;
					for (int i = 0; i < MAX_PARAMS; ++i)
					{
						if (!CShaderParser::GetToken(false)) break;
						f->p[i] = (float)atof(CShaderParser::token);
					}
				}
				else if (!stricmp(CShaderParser::token, "scroll"))
				{
					CResShader::funclist *f;
					layer->flags |= LF_HASTCMOD;
					shad->flags |= SF_HASTCMOD;
					if (!layer->tcMod) f = layer->tcMod = new CResShader::funclist;
					else
					{
						for (f = layer->tcMod; f->next; f = f->next){}
						f = f->next = new CResShader::funclist;
					}
					f->next = NULL;
					f->func = TCMOD_SCROLL;
					f->parameters = 2;
					for (int i = 0; i < MAX_PARAMS; ++i)
					{
						if (!CShaderParser::GetToken(false)) break;
						f->p[i] = (float)atof(CShaderParser::token);
					}
				}
				else if (!stricmp(CShaderParser::token, "stretch"))
				{
					CResShader::funclist *f;
					layer->flags |= LF_HASTCMOD;
					shad->flags |= SF_HASTCMOD;
					if (!layer->tcMod) f = layer->tcMod = new CResShader::funclist;
					else
					{
						for (f = layer->tcMod; f->next; f = f->next){}
						f=f->next = new CResShader::funclist;
					}
					f->next = NULL;
					f->func = TCMOD_STRETCH;
					f->parameters = 5;
					f->p[0] = ParseFunc();
					for (int i = 1; i < MAX_PARAMS; ++i)
					{
						if (!CShaderParser::GetToken(false)) break;
						f->p[i] = (float)atof(CShaderParser::token);
					}
				}
				else if (!stricmp(CShaderParser::token, "transform"))
				{
					CResShader::funclist *f;
					layer->flags |= LF_HASTCMOD;
					shad->flags |= SF_HASTCMOD;
					if (!layer->tcMod) f = layer->tcMod = new CResShader::funclist;
					else
					{
						for (f = layer->tcMod; f->next; f = f->next){}
						f=f->next = new CResShader::funclist;
					}
					f->next = NULL;
					f->func = TCMOD_TRANSFORM;
					f->parameters = 6;
					for (int i = 0; i < MAX_PARAMS; ++i)
					{
						if (!CShaderParser::GetToken(false)) break;
						f->p[i] = (float)atof(CShaderParser::token);
					}
				}
				else if (!stricmp(CShaderParser::token, "turb"))
				{
					CResShader::funclist *f;
					layer->flags |= LF_HASTCMOD;
					shad->flags |= SF_HASTCMOD;
					if (!layer->tcMod) f = layer->tcMod=new CResShader::funclist;
					else
					{
						for (f = layer->tcMod; f->next; f = f->next){}
						f = f->next = new CResShader::funclist;
					}
					f->next = NULL;
					f->func = TCMOD_TURB;
					f->parameters = 4;
					for (int i = 0; i < MAX_PARAMS; ++i)
					{
						if (!CShaderParser::GetToken(false)) break;
						f->p[i] = (float)atof(CShaderParser::token);
					}
				}

				while (CShaderParser::GetToken(false))
				{
					GENERAL::dprintf("^5WARNING: skipping \"%s\" in line %d", CShaderParser::token, CShaderParser::scriptline);
				}
				continue;
			}

			else
			{
				GENERAL::dprintf("^5WARNING: unknown parameter \"%s\" in line %d", CShaderParser::token, CShaderParser::scriptline);
			}

			// unknown keyword skip line
			while (CShaderParser::GetToken(false))
			{
				GENERAL::dprintf("^5WARNING: skipping \"%s\" in line %d", CShaderParser::token, CShaderParser::scriptline);
			}
		}
	}

	float CShaderFile::ParseFunc(void)
	{
		if (!CShaderParser::GetToken(false))
		{
			GENERAL::dprintf("^5WARNING: parameter expected in line %d", CShaderParser::scriptline);
			return 0;
		}
		
		if (!stricmp(CShaderParser::token, "sin")) return FUNC_SIN;
		else if (!stricmp(CShaderParser::token, "triangle")) return FUNC_TRIANGLE;
		else if (!stricmp(CShaderParser::token, "square")) return FUNC_SQUARE;
		else if (!stricmp(CShaderParser::token, "sawtooth")) return FUNC_SAWTOOTH;
		else if (!stricmp(CShaderParser::token, "inversesawtooth")) return FUNC_INVERSESAWTOOTH;
		else if (!stricmp(CShaderParser::token, "noise")) return FUNC_NOISE;
		else
		{
			GENERAL::dprintf("^5WARNING: unknown function \"%s\" in line %d", CShaderParser::token, CShaderParser::scriptline);
			return 0;
		}
	}

	bool CShaderFile::SkipShader()
	{
			char *tok = CShaderParser::token;

		if (!CShaderParser::GetToken(true))
			return false;

		if (stricmp(CShaderParser::token,"{")!=0)
		{
			return false;
		}
		int level=1;
		while (level>0)
		{
			if (!CShaderParser::GetToken(true))
				return false;
			if (stricmp(CShaderParser::token,"{")==0)
				level++;
			if (stricmp(CShaderParser::token,"}")==0)
				level--;
		}

		return true;
	}

/*******************************************************************************************************************/
/* CShaderParser                                                                                                   */
/*******************************************************************************************************************/
	char			 CShaderParser::token[MAXTOKEN];			// buffer for tokens
	int				 CShaderParser::scriptline;
	CShaderFile		*CShaderParser::file;
	const char		*CShaderParser::buffer;
	const char		*CShaderParser::buf_start;
	const char		*CShaderParser::buf_end;

	void CShaderParser::StartParseBuffer(const char *buff, const int size)
	{
		buf_start = buffer = buff;
		buf_end = buff + size;
		scriptline = 1;
	}

	void CShaderParser::StartParseString(const char *string)
	{
		buf_start = buffer = (char *) string;
		buf_end = (char *) string + strlen(string);
		scriptline = 1;
	}

	void CShaderParser::StopParseFile(void)
	{
		delete file;
	}

	// if crossline is false, GetToken returns false when a line break is found
	// GetToken(true) returns 0 only if end of file
	// GetToken(false) returns 0 if line break or end of file
	bool CShaderParser::GetToken(bool crossline)
	{
		char *token_p;

		if (!buffer) return 0;
		if (buffer == buf_end) return 0;

		*token = 0;	// init to zero

	// skip space
	skipspace:
		while (*buffer <= 32 || *buffer == '\n')
		{
			// no se si \n, \t y \r son menores que 32, pero supongo que s
			if (*buffer == '\n') ++scriptline;
			if (buffer >= buf_end)
			{
				if (!crossline)
				{
					//gConsole->InsertLine("Line %i is incomplete", scriptline);
				}
				return 0;
			}
			
			if (!crossline)
			{
				if (*buffer == '\n')
				{
					--scriptline;
					return 0;
				}
			}
			++buffer;
		}

		if (buffer >= buf_end) return 0;

		// # // comments
		if (*buffer == '#' ||
			(buffer[0] == '/' && buffer[1] == '/'))
		{
			while (*buffer++ != '\n')
			{
				if (buffer >= buf_end) return 0;
			}
			++scriptline;
			if (!crossline) return 0;
			goto skipspace;
		}

		// /* */ comments
		if (buffer[0] == '/' && buffer[1] == '*')
		{
			buffer += 2;
			while (buffer[0] != '*' &&
				buffer[1] != '/')
			{
				++buffer;
				if (buffer >= buf_end)
				{
					//gConsole->InsertLine("*** Error: Premature end of file, '/*' without '*/'");
					return 0;
				}
				if (*buffer == '\n') ++scriptline;
			}
			buffer += 2;
			goto skipspace;
		}

		// copy token
		token_p = token;

		if (*buffer == '"')
		{
			// quoted token
			++buffer;
			while (*buffer != '"')
			{
				*token_p++ = *buffer++;
				if (buffer == buf_end) break;
				if (token_p == &token[MAXTOKEN])
				{
					//gConsole->InsertLine("^1ERROR: Token too large on line %i",scriptline);
				}
			}
			++buffer;
		}
		else
		{
			// regular token
			if (*buffer == '{' ||
				*buffer == '}')
			{
				*token_p++ = *buffer++;
				*token_p = 0;
				return 1;
			}
			while (*buffer > 32 &&
				*buffer != '\n' &&
				*buffer !='{' &&
				*buffer !='}')
			{
				*token_p++ = *buffer++;
				if (buffer == buf_end) break;
				if (token_p == &token[MAXTOKEN])
				{
					//gConsole->InsertLine("^1ERROR: Token too large on line %i",scriptline);
				}
			}
		}

		*token_p = 0;

		return 1;
	}

	int CShaderParser::GetOffset(void)
	{
		return (int) (buffer - buf_start);
	}

	void CShaderParser::GoTo(const int offset)
	{
		buffer = buf_start + offset;

	}
} // namespace