#include "texture.h"
#include <stdio.h>

#include <glheaders.h>
#include "../implementation/resman.h"

namespace FRONTEND
{

	CResTexture::CResTexture(CResourceManager *_resMan,char *_file, int _targetstatus) : CFEResource(_resMan,_file, _targetstatus)
	{
		info.fname=NULL;
		data=NULL;
	};

	CResTexture::CResTexture(CResourceManager *_resMan,char *_name, int _channels, int _format, int _sizex, int _sizey, void *_data)
		: CFEResource(_resMan,_name,RESSTATUS_INMEMORY)
	{
		data=(unsigned char*)_data;
		info.channels=_channels;
		info.format=_format;
		info.sizex=_sizex;
		info.sizey=_sizey;
		info.fname=_name==0?NULL:strdup(_name);
/*		if (_targetstatus>RESSTATUS_INMEMORY) 
		{
			LoadIntoGraphicAdapter(_targetstatus);
		}*/
	}

	CResTexture::~CResTexture() 
	{
		SAFE_FREE(data);
		SAFE_FREE(info.fname);
		/// @todo free info.fname. info.fname is not set everx time
//		if (info.fname!=NULL) free(info.fname);
	};


	CResTextureGL::CResTextureGL(CResManagerGL *_resMan,char *_file, int _targetstatus) : CResTexture(_resMan,_file, _targetstatus)
	{
		wrapModeS = wrapModeT = GL_REPEAT;
		status = 0; 
		if (_targetstatus>=RESSTATUS_INMEMORY) 
		{
			APPLICATION::CApplication::GetApplication()->GetLoadingthread().LoadResource(this,_file,_targetstatus);
//			while (APPLICATION::CApplication::GetApplication()->GetLoadingthread().IsLoading())
//			{
//				Sleep(1);
//			}
//			LoadIntoGraphicAdapter(_targetstatus);
/*			if (!LoadFromFile(_file, _targetstatus))
			{
				LoadFromFile("notfound.tga", _targetstatus);
			}*/
		}
	}
		
	CResTextureGL::~CResTextureGL()
	{
		SAFE_FREE(info.fname);
	}
		
	bool CResTextureGL::LoadFromFile(char *_file, int _targetstatus)
	{
		bool result=false;
		if ((status==RESSTATUS_INMEMORY) || (status==RESSTATUS_INMEMORYANDGA))
			return TRUE;

		char strr[512];
		
		if(strstr(_file, ".tga") || strstr(_file, ".TGA"))
		{
			if (LoadTGA(_file))
			{
				sprintf(strr,"[  OK  ] ResourceManager: Loading Texture %s\n",_file);
				if (status==RESSTATUS_NOTLOADED)
				{
					status = RESSTATUS_INMEMORY;
					if (_targetstatus>RESSTATUS_INMEMORYANDGA)
					{
						LoadIntoGraphicAdapter(_targetstatus);
					}
				}
				result = true;
			}
			else sprintf(strr,"[FAILED] ResourceManager: Loading Texture %s\n",_file);
		} else if(strstr(_file, ".bmp") || strstr(_file, ".BMP"))
		{
			if (LoadBMP(_file))
			{
				sprintf(strr,"[  OK  ] ResourceManager: Loading Texture %s\n",_file);
				if (status==RESSTATUS_NOTLOADED)
				{
					status = RESSTATUS_INMEMORY;
					if (_targetstatus>RESSTATUS_INMEMORYANDGA)
					{
						LoadIntoGraphicAdapter(_targetstatus);
					}
				}
				result = true;
			}
			else sprintf(strr,"[FAILED] ResourceManager: Loading Texture %s\n",_file);
		}

		GENERAL::dprintf(strr);

		// unknown file extension
		if (!result)
		{
			/// @todo prevent possible stackoverflow
			return LoadFromFile("notfound.tga",_targetstatus);
		}
		return result;
	}

	bool CResTextureGL::LoadIntoGraphicAdapter(int _targetsatus)
	{
		
		if (status==RESSTATUS_INMEMORY)
		{

			int textureType = (info.channels==4)?GL_RGBA:GL_RGB;
			
			glGenTextures(1,&textureName);
			glBindTexture(GL_TEXTURE_2D, textureName);

			char strr[256];
			sprintf(strr,"[  OK  ] GLTexture: Binding %i\n",textureName);
			GENERAL::dprintf(strr);

			/****************/

			/*
			int i=0;
			while (info.sizex!=1) {i++; info.sizex>>=1;};
			info.sizex=info.sizex<<i;
			i=0;
			while (info.sizey!=1) {i++; info.sizey>>=1;};
			info.sizey=info.sizey<<i;
			*/

	//		while (info.sizex1024) info.sizex--;
	//		while (info.sizey!=0) info.sizex--;


			gluBuild2DMipmaps(GL_TEXTURE_2D, info.channels, info.sizex, 
		     				info.sizey, textureType, GL_UNSIGNED_BYTE, data);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_NEAREST);
			glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE); 
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, wrapModeS);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, wrapModeT);

			if (_targetsatus==RESSTATUS_ONLYINGA)
			{
				SAFE_FREE(data);
				status = RESSTATUS_ONLYINGA;
	            
			}
			else status = RESSTATUS_INMEMORYANDGA;
				
		}
		return true;
	}

	void CResTextureGL::SetWrapMode(int _sMode, int _tMode)
	{
		wrapModeS = _sMode;
		wrapModeT = _tMode==-1?_sMode:_tMode;

		if (status!=RESSTATUS_NOTLOADED)
		{
			glBindTexture(GL_TEXTURE_2D,GetTextureName());
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, wrapModeS);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, wrapModeT);
		}
	}


	bool CResTextureGL::UnloadFromGraphicAdapter()
	{
		glDeleteTextures(1,&textureName);
		return TRUE;
	}

	bool CResTextureGL::LoadTGA(char *_file)
	{
		WORD width = 0, height = 0;			
		byte length = 0;					
		byte imageType = 0;					
		byte bits = 0;						
		FILE *pFile = NULL;					
		int channels = 0;					
		int stride = 0;						
		int i = 0;

		if((pFile = GENERAL::morfOpenFile(_file, "rb")) == NULL) 
		{
			//file not found
//			GENERAL::dprintf("texture %s not found\n",_file);
			return FALSE;
		}

		fread(&length, sizeof(byte), 1, pFile);
	
		fseek(pFile,1,SEEK_CUR); 

		fread(&imageType, sizeof(byte), 1, pFile);
		
		fseek(pFile, 9, SEEK_CUR); 

		fread(&width,  sizeof(WORD), 1, pFile);
		fread(&height, sizeof(WORD), 1, pFile);
		fread(&bits,   sizeof(byte), 1, pFile);

		fseek(pFile, length + 1, SEEK_CUR); 

		if(imageType != TGA_RLE)
		{
			if(bits == 24 || bits == 32)
			{
				channels = bits / 8;
				stride = channels * width;
				data = ((unsigned char*)malloc(sizeof(unsigned char)*stride*height));

				for(int y = height-1; y>=0; y--)
				{
					unsigned char *pLine = &(data[stride * y]);

					fread(pLine, stride, 1, pFile);
				
					for(i = 0; i < stride; i += channels)
					{
						int temp     = pLine[i];
						pLine[i]     = pLine[i + 2];
						pLine[i + 2] = temp;
					}
				}
			}
/*	we do not use 16 bit TGA else if(bits == 16)
			{
				unsigned short pixels = 0;
				int r=0, g=0, b=0;

				channels = 3;
				stride = channels * width;
				data = ((unsigned char*)malloc(sizeof(unsigned char)*stride*height));

				for(int i = 0; i < width*height; i++)
				{
					fread(&pixels, sizeof(unsigned short), 1, pFile);
					
					b = (pixels & 0x1f) << 3;
					g = ((pixels >> 5) & 0x1f) << 3;
					r = ((pixels >> 10) & 0x1f) << 3;
					
					data[i * 3 + 0] = r;
					data[i * 3 + 1] = g;
					data[i * 3 + 2] = b;
				}
			}	*/
			else
				return NULL;
		}
		else
		{
			byte rleID = 0;
			int colorsRead = 0;
			channels = bits / 8;
			stride = channels * width;

			data = ((unsigned char*)malloc(sizeof(unsigned char)*stride*height));
			byte *pColors = ((byte*)malloc(sizeof(byte)*channels));

			while(i < width*height)
			{
				fread(&rleID, sizeof(byte), 1, pFile);
				if(rleID < 128)
				{
					rleID++;

					while(rleID)
					{
						fread(pColors, sizeof(byte) * channels, 1, pFile);

						data[colorsRead + 0] = pColors[2];
						data[colorsRead + 1] = pColors[1];
						data[colorsRead + 2] = pColors[0];

						if(bits == 32)
							data[colorsRead + 3] = pColors[3];

						i++;
						rleID--;
						colorsRead += channels;
					}
				}
				else
				{
					rleID -= 127;
					fread(pColors, sizeof(byte) * channels, 1, pFile);

					while(rleID)
					{
						data[colorsRead + 0] = pColors[2];
						data[colorsRead + 1] = pColors[1];
						data[colorsRead + 2] = pColors[0];

						if(bits == 32)
							data[colorsRead + 3] = pColors[3];

						i++;
						rleID--;
						colorsRead += channels;
					}
					
				}
					
			}
		}

		fclose(pFile);

		info.channels = channels;
		info.sizex    = width;
		info.sizey    = height;
		info.format	  = GL_UNSIGNED_BYTE;

		info.fname = strdup(_file);
	
		return TRUE;
	}

	bool CResTextureGL::LoadBMP(char *_file)
	{
		AUX_RGBImageRec *pBitmap = NULL;
		FILE *pFile = NULL;

		char *strr;

		if ((strr= GENERAL::morfFindFile(_file))==0) return false;

		pBitmap = auxDIBImageLoad(strr);				

		//tImage *pImage = (tImage *)malloc(sizeof(tImage));

		

		info.channels = 3;
		info.sizex = pBitmap->sizeX;
		info.sizey = pBitmap->sizeY;

		data = ((unsigned char*)malloc(sizeof(unsigned char)*info.sizex*info.sizey*3));

		for (int i=0; i<info.sizey; i++)
		{
			memcpy(data+(info.sizey-i-1)*pBitmap->sizeX*info.channels,pBitmap->data+pBitmap->sizeX*info.channels*i,pBitmap->sizeX*info.channels);
		}
		//memcpy(data,pBitmap->data,info.channels*info.sizex*info.sizey);

		//data = pBitmap->data;

		/// @todo bin nicht sicher ob das free notwqendig ist 
		free(pBitmap->data);
		free(pBitmap);

		return true;
	}









}