#ifndef RES_TEXTURE_H
#define RES_TEXTURE_H

#include <morfbase.h>

#include <glheaders.h>



#define TGA_RGB		 2		
#define TGA_A		 3		
#define TGA_RLE		10

namespace FRONTEND
{

	class CResManagerGL;

	struct TResTextureInfo
	{
		char *fname;

		int channels;
		int sizex,sizey;
		int format;

		TResTextureInfo() {};

		TResTextureInfo(char *_fname, 
						int _channels, int _sizex, int _sizey, int _format)
		{
			
			fname = _fname;
			channels = _channels;
			sizex = _sizex;
			sizey = _sizey;
			format = _format;
		};


	};


	class CResTexture : public CFEResource
	{
	protected:

		TResTextureInfo info;
		unsigned char *data;
		int wrapModeS;
		int wrapModeT;

	public:

		

		CResTexture(CResourceManager *_resMan,char *_file, int _targetstatus);
		CResTexture(CResourceManager *_resMan,char *_name, int _channels, int _format, int _sizex, int _sizey, void *_data);
		
		virtual ~CResTexture();
		
		virtual bool LoadFromFile(char *_file, int _targetstatus) = 0;
		virtual bool LoadIntoGraphicAdapter(int _targetsatus) = 0;
		virtual bool UnloadFromGraphicAdapter() = 0;
		virtual void SetWrapMode(int _sMode, int _tMode = -1) = 0;

		inline TResTextureInfo &GetTextureInfo() { return info; };
	};

	class CResTextureGL : public CResTexture
	{
	protected:

		GLuint textureName;

		bool LoadTGA(char *_file);
		bool LoadBMP(char *_file);


	public:
		CResTextureGL(CResManagerGL *_resMan,char *_file, int _targetstatus);
		inline CResTextureGL(CResManagerGL *_resMan,char *_name, int _channels, int _format, int _sizex, int _sizey, void *_data)
			: CResTexture((CResourceManager*)_resMan,_name,_channels,_format,_sizex,_sizey,_data) {};
		
		virtual ~CResTextureGL();
		
		virtual bool LoadFromFile(char *_file, int _targetstatus);
		virtual bool LoadIntoGraphicAdapter(int _targetsatus);
		virtual bool UnloadFromGraphicAdapter();
		virtual void SetWrapMode(int _sMode, int _tMode = -1);

		inline GLuint GetTextureName() 
			{ 
				if (status==RESSTATUS_NOTLOADED)
					return 0;
				if (status==RESSTATUS_INMEMORY) 
					LoadIntoGraphicAdapter(RESSTATUS_ONLYINGA);

				return textureName; 
			};
		

	};

	typedef sptr<CResTexture> HResTexture;
	typedef sptr<CResTextureGL> HResTextureGL;






}
#endif