#ifndef APPLICATION_H
#define APPLICATION_H


#include "../../general/base/morfforward.h"
#include "../../general/base/morfmodules.h"
#include "../../frontend/base/frontend.h"
#include "loadingthread.h"

namespace APPLICATION
{

class appexception : public exception
{
public:
	typedef enum {
		ERR_REGISTER_WNDCLASS,
		ERR_CREATE_WINDOW,
		ERR_DX_CREATE,
		ERR_DXDEVICE_CREATE,
		ERR_UNKNOWN = 0xFFFF
	} TErrorTyp;
protected:
	TErrorTyp __typ;

public:
	appexception(TErrorTyp _typ, LPCSTR szMessage);
	appexception(LPCSTR szMessage);
	appexception();

	inline TErrorTyp typ() {return __typ;};
};

class CApplication
{
private:
	DOMAIN::CDomain *domain;
	NETWORK::CNetwork *network;
			
	GENERAL::CGameTick gameTicker;
	CLoadingThread loadingThread;

public:
	inline CApplication()
		: domain(NULL),
		  network(NULL)//,
//		  gamedb(NULL)
	  {}
	virtual ~CApplication();

	virtual void RegisterClasses();
	virtual bool Init();
	virtual void ExecuteSlice() = 0;
//	virtual void Render() = 0;  ///< Should be deleted Later and a Frontendclass inserted
	virtual int Run();

	static CApplication *app;
	inline static CApplication *GetApplication() {return app;}
	inline static void SetApplication(CApplication *_app) {app=_app;}

	inline DOMAIN::CDomain *GetDomain() {return domain;}
	inline void SetDomain(DOMAIN::CDomain *_domain) {domain=_domain;}
	inline NETWORK::CNetwork *GetNetwork() {return network;}
	inline void SetNetwork(NETWORK::CNetwork *_network) {network=_network;}
	//inline GAMEDB::CGameDatabase *GetGameDB() {return gamedb;}
	//inline void SetGameDB(GAMEDB::CGameDatabase *_gamedb) {gamedb=_gamedb;}
	//inline FRONTEND::CFrontEnd *GetFrontend() {return frontend;}
	//inline void SetFrontend(FRONTEND::CFrontEnd *_frontend) {frontend=_frontend;}
	
	inline CLoadingThread &GetLoadingthread() {return loadingThread;}
	inline GENERAL::CGameTick &GetTicker() {return gameTicker;}
	virtual FRONTEND::CFrontEnd *GetFE();

};

class CWinApplication : public CApplication
{
protected:
	
public:
	static LRESULT CALLBACK staticWindowProcedure(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam );
	virtual LRESULT CALLBACK WindowProcedure(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam ) = 0;
protected:	

	virtual ATOM RegisterWindowClass(HINSTANCE hInstance) = 0;

	HINSTANCE hInstance;
	HWND hWnd;
	LPCSTR szClassName;
	LPCSTR szTitle;

public:
	inline CWinApplication(HINSTANCE _hInstance,LPCSTR _szClassName)
		: CApplication(),
		  hInstance(_hInstance),
		  szClassName(_szClassName)
		{}
	virtual ~CWinApplication();

	virtual bool Init();

	virtual void ExecuteSlice();
};

}
#endif