#ifndef LOADINGTHREAD_H
#define LOADINGTHREAD_H

#include <windows.h>

#include "../../frontend/base/resource.h"
#include "../../domain/objects/objects.h"

#include <queue>

namespace APPLICATION
{
	class CLoadingThread;

	class CLoadingQueuesItem
	{
	public:
		inline CLoadingQueuesItem() {};
		virtual ~CLoadingQueuesItem();

		virtual bool Load() = 0;
		virtual char *GetStatusString() = 0;
	};

	class CLoadingQueuesItemResource : public CLoadingQueuesItem
	{
	protected:
		char *file;
		char statusString[100];
		FRONTEND::CFEResource *res;
		int targetstatus;
		CLoadingThread *loadingThread;

	public:
		inline CLoadingQueuesItemResource(CLoadingThread *_loadingThread, FRONTEND::CFEResource *_res, char *_file, int _targetstatus) 
			: CLoadingQueuesItem()
			{
				loadingThread = _loadingThread;
				res = _res;
				targetstatus = _targetstatus;
				file = strdup(_file);
				//statusstring = ;
				sprintf(statusString, "Loading %s", _file);
			}
		virtual ~CLoadingQueuesItemResource();


		virtual bool Load();

		inline int GetTargetStatus() { return targetstatus; };
		inline FRONTEND::CFEResource *GetResource() { return res; };
		
		virtual char *GetStatusString();

	};

	class CLoadingQueuesItemGameObject : public CLoadingQueuesItem
	{
	protected:
		DOMAIN::CGameObject *go;
		char statusString[100];

	public:
		inline CLoadingQueuesItemGameObject(DOMAIN::CGameObject *_go) 
			: CLoadingQueuesItem()
			{
				go = _go;
				sprintf(statusString, "Loading Object %i", _go->GetID());
			}
		virtual bool Load();
		virtual char *GetStatusString();

	};

	class CLoadingThread
	{
	public:
		typedef std::queue<CLoadingQueuesItemResource*> TLoadedResources;

	protected:
		std::queue<CLoadingQueuesItem*> loadingQueues;
		TLoadedResources loadedResources;

		HANDLE semaphore;
		HANDLE mutex;
		HANDLE thread;

		bool exitThread;
		bool loading;

		char *statusString;

	public:
		CLoadingThread();
		virtual ~CLoadingThread();

		bool IsLoading();
		void LoadResource(FRONTEND::CFEResource *_res,  char *_file, int _targetstatus);
		void LoadGameObject(DOMAIN::CGameObject *_go);
		void GetLoadedResources(TLoadedResources &_loadedResources);

		static void staticThread(CLoadingThread *obj);
		void loadingThread();
		int GetQueuesSize();
		inline void WaitFor()
		{
			while (loading)
				Sleep(10);
		}

		inline char *GetStatusString() { return statusString; }

		friend CLoadingQueuesItemResource;
	};
} // namespace
#endif