#ifndef OBJECTMAN_H
#define OBJECTMAN_H

#include <map>
#include <list>
#include <vector>
#include "domain.h"

//#include "../objects/objects.h"
//#include "../../game/objectdescriptor.h"


namespace DOMAIN
{

	/** Object types:
	*
	*	Transient Objects
	*   -----------------
	*	TObjects are immediatly deleted after creation and use.
	*
	*	Semipersistant Objects
	*   ----------------------
	*	SPObjects will be delete when the Aplication goes down or before.
	*
	*	Examples: GUI-popups
	*
	*
	*	Persistant Objects
	*   ------------------
	*	PObjects remain even if a server or client goes down, and are stored in 
	*	the game database (serverside) or the local filesystem.
	*
	*	Examples: Bitmaps, Playerdata, Items
	*	
	*/

	/** Object properties
	*
	*	Local Objects
	*	-------------
	*	Data is stored on local filesystem or database.
	*
	*	Remote Objects
	*	--------------
	*	Data is stored on another machine.
	*/

	class CDomain;

	class CObjectmanager
	{
	protected:

		CDomain *domain;

	public:

		inline CObjectmanager(CDomain *_domain) : domain(_domain) {};
		virtual ~CObjectmanager();

		virtual sptr<DOMAIN::CGameObject> GetRef(long _id) = 0;
		virtual sptr<DOMAIN::CGameObject> GetPtr(long _id) = 0;

		virtual long GetId(sptr<DOMAIN::CGameObject> _obj) = 0;
		virtual void Add(sptr<DOMAIN::CGameObject> _obj) = 0;
        virtual void Release(sptr<DOMAIN::CGameObject> _obj) = 0;
		virtual void Release(long _id) = 0;
//		virtual void ReleaseID(long _id) = 0;
		virtual void PrintDebug(char *st = "objs") = 0;

		virtual sptr<GAME::CObjectDescriptor> GetDescriptor(char *_name) = 0;
		virtual sptr<GAME::CObjectDescriptor> GetDescriptor(int _id) = 0;
		virtual long GetDescriptorID(char *name) = 0; 

		inline CDomain *GetDomain() { return domain; };

		virtual long CreateGlobalId();
	};
}

#endif