#ifndef DOMAIN_LIGHT_H
#define DOMAIN_LIGHT_H

#include "../../general/base/serialize.h"
#include "../../general/base/morfmath.h"
#include "levelobjects.h"

namespace DOMAIN
{
	class CLight : public CLevelObject
	{
	public:
		DECLARE_SERIALIZEABLE(CLight);

	protected:

		GENERAL::CVector4f ambientColor;
		GENERAL::CVector4f diffuseColor;
		GENERAL::CVector4f specularColor;
		/// todo: pos is vec3 should be vec4 for GL_LIGHTS
		
		float spotExponent;
		float spotCutoff;

		float constantAtt;
		float linearAtt;
		float quadAtt;

	public:

		inline CLight(CDomain *_domain)
			  : CLevelObject (_domain) , 
				ambientColor (0.0f,0.0f,0.0f,1.0f),
				diffuseColor (1.0f,1.0f,1.0f,1.0f),
				specularColor(1.0f,1.0f,1.0f,1.0f),
				spotExponent(0.0f),
				spotCutoff(180.0f),
				constantAtt(1.0f),
				linearAtt(0.0f),
				quadAtt(0.0f)
				
		{
			SetClassKey(GET_KEY(CLight));
		}

		inline CLight(CDomain *_domain, int _odid,long _id, CLevel *_lvl) 
			  : CLevelObject (_domain,_odid,_id, _lvl) , 
				ambientColor (0.0f,0.0f,0.0f,1.0f),
				diffuseColor (1.0f,1.0f,1.0f,1.0f),
				specularColor(1.0f,1.0f,1.0f,1.0f),
				spotExponent(0.0f),
				spotCutoff(180.0f),
				constantAtt(1.0f),
				linearAtt(0.0f),
				quadAtt(0.0f)
		{
			SetClassKey(GET_KEY(CLight));
		}

		virtual long GetSerializedSize();
		virtual long Serialize(GENERAL::CBuffer &buffer);
		virtual long DeSerialize(GENERAL::CBuffer &buffer);

		inline GENERAL::CVector4f &GetAmbientColor() { return ambientColor; };
		inline GENERAL::CVector4f &GetDiffuseColor() { return diffuseColor; };
		inline GENERAL::CVector4f &GetSpecularColor() { return specularColor; };

		inline float GetSpotExponent()	{ return spotExponent; };
		inline float GetSpotCutoff()	{ return spotCutoff; };
		inline float GetConstantAtt()	{ return constantAtt; };
		inline float GetLinearAtt()		{ return linearAtt; };
		inline float GetQuadAtt()		{ return quadAtt; };

		inline void SetAmbientColor(GENERAL::CVector4f &_val)  {  ambientColor= _val; };
		inline void SetDiffuseColor(GENERAL::CVector4f &_val)  {  diffuseColor= _val; };
		inline void SetSpecularColor(GENERAL::CVector4f &_val) {  specularColor= _val; };

		inline void SetSpotExponent(float _val)	{  spotExponent= _val; };
		inline void SetSpotCutoff(float _val)	{  spotCutoff= _val; };
		inline void SetConstantAtt(float _val)	{  constantAtt= _val; };
		inline void SetLinearAtt(float _val)	{  linearAtt= _val; };
		inline void SetQuadAtt(float _val)		{  quadAtt= _val; };
	};
}

#endif