// ijhuhuhu
#ifndef DOMAIN_BASEOBJECTS_H
#define DOMAIN_BASEOBJECTS_H

#include "../../general/base/serialize.h"
#include "../../general/base/smartpointer.h"

#include "../../general/base/morfmath.h"
#include "../../general/base/morftypes.h"
#include "../../general/base/morftools.h"
#include "../../game/objectdescriptor.h"

#include "../../frontend/base/renderable.h"

#include <deque>

using namespace GENERAL;

namespace FRONTEND
{
	class CRenderAble;
}

namespace GAME
{
	class CGameFSM;
}


#define IMPLEMENT_SERIALIZEABLE_DOMAIN(classname) \
	sptr<GENERAL::CSerializeableObject> classname::CreateObject(void *data) \
		{ return sptr<GENERAL::CSerializeableObject>(new classname((CDomain*)data)); }

#define DOMAINCMD_CONTROLLOBJECT		1
#define DOMAINCMD_PINGREQUEST			2
#define DOMAINCMD_PINGREPLAY			3
#define DOMAINCMD_SYNCHTIME				4
#define DOMAINCMD_GOTCONTROLLOBJECT     5
#define DOMAINCMD_PDU					6

#define DOMAINCMD_OBJECTACTION_REQUEST	7				
#define DOMAINCMD_OBJECTACTION_REPLY	8	
#define DOMAINCMD_REQUEST_OBJECT		9
#define DOMAINCMD_REQUEST_OBJECT_REPLY	10
#define DOMAINCMD_KILLED				11

#define PY_GETSET_ACCESSOR(func) \
	static PyObject *Py##func(PyObject *self, PyObject *args)
#define PY_GET_ACCESSOR_IMPL(classname,func,convertor) \
	PyObject *classname::Py##func(PyObject *self, PyObject *args) \
	{ \
		classname *obj = (classname*)PyCObject_AsVoidPtr(PyTuple_GetItem( args, 0)); \
		return convertor(obj->func()); \
	} 
#define PY_SET_ACCESSOR_IMPL(classname,func,convertor) \
	PyObject *classname::Py##func(PyObject *self, PyObject *args) \
	{ \
		classname *obj = (classname*)PyCObject_AsVoidPtr(PyTuple_GetItem( args, 0)); \
		obj->func(convertor(PyTuple_GetItem( args, 1))); \
		return PyInt_FromLong(1); \
	} 
#define PY_FUNC_DECLARATION(classname,func,args) { #func, DOMAIN::classname::Py##func, args, "" }

namespace DOMAIN
{


#define ACTIONLISTTYPE_NORMAL	-1
#define ACTIONLISTTYPE_POPUP	 1





	class CDomain;

	

	class CLight;
	enum  {
		ACT = 0,
		PDU = 1,
		numPDU
	};

	class CDomainCommand;

	struct TDomainObjectAction
	{
		long sourceObj;
		long action;

		inline TDomainObjectAction(long _action, long _sourceObj) : 
				action(_action), 
				sourceObj(_sourceObj) {};
	};
	struct TDomainObjectAction1P : TDomainObjectAction
	{
		long destObj;

		inline TDomainObjectAction1P(long _action, long _sourceObj, long _destObj) : 
				TDomainObjectAction(_action, _sourceObj),
				destObj(_destObj) {};
	};
	struct TDomainObjectAction1P1S : TDomainObjectAction1P
	{
		char text[128];

		inline TDomainObjectAction1P1S(long _action, long _sourceObj, long _destObj, char *_text) 
				: TDomainObjectAction1P(_action, _sourceObj,_destObj)
					{ strncpy(text,_text,128); }
				/// @todo Memorleak created
	};


	class CDomainObject : public GENERAL::CSerializeableObject
	{
	protected:
		CDomain *domain;

	public:
		DECLARE_SERIALIZEABLE(DOMAIN::CDomainObject);
		
		inline CDomainObject(CDomain *_domain)
			: GENERAL::CSerializeableObject(), domain(_domain) { SetClassKey(GET_KEY(CDomainObject));}

		virtual ~CDomainObject();

		inline void SetDomain(CDomain *_domain) { domain = _domain; };
		inline CDomain *GetDomain() { return domain; };
	};

	/** The CGameObject is the Baseclass of all Game Relevant Objects
	*	
	*	the new variable id stores a global unique id of the Object
	*/
	class CGameObject : public CDomainObject
	{
	public:
		static CGameObject *selectedGameObject;

		

		class tActionList : public std::vector<long>
		{
		public:
			tActionList operator & (tActionList &al);
			tActionList operator | (tActionList &al);
		};

		typedef morfmap<long,tActionList > tActions;


		typedef struct _TSerializedData : public GENERAL::CSerializeableObject::TSerializedData
		{
			long id;
			GENERAL::TGameTick timestamp;
			long who;
		} TSerializedData;

		struct TPDU
		{
			GENERAL::TGameTick timestamp;
			unsigned short fsmState;
		};
	protected:
		/// Global unique id of the Object

		bool canDelete;
		long id;
		long odid;
		
		unsigned short fsmState[numPDU];
		GENERAL::TGameTick timestamp[numPDU];

		long who;
		bool doDeathReckoning;
		bool loaded;

		tActions actions;
		tActionList popupactions;

		std::string caption;

		long inContainerType;

	public:
		
		DECLARE_SERIALIZEABLE(CGameObject);
		
		CGameObject(CDomain *_domain);
		CGameObject(CDomain *_domain, int _odid, long _id);

		virtual ~CGameObject();
		inline bool Loaded(){return loaded;};
		virtual void Done();
		virtual bool Load();

		virtual long GetSerializedSize();
		virtual long Serialize(GENERAL::CBuffer &buffer);
		virtual long DeSerialize(GENERAL::CBuffer &buffer);
		virtual void CopyPDUto(char *data);
		virtual void SetPDU();
		virtual void CopyPDUfrom(char *data);
		virtual int GetPDUSize();
		virtual float GetPDUTreshHold();
		virtual void GetSerializeObjectsBefore(std::vector<sptr<CGameObject> > &_objs);

		virtual void TimeStep(GENERAL::TGameTick dt);

		inline bool GetCanDelete() {return canDelete;};
		inline void SetCanDelete(bool _canDelete) {canDelete=_canDelete;};
		inline long GetID() {return id;};
		inline void SetID(long _id) { id = _id; };
		inline long GetWho() {return who;};
		inline void SetWho(long _who) { who = _who; };
		inline void SetTimeStamp(GENERAL::TGameTick _timestamp) {timestamp[ACT]=_timestamp;};
		inline void SetControl() {doDeathReckoning=false;};
		inline void SetDeathReckoning() {doDeathReckoning=true;};
		inline void SetFSMState(unsigned short _fsmState) {fsmState[ACT]=_fsmState;};
		inline unsigned short GetFSMState() {return fsmState[ACT];};
		inline bool IsControled() {return !doDeathReckoning;};
		inline GENERAL::TGameTick &GetTimeStamp() {return timestamp[ACT];};
		virtual void Print(FILE *f,bool newline=true);
		virtual void GetActionList(tActionList &_actions, int witch=-1);// current


		virtual int ReadObjectDescriptor(sptr<GAME::CObjectDescriptor> _od);

		inline const char *GetCaption() { return caption.data(); };
		inline void SetCaption(char *_cap) { caption = _cap; };

		inline long GetContainerType() { return inContainerType; };
		inline void SetContainerType(long _ct) { inContainerType = _ct; };

		
		virtual void InitiateAction(TDomainObjectAction *act, unsigned short size);
		virtual void HandleDomainObjectActionRequest(sptr<CDomainCommand> &cmd);		
		virtual void HandleDomainObjectActionReply(sptr<CDomainCommand> &cmd);

		
	};

	class CGraphObject : public CGameObject
	{
	public:

		typedef std::deque<sptr<CGraphObject> > tChildsType;
		typedef std::deque<sptr<CGraphObject> > tLightsType;
		typedef std::deque<FRONTEND::CRenderAble*> tRenderableList;


	protected:
	
		GAME::CGameFSM *fsm;

		sptr<FRONTEND::CRenderAble> feData;
		tRenderableList updateRenderables;
		long feRepresentation;
		unsigned short int feStatus;

		/* SCENEGRAPH functionality */
		CGraphObject *parent;
		tChildsType childs;
		tLightsType lights; 
		/* end SCENEGRAPH functionality */

		long parentId;

		long ownedChatChannel;

		std::string icon;


	public:
		
		DECLARE_SERIALIZEABLE(CGameObject);
		CGraphObject(CDomain *_domain);
		CGraphObject(CDomain *_domain, int _odid,long _id);
		virtual ~CGraphObject();

		void Command(char *_cmd);
		
		inline sptr<FRONTEND::CRenderAble> &GetfeData() {return feData;}
		inline void SetfeData(sptr<FRONTEND::CRenderAble> &_feData) {feData=_feData;}

		inline long GetfeRepresentation() {return feRepresentation;}
		inline void SetfeRepresentation(long _feRep) {feRepresentation=_feRep;}

		inline unsigned int GetfeStatus() {return feStatus;}
		inline void SetfeStatus(unsigned short int _feStatus) {feStatus=_feStatus;}

		virtual long GetSerializedSize();
		virtual long Serialize(GENERAL::CBuffer &buffer);
		virtual long DeSerialize(GENERAL::CBuffer &buffer);
		virtual void Print(FILE *f,bool newline=true);

		/* SCENEGRAPH functionality */

		inline CGraphObject *GetParent() { return parent; };
		inline tChildsType &GetChilds() { return childs; };
		inline bool HasChild(sptr<CGraphObject> _child)
		{
			return (std::find(childs.begin(),childs.end(),_child)!=childs.end());
		}

		inline long GetParentId() { return parentId; };

		inline GAME::CGameFSM *GetFSM() { return fsm; };

		virtual void SetParent(CGraphObject *_parent);
		virtual void AddChild(sptr<CGraphObject> _child);
		virtual void RemoveChild(sptr<CGraphObject> _child);
		virtual void HandoverChild(sptr<CGraphObject> _target, sptr<CGraphObject> _child);

        virtual void RegisterCamera(sptr<CGraphObject> _camera);
		virtual void RegisterLight(sptr<CLight> _light);
        virtual void UnRegisterCamera(sptr<CGraphObject> _camera);
		virtual void UnRegisterLight(sptr<CLight> _light);

		inline char *GetIcon() { return (char *)icon.data(); };
		inline void SetIcon(char *_icon) { icon = _icon; };

		virtual int ReadObjectDescriptor(sptr<GAME::CObjectDescriptor> _od);
		inline void AddUpdateRenderable(FRONTEND::CRenderAble *_ro) { updateRenderables.push_back(_ro); }
		inline void RemoveUpdateRenderable(FRONTEND::CRenderAble *_ro) 
		{
			tRenderableList::iterator it = std::find(updateRenderables.begin(),updateRenderables.end(),_ro); 
			if (it!=updateRenderables.end())
				updateRenderables.erase(it);
		}
		void UpdateRenderables(int _flags);

		/* end SCENEGRAPH functionality */

		inline void SetOwnedChatChannel(long _id) { ownedChatChannel = _id; }
		inline long GetOwnedChatChannel() { return ownedChatChannel; };
		virtual long CreateOwnedChatChannel();

		virtual void GetSerializeObjectsBefore(std::vector<sptr<CGameObject> > &_objs); 

		//virtual int ExecuteAction(int _actionId, sptr<CGraphObject> &_caller, long lParam = 0, void *pParam = 0);
		virtual int ExecuteAction(int _actionId, CGraphObject *_caller, long lParam = 0, void *pParam = 0);
		virtual int ActionChatReply(CGraphObject *_caller, long _whoId,char *_msg);

		PY_GETSET_ACCESSOR(GetParent);
	};

	class CWorldObject : public CGraphObject
	{
	public:
		struct TPDU : public CGameObject::TPDU
		{
			GENERAL::CVector3f pos;
			GENERAL::CVector3f rotation;
		};

	protected:
		GENERAL::CVector3f pos[numPDU];
		GENERAL::CVector3f scale;
		GENERAL::CVector3f rotation[numPDU];

	public:
		DECLARE_SERIALIZEABLE(CGameObject);
		inline CWorldObject(CDomain *_domain) 
			: CGraphObject(_domain), 
			  pos(), 
			  scale(1.0f,1.0f,1.0f), 
			  rotation() 
		{
			SetClassKey(GET_KEY(CWorldObject));
			pos[PDU]=GENERAL::CVector3f(0.0f,0.0f,0.0f);
			rotation[PDU]=GENERAL::CVector3f(1.0f,0.0f,0.0f);
		}
		inline CWorldObject(CDomain *_domain, int _odid, long _id) 
			: CGraphObject(_domain,_odid, _id), 
			  pos(), 
			  scale(1.0f,1.0f,1.0f), 
			  rotation() 
		{
			SetClassKey(GET_KEY(CWorldObject));
		}

		virtual long GetSerializedSize();
		virtual long Serialize(GENERAL::CBuffer &buffer);
		virtual long DeSerialize(GENERAL::CBuffer &buffer);
		virtual void CopyPDUto(char *data);
		virtual void SetPDU();
		virtual void CopyPDUfrom(char *data);
		virtual int GetPDUSize();
		virtual float GetPDUTreshHold();
		virtual void Print(FILE *f,bool newline=true);
//		virtual void Assign(CGameObject *o);

		inline SetPos(GENERAL::CVector3f &_pos) {pos[ACT]=_pos;};
		inline SetScale(GENERAL::CVector3f &_scale) {scale=_scale;};
		inline SetRotation(GENERAL::CVector3f &_rotation) {rotation[ACT]=_rotation;};
		inline GENERAL::CVector3f &GetPos() {return pos[ACT];};
		inline GENERAL::CVector3f &GetScale() {return scale;};
		inline GENERAL::CVector3f &GetRotation() {return rotation[ACT];};

		PY_GETSET_ACCESSOR(GetPos);
		PY_GETSET_ACCESSOR(GetRotation);
//		PY_GETSET_ACCESSOR(SetPos);
	};
}

#endif