#include "gamefsm.h"

#include "../../general/base/morftools.h"
#include "../../domain/objects/objects.h"
#include "../../general/base/trace.h"


using namespace GENERAL;

namespace GAME
{
	CGameFSM::CGameFSM(DOMAIN::CGraphObject *_owner)
	{
		owner = _owner;
	}

	CGameFSM::~CGameFSM()
	{
	}

	bool CGameFSM::Init(char *_filename)
	{
		GENERAL::CParser parser;
		parser.LoadFromFile(_filename);

		CParser::TNode *root = parser.GetRoot();
		CParser::TNode *constNode = NULL;
		CParser::TNode *fsmNode = NULL;
		CParser::TNode *commandsNode = NULL;
		CParser::TNode::TChildren::iterator it,it2,fit,lit;

		it = root->children.find("CONST");
		if (it!=root->children.end())
			constNode = it->second;
		it = root->children.find("FSM");
		if (it!=root->children.end())
			fsmNode = it->second;
		it = root->children.find("COMMANDS");
		if (it!=root->children.end())
			commandsNode = it->second;

		// Load Constants;
		if (constNode!=NULL)
		{
			for (it = constNode->children.begin(); it != constNode->children.end(); it++)
			{
				constants[it->first] = it->second->val;
			}
		}

		if (commandsNode!=NULL)
		{
			int commandNum=1;
			
			for (it = commandsNode->children.begin(); it != commandsNode->children.end(); it++)
			{
				commandNames[it->first] = commandNum;
				commandNum++;
			}
		} 
		else 
		{
			dprintf("[ERROR] CGameFSM: %s, no commands defined\n",_filename);
			return false;
		}

		if (fsmNode!=NULL)
		{
			for (it = fsmNode->children.begin(); it != fsmNode->children.end(); it++)
			{
				if (it->second->val!=0) continue;

				TStateSet *currentSet = new TStateSet;
				currentSet->stateSetName=strdup(it->first.data());
				stateSets[it->first] = currentSet;
			}

			for (it = fsmNode->children.begin(); it != fsmNode->children.end(); it++)
			{
				if (it->second->val!=0) continue;

				TStateSet *currentSet = stateSets[it->first];

				ReadSet(it->second,currentSet,_filename,parser);
			}




	
		} 
		else 
		{
			dprintf("[ERROR] CGameFSM: %s, no statesets defined\n",_filename);
			return false;
		}

		return true;
	}

	int CGameFSM::ReadSet(GENERAL::CParser::TNode *fsmNode, TStateSet *set, char *_filename, GENERAL::CParser &parser)
	{
		CParser::TNode::TChildren::iterator it,it2,fit,lit;

		if (fsmNode!=NULL)
		{
			TState *curState;
			
			for (it = fsmNode->children.begin(); it != fsmNode->children.end(); it++)
			{
				if (it->second->val!=NULL)
				{
					continue;
				}
				curState = new TState();
				curState->stateName = strdup(it->first.data());
				
				it2 = it->second->children.find("params");
				if (it2!=it->second->children.end())
				{
					char *paramnames = strdup(it2->second->val);
					char *token = strtok(paramnames,",");

					while (token!=NULL)
					{
						curState->params[token]="";
						token = strtok(NULL,",");
					}

					SAFE_FREE(paramnames);
				}

				curState->enabled = STATE_ENABLED;
				curState->disabledBy = "";

				set->states[curState->stateName] = curState;
			}

			for (it = fsmNode->children.begin(); it != fsmNode->children.end(); it++)
			{
				if (it->second->val!=NULL)
				{
					if (stricmp(it->first.data(),"defaultState")==0)
					{
						set->currentState = set->GetState(it->second->val);
					}
					if (stricmp(it->first.data(),"parent")==0)
					{
						set->parent = GetStateSet(it->second->val);

						if (set->parent!=0)
						{
							set->parent->childs.push_back(set);
						}
					}
					continue;
				}

				curState = set->GetState((char*)it->first.data());

				parser.GetNodes(it->second,"nextstate",fit,lit);
				while (fit!=lit)
				{
					char *val = strdup(fit->second->val);
					char *command, *stateName;

					command = strtok(val,":");
					if (command!=NULL)
					{
						stateName = strtok(NULL,"(");
						if (stateName!=NULL)
						{
							int cmd = GetCommand(command);
							TState *nState = set->GetState(stateName);

							if ((cmd!=0) && (nState!=NULL))
							{
								curState->nextStates[cmd]=nState;
							}
							else
							{
								dprintf("[ERROR] invalid nextState in %s\n",_filename);
							}
							
						} 
						else
						{
							dprintf("[ERROR] invalid nextState in %s\n",_filename);
						}
					} 
					else
					{
						dprintf("[ERROR] invalid nextState in %s\n",_filename);
					}
					
					SAFE_FREE(val);
					fit++;
				}

				parser.GetNodes(it->second,"script",fit,lit);
				while (fit!=lit)
				{
					char *val = strdup(fit->second->val);
					char *event, *scriptModule,*scriptFunction;

					event = strtok(val,":");
					if (event!=NULL)
					{
						scriptModule = strtok(NULL,".");
						if (scriptModule!=NULL)
						{
							scriptFunction = strtok(NULL,".");
						}
						if ((scriptModule==0) || (scriptFunction==0))
						{
							dprintf("[WARN] invalid script parameters in %s -> skipping\n",_filename);
							fit++;
							continue;
						}
						TScriptEvent se;

						se.module=scriptModule;
						se.func=scriptFunction;
                        
						curState->EventMap[event]=se;
					}
					SAFE_FREE(val);
					fit++;
			
				}
			
			}
			


		}
		return 0;
	}

	std::vector<std::string> CGameFSM::GetStateSets()
	{
		TStateSetMap::iterator it;
		std::vector<std::string> res;

		for (it = stateSets.begin(); it != stateSets.end(); it++)
		{
			res.push_back(it->first);
		}
		return res;
	}


	void CGameFSM::Command(char *_cmd)
	{
		
		int cmd=GetCommand(_cmd);

		TStateSetMap::iterator it;
		for (it=stateSets.begin(); it!=stateSets.end(); it++)
		{
			TState *nextState = it->second->GetNextState(cmd);
			if (nextState!=NULL)
			{
				it->second->currentState=nextState;
				if (nextState->HasEventScript("onEnter"))
				{
					char *module = (char*)nextState->EventMap["onEnter"].module.data();
					char *func = (char*)nextState->EventMap["onEnter"].func.data();

					CallScript(module,func,"g",owner);
				}
			}

		}

		owner->UpdateRenderables(0);
		
		//owner->SetOwnedChatChannel(5);
			
	}





}