#ifndef MORFMATH_H
#define MORFMATH_H

//#include <windows.h>
//#include <windows.h>
//#include <math.h>

#include <math.h>
#undef DOMAIN

namespace GENERAL
{

	#define EPSILON 1E-10f
	#define BSPEPSILON 0.0325f
	//0.3125f 
	#define PI 3.14159265359f

	typedef float SCALAR;

	template <typename A> class TQuaternion;
	template <typename A> class TVector2D;
	template <typename A> class TVector3D;
	template <typename A> class TVector4D;
	template <typename A> class TMatrix3D;
	template <typename A> class TMatrix4D;
	template <typename A> class TLine;
	template <typename A> class TPlane;
	template <typename A,typename TYP> class TQuadricBezierPatch;
	template <typename A,typename TYP> class TBiQuadricBezierPatch;
	template <typename A,typename TYP> class TCubicBezierPatch;

	typedef void  (_fastcall *TMult_V4xM44f)(float *matrix4, float *vector4, float *res);
	typedef void (_fastcall *TTranspose_M44f)(float *matrix4, float *res);

	extern TMult_V4xM44f   Mult_V4xM44f;
	extern TTranspose_M44f Transpose_M44f;

	void _fastcall Mult_V4xM44f_SSE(float *matrix4, float *vector4, float *res);
	void _fastcall Transpose_M44f_SSE(float *matrix4, float *res);
	void _fastcall Mult_V4xM44f_C(float *matrix4, float *vector4, float *res);
	void _fastcall Transpose_M44f_C(float *matrix4, float *res);

	typedef TQuaternion<SCALAR> CQuaternionf;
	typedef TVector2D<SCALAR> CVector2f;
	typedef TVector3D<SCALAR> CVector3f;
	typedef TVector4D<SCALAR> CVector4f;
	typedef TMatrix3D<SCALAR> CMatrix3f;
	typedef TMatrix4D<SCALAR> CMatrix4f;
	typedef TLine<SCALAR> CLinef;
	typedef TPlane<SCALAR> CPlanef;
	typedef TVector3D<int> CVector3i;

	typedef TQuadricBezierPatch<SCALAR,CVector3f> CQuadricBezierPatchVec3f;
	typedef TBiQuadricBezierPatch<SCALAR,CVector3f> CBiQuadricBezierPatchVec3f;
	typedef TCubicBezierPatch<SCALAR,CVector3f> CCubicBezierPatchVec3f;

	double Sin(double _alpha);
	double Cos(double _alpha);
	double Sqrt(double _alpha);

		


	// INFOS @
	// http://www.gamasutra.com/features/19980703/quaternions_01.htm
	// http://www.euclideanspace.com/maths/geometry/rotations/conversions/eulerToQuaternion/index.htm

	// BAD INFOS (be careful) @
	// http://www.j3d.org/matrix_faq/matrfaq_latest.html
	template <typename A> class TQuaternion
	{
	public:
		A x,y,z,w;

		TQuaternion() : x((A)0),y((A)0),z((A)0),w((A)0) {};
		TQuaternion(TVector3D<A> vec3d) { CreateFormEulerAngle(vec3d); };
		TQuaternion(A _rx, A _ry ,A _rz) { CreateFormEulerAngle(_rx,_ry,_rz ); };
		TQuaternion(A _x, A _y ,A _z, A _w) : x(_x), y(_y), z(_z), w(_w) {};

		TQuaternion(TMatrix4D<A> mat4f) { CreateFromMartrix4f(mat4f); };
		TQuaternion(TMatrix3D<A> mat3f) { CreateFromMartrix3f(mat3f); };


		TQuaternion<A> & operator = (TQuaternion<A> obj)
			{
				x=obj.x; y=obj.y; z=obj.z; w=obj.w;
				return *this;
			}

		inline TQuaternion<A> &Clear()
			{
				x=y=z=(A)0;
				w=(A)1;
				return *this;
			}

		inline TQuaternion<A> &CreateFromMartrix4D(TMatrix4D<A> mat4f)
			{
				//not implemented yet
			}

		inline TQuaternion<A> &CreateFromMartrix3D(TMatrix3D<A> mat3f)
			{
				//not implemented yet
			}

		inline TQuaternion<A> &CreateFormEulerAngle(A rx, A ry, A rz) 
			{
				A cr, cp, cy, sr, sp, sy, cpcy, spsy;
				A roll,pitch,yaw;


				roll  = rx/(A)2;
				pitch = ry/(A)2;
				yaw   = rz/(A)2;

				// calculate trig identities
				cr = (A)cos(roll);
				cp = (A)cos(pitch);
				cy = (A)cos(yaw);


				sr = (A)sin(roll);
				sp = (A)sin(pitch);
				sy = (A)sin(yaw);
				
				cpcy = cp * cy;
				spsy = sp * sy;


				w = cr * cpcy + sr * spsy;
				x = sr * cpcy - cr * spsy;
				y = cr * sp * cy + sr * cp * sy;
				z = cr * cp * sy - sr * sp * cy;
		
	/*			A		angle;
				A		sr, sp, sy, cr, cp, cy;
				angle = rz * 0.5f;
				sy = sinf(angle);
				cy = cosf(angle);
				angle = ry * 0.5f;
				sp = sinf(angle);
				cp = cosf(angle);
				angle = rx * 0.5f;
				sr = sinf(angle);
				cr = cosf(angle);

				x = sr*cp*cy-cr*sp*sy; 
				y = cr*sp*cy+sr*cp*sy; 
				z = cr*cp*sy-sr*sp*cy;
				w = cr*cp*cy+sr*sp*sy; */

				return *this;
			}
		inline TQuaternion<A> &CreateFormEulerAngle(TVector3D<A> rot) 
			{
				return CreateFormEulerAngle(rot.x,rot.y,rot.z);
			}
		inline operator TVector3D<A> (void)
			{
				A sqw = w*w;
				A sqx = x*x;
				A sqy = y*y;
				A sqz = z*z;

				return TVector3D<A> (
					(A)atan2(2.0 * (y*z + x*w),(-sqx - sqy + sqz + sqw)),
					(A)asin(-2.0 * (x*z - y*w)),
					(A)atan2(2.0 * (x*y + z*w),(sqx - sqy - sqz + sqw)));
			}

		inline operator TMatrix4D<A> ()
			{
				A wx, wy, wz, xx, yy, yz, xy, xz, zz, x2, y2, z2;



				// calculate coefficients
				x2 = x + x; y2 = y + y;
				z2 = z + z;
				xx = x * x2; xy = x * y2; xz = x * z2;
				yy = y * y2; yz = y * z2; zz = z * z2;
				wx = w * x2; wy = w * y2; wz = w * z2;

				return TMatrix4D<A> (
				// row 1
					1.0 - (yy + zz), 
					xy - wz,
					xz + wy, 
					0.0,
				// row 2
					xy + wz, 
					1.0 - (xx + zz),
					yz - wx, 
					0.0,
				// row 3
					xz - wy, 
					yz + wx,
					1.0 - (xx + yy), 
					0.0,
				// row 4
					0, 
					0,
					0, 
					1);

				/*			// row 1
					1.0 - (yy + zz), 
					xy + wz, 
					xz - wy, 
					0, 
				// row 2
					xy - wz,
					1.0 - (xx + zz),
					yz + wx,
					0,
				// row 3
					xz + wy, 
					yz - wx, 
					1.0 - (xx + yy), 
					0, 
				// row 4
					0.0,
					0.0,
					0.0,
					1);
	*/
			}

		inline TQuaternion<A> &operator *= (TQuaternion<A> q)
			{
				float A, B, C, D, E, F, G, H;

				A = (w + x)*(q->w + q->x);
				B = (z - y)*(q->y - q->z);
				C = (w - x)*(q->y + q->z); 
				D = (y + z)*(q->w - q->x);
				E = (x + z)*(q->x + q->y);
				F = (x - z)*(q->x - q->y);
				G = (w + y)*(q->w - q->z);
				H = (w - y)*(q->w + q->z);

				x = A - (E + F + G + H)/2,
				y = C + (E - F + G - H)/2, 
				z = D + (E - F - G + H)/2,
				w = B + (-E - F + G + H) /2);
				return *this;
			}
		inline TQuaternion<A> operator * (TQuaternion<A> q)
			{
				A  B, C, D, E, F, G, H, I;

				B = (z - y)*(q.y - q.z);
				C = (w - x)*(q.y + q.z); 
				D = (y + z)*(q.w - q.x);
				E = (x + z)*(q.x + q.y);
				F = (x - z)*(q.x - q.y);
				G = (w + y)*(q.w - q.z);
				H = (w - y)*(q.w + q.z);
				I = (w + x)*(q.w + q.x);

				return TQuaternion<A>(
					I - ( E + F + G + H)/(A)2,
					C + ( E - F + G - H)/(A)2, 
					D + ( E - F - G + H)/(A)2,
					B + (-E - F + G + H)/(A)2);
			}
		inline TQuaternion<A> Slerp(A t,TQuaternion<A> &to)
			{
				
				A  to1[4];
				A omega, cosom, sinom, scale0, scale1;

				// calc cosine
				cosom = x * to.x + y * to.y + z * to.z
						+ w * to.w;

				// adjust signs (if necessary)
				if ( cosom <0.0 ){ cosom = -cosom; to1[0] = - to.x;
				to1[1] = - to.y;
				to1[2] = - to.z;
				to1[3] = - to.w;
				} else  {
				to1[0] = to.x;
				to1[1] = to.y;
				to1[2] = to.z;
				to1[3] = to.w;
				}

				// calculate coefficients
			if ( (1.0 - cosom) > EPSILON ) {
						// standard case (slerp)
						omega = (A)acos(cosom);
						sinom = (A)sin(omega);
						scale0 = (A)sin((1.0 - t) * omega) / sinom;
						scale1 = (A)sin(t * omega) / sinom;
				} else {        
			// "from" and "to" quaternions are very close 
				//  ... so we can do a linear interpolation
						scale0 = (A)1 - t;
						scale1 = t;
				}
			// calculate final values
				return TQuaternion<A>(
					scale0 * x + scale1 * to1[0],
					scale0 * y + scale1 * to1[1],
					scale0 * z + scale1 * to1[2],
					scale0 * w + scale1 * to1[3]);
			}
	};

	template <typename A> class TVector2D
	{
	public:
		A x,y;
		TVector2D() {x=y=(A)0;};
		TVector2D(A _x, A _y) {x=_x; y=_y;};
		TVector2D(TVector2D<A> &_v) {x=_v.x; y=_v.y;};

		TVector2D & operator = (TVector2D<A> _v)
			{ x=_v.x; y=_v.y; };
	};

	template <typename A> class TVector3D
	{
	public:
		A x,y,z;
	
		TVector3D() {};
		TVector3D(A _x, A _y ,A _z) : x(_x), y(_y), z(_z) {};
		TVector3D(A a[]) {memcpy(&x,a,3*sizeof(A));};

		inline TVector3D<A> &Clear()
			{
				x=y=z=(A)0;
				return *this;
			}

		operator TVector4D<A> (void)
			{
				return TVector4D<A>(x, y, z, (A)1);
			}

		TVector3D<A> & operator = (const TVector4D<A> &obj) 
			{
				x=obj.x; y=obj.y; z=obj.z;
				return *this;
			}
		TVector3D<A> & operator = (const TVector3D<A> &obj) 
			{ 
				x=obj.x; y=obj.y; z=obj.z;
				return *this;
			};
		TVector3D<A> & operator = (const A obj[]) 
			{ 
				x=obj[0]; y=obj[1]; z=obj[2];
				return *this;
			};
		TVector3D<A> &operator += (const TVector3D<A> &obj) 
			{ 
				x+=obj.x; y+=obj.y; z+=obj.z;
				return *this;
			};
		TVector3D<A> & operator -= (const TVector3D<A> &obj) 
			{ 
				x-=obj.x; y-=obj.y; z-=obj.z;
				return *this;
			};
		TVector3D<A> & operator *= (const A obj) 
			{ 
				x*=obj; y*=obj; z*=obj;
				return *this;
			};
		TVector3D<A> operator *= (const TVector3D<A> &obj)
			{ 
				return ((*this)=VectorProduct(obj));
			};
		TVector3D<A> & operator /= (const A obj) 
			{ 
				x/=obj; y/=obj; z/=obj;
				return *this;
			};

		TVector3D<A> operator - (void)
			{ 
				return TVector3D(-x, -y, -z);
			};
		TVector3D<A> operator + (const TVector3D<A> &obj)
			{ 
				return TVector3D(x+obj.x, y+obj.y, z+obj.z);
			};
		TVector3D<A> operator - (const TVector3D<A> &obj)
			{ 
				return TVector3D(x-obj.x, y-obj.y, z-obj.z);
			};
		TVector3D<A> operator * (const A obj)
			{ 
				return TVector3D(x*obj, y*obj, z*obj);
			};


		TVector3D<A> operator * (const TVector3D<A> &obj)
			{ 
				return VectorProduct((TVector3D<A>)obj);
				//return DotProduct(obj);
			};
		TVector3D<A> operator / (const A obj)
			{ 
				return TVector3D(x/obj, y/obj, z/obj);
			};

		inline bool operator == (const TVector3D<A> obj)
			{
				return ((x==obj.x) && (y==obj.y) && (z==obj.z)); 
			};

		inline bool operator != (const TVector3D<A> obj)
			{
				return ((x!=obj.x) || (y!=obj.y) || (z!=obj.z)); 
			};


		inline A abs2()
			{
				return x*x+y*y+z*z;
			}
		inline A abs()
			{
				return (A) Sqrt(abs2());
			};
		inline TVector3D<A> &Normal()
			{
				A len = abs();
				if (len) return (*this)/=len;
				else return (*this)=TVector3D<A>(0.0,0.0,0.0); 
			}

		inline A ScalarMultiply(TVector3D<A> &vec)
			{
				return
					x*vec.x+y*vec.y+z*vec.z;
			}

		inline A Angle(TVector3D<A> &vec)
			{
				return acos(ScalarMultiply(vec));
			}
/////
//		inline A DotProduct(TVector3D<A> &v1, TVector3D<A> &v2)
		inline A DotProduct(TVector3D<A> &v)
			{
				return x*v.x + y * v.y + z*v.z;
			}

		inline TVector3D<A> VectorProduct(TVector3D<A> &vec)
			{
				return TVector3D<A>(
					y*vec.z-z*vec.y,
					z*vec.x-x*vec.z,
					x*vec.y-y*vec.x);
			}	
		inline TVector3D<A> MatrixMultiply(TMatrix4D<A> &mat)
			{
				return Vector3D(
					x*mat.m[ 0] +	y*mat.m[ 4] +	z*mat.m[ 8] +	((A)1)*mat.m[12],
					x*mat.m[ 1] +	y*mat.m[ 5] +	z*mat.m[ 9] +	((A)1)*mat.m[13],
					x*mat.m[ 2] +	y*mat.m[ 6] +	z*mat.m[10] +	((A)1)*mat.m[14])
			}
		inline A GetAlpha(TVector3D<A> &vec)
			{
				A res = ScalarMultiply(vec)/(abs()*vec.abs());
				res = res>(A)1?(A)1: (res<(A)-1?(A)-1:res);
				return (A)acos(res);
			}
		inline bool isNull()
			{
				return ((x==(A)0) && (y==(A)0) && (z==(A)0));
			}
		inline void getRotation(TVector3D<A> &xAxis,TVector3D<A> &yAxis, A *rotX, A *rotY)
			{
				CPlanegl pYZ(xAxis  ,TVector3D<A>(0.0f,0.0f,0.0f));
				CPlanegl pPY(yAxis*(*this),TVector3D<A>(0.0f,0.0f,0.0f));

				*rotY = _copysign( pPY.n.isNull()?0.0f:pYZ.GetAlpha(pPY) ,pYZ.Solve(*this) );

				CPlanegl pXZ(yAxis,TVector3D<A>(0.0f,0.0f,0.0f));
				CVector3gl Ip; pXZ.Intersect(*this,-yAxis,&Ip);
				*rotX = _copysign( Ip.isNull()?PI/2:GetAlpha(Ip) ,-pXZ.Solve(*this) );
			}
		inline TVector3D<A> Min(TVector3D<A> &min)
			{
				return TVector3D<A> 
					(	x<min.x?x:min.x,
						y<min.y?y:min.y,
						z<min.z?z:min.z	);
			}
		inline TVector3D<A> Max(TVector3D<A> &max)
			{
				return TVector3D<A> 
					(	x>max.x?x:max.x,
						y>max.y?y:max.y,
						z>max.z?z:max.z	);
			}
	};

	template <typename A> class TVector4D
	{
	public:
		A x,y,z,w;

		TVector4D() {};
		TVector4D(TVector3D<A> vec3d) : x(vec3d.x), y(vec3d.y), z(vec3d.z), w((A)1) {};
		TVector4D(A _x, A _y ,A _z) : x(_x), y(_y), z(_z), w((A)1) {};
		TVector4D(A _x, A _y ,A _z, A _w) : x(_x), y(_y), z(_z), w(_w) {};

		inline TVector4D<A> &Clear()
			{
				x=y=z=w=(A)0;
				return *this;
			}

		operator TVector3D<A> (void) { return TVector3D<A>(x,y,z); }

		TVector4D<A> & operator = (TVector4D<A> &obj) 
			{ 
				x=obj.x; y=obj.y; z=obj.z; w=obj.w;
				return *this;
			};
		TVector4D<A> & operator = (TVector3D<A> &obj)
			{ 
				x=obj.x; y=obj.y; z=obj.z; w=(A)1;
				return *this;
			};
		TVector4D<A> & operator = (const A obj[]) 
			{ 
				x=obj[0]; y=obj[1]; z=obj[2]; w=obj[3];
				return *this;
			};
		TVector4D<A> & operator += (TVector4D<A> &obj) 
			{ 
				x+=obj.x; y+=obj.y; z+=obj.z; w+=obj.w;
				return *this;
			};
		TVector4D<A> & operator -= (TVector4D<A> &obj) 
			{ 
				x-=obj.x; y-=obj.y; z-=obj.z; w-=obj.w;
				return *this;
			};
		TVector4D<A> & operator *= (A obj) 
			{ 
				x*=obj; y*=obj; z*=obj; w*=obj;
				return *this;
			};
		TVector4D<A> & operator /= (A obj) 
			{ 
				x/=obj; y/=obj; z/=obj; w/=obj;
				return *this;
			};

		TVector4D<A> operator - (void)
			{ 
				return TVector4D(-x, -y, -z, -w);
			};
		TVector4D<A> operator + (TVector4D<A> &obj)
			{ 
				return TVector4D(x+obj.x, y+obj.y, z+obj.z, w+obj.w);
			};
		TVector4D<A> operator - (TVector4D<A> &obj)
			{ 
				return TVector4D(x-obj.x, y-obj.y, z-obj.z, w-obj.w);
			};
		TVector4D<A> operator * (A obj)
			{ 
				return TVector4D(x*obj, y*obj, z*obj, w*obj);
			};
		TVector4D<A> operator * (TVector4D<A> obj)
			{ 
				return VectorProduct(obj);
			};
		TVector4D<A> operator / (A obj)
			{ 
				return TVector4D(x/obj, y/obj, z/obj, w/obj);
			};

		inline A abs2()
			{
				return x*x+y*y+z*z;
			}
		inline A abs()
			{
				return sqrt(abs2);
			};
		inline TVector4D<A> &Normal()
			{
				A len = abs();
				if (len) return (*this)/=len;
				else return (*this)=TVector4D<A>(0.0,0.0,0.0); 
			}

		inline A ScalarMultiply(TVector4D<A> &vec)
			{
				return
					x*vec.x+y*vec.y+z*vec.z+w*vec.w;
			}

		inline TVector4D<A> VectorProduct(TVector4D<A> &vec)
			{
				return Vector4D(
					y*vec.z-z*vec.y,
					z*vec.x-x*vec.z,
					x*vec.y-y*vec.x);
			}
		inline TVector4D<A> MatrixMultiply(TMatrix4D<A> &mat)
			{
				return Vector4D(
					x*mat.m[ 0] +	y*mat.m[ 4] +	z*mat.m[ 8] +	w*mat.m[12],
					x*mat.m[ 1] +	y*mat.m[ 5] +	z*mat.m[ 9] +	w*mat.m[13],
					x*mat.m[ 2] +	y*mat.m[ 6] +	z*mat.m[10] +	w*mat.m[14],
					x*mat.m[ 3] +	y*mat.m[ 7] +	z*mat.m[11] +	w*mat.m[15])
			}
		inline bool isNull()
			{
				return ((x!=(A)0) && (y!=(A)0) && (z!=(A)0))
			}
	};

	template <typename A> class TMatrix3D
	{
	public:
		A m[9];			// 0,1,2 col0 3,4,5 col1 6,7,8 col2
		enum {
			a11 = 0, a12 = 3, a13 = 6,
			a21 = 1, a22 = 4, a23 = 7,
			a31 = 2, a32 = 5, a33 = 8};
		enum {
			a1 = 0, a2 = 3, a3 = 4,
			b1 = 1, b2 = 4, b3 = 7,
			c1 = 2, c2 = 5, c3 = 8};

		TMatrix3D()
			{
				Clear ();
			}
		TMatrix3D(
			A m11,	A m12, A m13,
			A m21,	A m22, A m23,
			A m31,	A m32, A m33)
			{
				m[0] = m11; m[3] = m12; m[6] = m13;
				m[1] = m21; m[4] = m22; m[7] = m23;
				m[2] = m31; m[5] = m32; m[8] = m33;			
			}
		inline operator TMatrix4D<A> (void)
			{
				TMatrix4D<A> res;
				res.m[TMatrix4D<A>::a11] = m[a11];
				res.m[TMatrix4D<A>::a12] = m[a12];
				res.m[TMatrix4D<A>::a13] = m[a13];
				res.m[TMatrix4D<A>::a21] = m[a21];
				res.m[TMatrix4D<A>::a22] = m[a22];
				res.m[TMatrix4D<A>::a23] = m[a23];
				res.m[TMatrix4D<A>::a31] = m[a31];
				res.m[TMatrix4D<A>::a32] = m[a32];
				res.m[TMatrix4D<A>::a33] = m[a33];
				return res;
			}

		inline TMatrix3D &Clear()
			{
				m[0] = m[4] = m[8] = (A)1;
				m[1] = m[2] = m[3] = m[5] = m[6] = m[7]  = (A)0;

				return *this;
			}

		inline A Determinant()
		{
			return 
				+ m[a11] * m[a22] * m[a33]
				- m[a11] * m[a23] * m[a32]
				- m[a12] * m[a21] * m[a33]
				+ m[a12] * m[a23] * m[a31]
				+ m[a13] * m[a21] * m[a32]
				- m[a13] * m[a22] * m[a31];
		}
	};

	template <typename A> class TMatrix4D
	{
	public:
		A m[16];		// 0,1,2,3 col0 4,5,6,7 col1 ...

		enum {
			a11 =  0, a12 =  4, a13 =  8, a14 = 12,
			a21 =  1, a22 =  5, a23 =  9, a24 = 13,
			a31 =  2, a32 =  6, a33 = 10, a34 = 14,
			a41 =  3, a42 =  7, a43 = 11, a44 = 15};
		enum {
			a1 =  0, a2 =  4, a3 =  8, a4 = 12,
			b1 =  1, b2 =  5, b3 =  9, b4 = 13,
			c1 =  2, c2 =  6, c3 = 10, c4 = 14,
			d1 =  3, d2 =  7, d3 = 11, d4 = 15};

		TMatrix4D()
			{
				Clear();
			}
		TMatrix4D(
			A m11,	A m12,	A m13,	A m14,
			A m21,	A m22,	A m23,	A m24,
			A m31,	A m32,	A m33,	A m34,
			A m41,	A m42,	A m43,	A m44)
			{
				m[ 0] = m11;	m[ 4] = m12;	m[ 8] = m13;	m[12] = m14;	
				m[ 1] = m21;	m[ 5] = m22;	m[ 9] = m23;	m[13] = m24;	
				m[ 2] = m31;	m[ 6] = m32;	m[10] = m33;	m[14] = m34;	
				m[ 3] = m41;	m[ 7] = m42;	m[11] = m43;	m[15] = m44;	
			}
		TMatrix4D(A _m[])
			{
				memcpy(m,_m,sizeof(A)*16);
			}
		TMatrix4D(TVector3D<A> x,TVector3D<A> y,TVector3D<A> z,TVector3D<A> t)
			{
				m[ 0] = x.x;	m[ 4] = y.x;	m[ 8] = z.x;	m[12] = t.x;
				m[ 1] = x.y;	m[ 5] = y.y;	m[ 9] = z.y;	m[13] = t.y;
				m[ 2] = x.z;	m[ 6] = y.z;	m[10] = z.z;	m[14] = t.z;
				m[ 3] = (A)0;	m[ 7] = (A)0;	m[11] = (A)0;	m[15] = (A)1;
			}
		inline TMatrix4D &Clear()
			{
				m[ 0] =	m[ 5] =	m[10] = m[15] = (A)1;

						m[ 4] =	m[ 8] =	m[12] = 
				m[ 1] =			m[ 9] =	m[13] =
				m[ 2] =	m[ 6] =			m[14] =
				m[ 3] =	m[ 7] = m[11] =			(A)0;
				return *this;
			}

		inline A &operator [] (int index)
			{
				return m[index];
			}
		inline A &operator [] (char *rowcol)
			{
				return m[(rowcol[0]-'0')+4*(rowcol[1]-'0')];
			}

		inline TMatrix4D<A> &operator = (TMatrix4D<A> &obj)
			{
				memcpy(m,obj.m,sizeof(A)*16);
				return *this;
			}
		inline TMatrix4D<A> &operator = (A obj[])
			{
				memcpy(m,obj,sizeof(A)*16);
				return *this;
			}

		inline TMatrix4D<A> &operator *= (TMatrix4D<A> &obj)
			{
				return (*this=Multiply(obj.m));
			}
		inline TMatrix4D<A> &operator *= (A obj[])
			{
				return (*this=Multiply(obj));
			}

		inline TMatrix4D<A> operator * (TMatrix4D<A> &obj)
			{
				return Multiply(obj.m);
			}
		inline TMatrix4D<A> operator * (A obj[])
			{
				return Multiply(obj);
			}
		inline TVector4D<A> operator * (TVector4D<A> &vec)
			{
				return Multiply(vec);
			}
		inline TVector3D<A> operator * (TVector3D<A> &vec)
			{
				return Multiply(vec);
			}

		inline TMatrix4D<A> Multiply(TMatrix4D<A> &obj)
			{
				return Multiply(obj.m);
			}
		inline TMatrix4D<A> Multiply(A obj[])
			{
				int c,r,n;
				TMatrix4D<A> res;
				for (c=0; c<4; c++)
					for (r=0; r<4; r++)
					{
						res.m[r+4*c]=0;
						for (n=0;n<4;n++)
							res.m[r+4*c]+=m[r+4*n]*obj[n+4*c];
					}
				return res;
	/*			return TMatrix4D<A>(
				// Line 0
					m[ 0]*obj[ 0] +	m[ 4]*obj[ 1] +	m[ 8]*obj[ 2] +	m[12]*obj[ 3],
					m[ 0]*obj[ 4] +	m[ 4]*obj[ 5] +	m[ 8]*obj[ 6] +	m[12]*obj[ 7],
					m[ 0]*obj[ 8] +	m[ 4]*obj[ 9] +	m[ 8]*obj[10] +	m[12]*obj[11],
					m[ 0]*obj[12] +	m[ 4]*obj[13] +	m[ 8]*obj[14] +	m[12]*obj[15],
				// Line 1
					m[ 1]*obj[ 0] +	m[ 5]*obj[ 1] +	m[ 9]*obj[ 2] +	m[13]*obj[ 3],
					m[ 1]*obj[ 4] +	m[ 5]*obj[ 5] +	m[ 9]*obj[ 6] +	m[13]*obj[ 7],
					m[ 1]*obj[ 8] +	m[ 5]*obj[ 9] +	m[ 9]*obj[10] +	m[13]*obj[11],
					m[ 1]*obj[12] +	m[ 5]*obj[13] +	m[ 9]*obj[14] +	m[13]*obj[15],
				// Line 2
					m[ 2]*obj[ 0] +	m[ 6]*obj[ 1] +	m[10]*obj[ 2] +	m[14]*obj[ 3],
					m[ 2]*obj[ 4] +	m[ 6]*obj[ 5] +	m[10]*obj[ 6] +	m[14]*obj[ 7],
					m[ 2]*obj[ 8] +	m[ 6]*obj[ 9] +	m[10]*obj[10] +	m[14]*obj[11],
					m[ 2]*obj[12] +	m[ 6]*obj[13] +	m[10]*obj[14] +	m[14]*obj[15],
				// Line 3
					m[ 3]*obj[ 0] +	m[ 7]*obj[ 1] +	m[11]*obj[ 2] +	m[15]*obj[ 3],
					m[ 3]*obj[ 4] +	m[ 7]*obj[ 5] +	m[11]*obj[ 6] +	m[15]*obj[ 7],
					m[ 3]*obj[ 8] +	m[ 7]*obj[ 9] +	m[11]*obj[10] +	m[15]*obj[11],
					m[ 3]*obj[12] +	m[ 7]*obj[13] +	m[11]*obj[14] +	m[15]*obj[15]
					);*/
			}
		inline TVector4D<A> Multiply(TVector4D<A> &obj)
			{
				return TVector4D<A>(
					obj.x*m[ 0] +	obj.y*m[ 4] +	obj.z*m[ 8] +	obj.w*m[12],
					obj.x*m[ 1] +	obj.y*m[ 5] +	obj.z*m[ 9] +	obj.w*m[13],
					obj.x*m[ 2] +	obj.y*m[ 6] +	obj.z*m[10] +	obj.w*m[14],
					obj.x*m[ 3] +	obj.y*m[ 7] +	obj.z*m[11] +	obj.w*m[15]);
			}
		inline TVector3D<A> Multiply(TVector3D<A> &obj)
			{
				return TVector3D<A>(
					obj.x*m[ 0] +	obj.y*m[ 4] +	obj.z*m[ 8] +	((A)1)*m[12],
					obj.x*m[ 1] +	obj.y*m[ 5] +	obj.z*m[ 9] +	((A)1)*m[13],
					obj.x*m[ 2] +	obj.y*m[ 6] +	obj.z*m[10] +	((A)1)*m[14]);
			}

		inline TMatrix4D<A> &RotateX(double angle)
			{
				TMatrix4D<A> old(*this);
				A s=(A)sin(-angle);
				A c=(A)cos(-angle);
				m[ 4]=old.m[ 4]*c-old.m[ 8]*s;	m[ 8]=old.m[ 4]*s+old.m[ 8]*c;
				m[ 5]=old.m[ 5]*c-old.m[ 9]*s;	m[ 9]=old.m[ 5]*s+old.m[ 9]*c;
				m[ 6]=old.m[ 6]*c-old.m[10]*s;	m[10]=old.m[ 6]*s+old.m[10]*c;
				m[ 7]=old.m[ 7]*c-old.m[11]*s;	m[11]=old.m[ 7]*s+old.m[11]*c;
				return *this;
			}
		inline TMatrix4D<A> &RotateY(double angle)
			{
				TMatrix4D<A> old(*this);
				A s=(A)sin(-angle);
				A c=(A)cos(-angle);
				m[ 0]=old.m[ 0]*c+old.m[ 8]*s;	m[ 8]=old.m[ 8]*c-old.m[ 0]*s;
				m[ 1]=old.m[ 1]*c+old.m[ 9]*s;	m[ 9]=old.m[ 9]*c-old.m[ 1]*s;
				m[ 2]=old.m[ 2]*c+old.m[10]*s;	m[10]=old.m[10]*c-old.m[ 2]*s;
				m[ 3]=old.m[ 3]*c+old.m[11]*s;	m[11]=old.m[11]*c-old.m[ 3]*s;
				return *this;
			}
		inline TMatrix4D<A> &RotateZ(double angle)
			{
				TMatrix4D<A> old(*this);
				A s=(A)sin(-angle);
				A c=(A)cos(-angle);
				m[ 0]=old.m[ 0]*c-old.m[ 4]*s;	m[ 4]=old.m[ 0]*s+old.m[ 4]*c;
				m[ 1]=old.m[ 1]*c-old.m[ 5]*s;	m[ 5]=old.m[ 1]*s+old.m[ 5]*c;
				m[ 2]=old.m[ 2]*c-old.m[ 6]*s;	m[ 6]=old.m[ 2]*s+old.m[ 6]*c;
				m[ 3]=old.m[ 3]*c-old.m[ 7]*s;	m[ 7]=old.m[ 3]*s+old.m[ 7]*c;
				return *this;
			}

		inline bool Invert()
			{
				TMatrix4D<A> mat;
				TMatrix3D<A> m3;
				A det = Derterminant();
				if (det<EPSILON) return false;
				int i,n,k,j,l;

				for (i=0;i<4;i++)
				{
					for (n=0;n<4;n++)
					{
						l=0;
						for (k=0; k<4; k++)
						{
							if (k==i) continue;
							for (j=0; j<4; j++)
							{
								if (j==n) continue;
								m3.m[l] = m[j+4*k];
								l++;
							}
						}

						mat.m[i+4*n] = ((i+n)%2==0?(A)1:(A)-1) * m3.Determinant() / det;
					}
				}

				*this = mat;
				return true;
			}

		inline SCALAR Derterminant()
			{
				return 
					m[a11] * m[a22] * m[a33] * m[a44]
					- m[a11] * m[a22] * m[a34] * m[a43]
					- m[a11] * m[a23] * m[a32] * m[a44]
					+ m[a11] * m[a23] * m[a34] * m[a42]
					+ m[a11] * m[a24] * m[a32] * m[a43]
					- m[a11] * m[a24] * m[a33] * m[a42]
					- m[a12] * m[a21] * m[a33] * m[a44]
					+ m[a12] * m[a21] * m[a34] * m[a43]
					+ m[a12] * m[a23] * m[a31] * m[a44]
					- m[a12] * m[a23] * m[a34] * m[a41]
					- m[a12] * m[a24] * m[a31] * m[a43]
					+ m[a12] * m[a24] * m[a33] * m[a41]
					- m[a13] * m[a22] * m[a31] * m[a44]
					+ m[a13] * m[a22] * m[a34] * m[a41]
					+ m[a13] * m[a21] * m[a32] * m[a44]
					- m[a13] * m[a21] * m[a34] * m[a42]
					- m[a13] * m[a24] * m[a32] * m[a41]
					+ m[a13] * m[a24] * m[a31] * m[a42]
					- m[a14] * m[a22] * m[a33] * m[a41]
					+ m[a14] * m[a22] * m[a31] * m[a43]
					+ m[a14] * m[a23] * m[a32] * m[a41]
					- m[a14] * m[a23] * m[a31] * m[a42]
					- m[a14] * m[a21] * m[a32] * m[a43]
					+ m[a14] * m[a21] * m[a33] * m[a42];
			}

		inline TMatrix4D<A> &InvertGauss()
			{
				TMatrix4D<A> res;
				float di,mu;
				int i,j,k,l;

				for (k=0; k<4; k++)
				{
					di = m[5*k];
					for (i=0; i<4; i++)
					{
						res[k+4*i] /= di;
						m[k+4*i]   /= di;
					}
					for (i=k+1; i<4; i++)
					{
						mu = m[i+k*4];
						for (j=0; j<4; j++)
						{
							res[i+j*4] -= mu*res[k+j*4];
							m[i+j*4]   -= mu*m[k+j*4];
						}
					}
				}
				for (k=1; k<4; k++)
				{
					for (i=0; i<k; i++)
					{
						mu = m[i+4*k];
						for (j=0; j<4; j++)
						{
							res[i+4*j] -= mu*res[k+4*j];
							m[i+4*j]   -= mu*m[k+4*j];
						}
					}
				}

				
				

				memcpy(m,res.m,16*sizeof(A));
				return *this;
			}
	/*	inline TMatrix4D<A> &Invert()
		{
				TMatrix4D<A> old(*this);
				TMatrix4D<A> dummy(*this);
				A			 d=0,q=0;

				int i,j,k,n;

				//A s=(A)sin(-angle);
				//A c=(A)cos(-angle);
				//m[ 0]=old.m[ 0]*c-old.m[ 4]*s;	m[ 4]=old.m[ 0]*s+old.m[ 4]*c;
				//m[ 1]=old.m[ 1]*c-old.m[ 5]*s;	m[ 5]=old.m[ 1]*s+old.m[ 5]*c;
				//m[ 2]=old.m[ 2]*c-old.m[ 6]*s;	m[ 6]=old.m[ 2]*s+old.m[ 6]*c;
				//m[ 3]=old.m[ 3]*c-old.m[ 7]*s;	m[ 7]=old.m[ 3]*s+old.m[ 7]*c;

				// A original matrix (old)
				// B inverse matrix  (m)
				// C umrechnung		 (dummy);
				// n nxn Matrix		 (4);

				for (i=0; i<4;i++)
				{
					for (j=0; j<4; j++)
					{
						if (i==j) m[i+(j*4)]=1;
						else m[i+(j*4)]=0;
					}
				}
				if (old[0]==0)
				{
					for (j=0; j<4; j++)
					{
				

						old.m[0] = 0; 

						dummy.m[0+j*4]=  old.m[1+j*4];
						old.m[1+j*4]=  old.m[0+j*4];
						old.m[0+j*4]=dummy.m[0+j*4];

						dummy.m[0+j*4]=      m[1+j*4];
							m[1+j*4]=      m[0+j*4];
							m[0+j*4]=dummy.m[1+j*4];
					}
				}
				for (i=0; i<4; i++)
				{
					d = (A)old.m[i+4*i];
					for (j=0; j<4; j++)
					{
						if (d!=(A)0)
						{
							old.m[i+4*j]=(A) (old.m[i+4*j]/(A)d);
							m[i+4*j]=(A) (m[i+4*j]/(A)d);
						}
					}
					for (k=0; k<4; k++)
					{
						if (k!=i)
						{
							q = (A)old.m[k+i*4];
							for (j=0; j<4; j++)
							{
								old.m[k+j*4]=(A)old.m[k+j*4]-((A)old.m[i+j*4]*q);
								m    [k+j*4]=(A)    m[k+j*4]-((A)    m[i+j*4]*q);
							}
						}
					}

				};
				return *this;
			}*/



		inline TMatrix4D<A> &SetRotate(double angle, TVector3D<A> axisNormalized)
			{
				float s = (float)GENERAL::Sin(angle);
				float c = (float)GENERAL::Cos(angle);

				m[ 0]=axisNormalized.x*axisNormalized.x*(1-c)+c;					
				m[ 4]=axisNormalized.x*axisNormalized.y*(1-c)-axisNormalized.z*s;	
				m[ 8]=axisNormalized.x*axisNormalized.z*(1-c)+axisNormalized.y*s;

				m[ 1]=axisNormalized.x*axisNormalized.y*(1-c)+axisNormalized.z*s;	
				m[ 5]=axisNormalized.y*axisNormalized.y*(1-c)+c;					
				m[ 9]=axisNormalized.y*axisNormalized.z*(1-c)-axisNormalized.x*s;

				m[ 2]=axisNormalized.x*axisNormalized.z*(1-c)-axisNormalized.y*s;	
				m[ 6]=axisNormalized.y*axisNormalized.z*(1-c)+axisNormalized.x*s;	
				m[10]=axisNormalized.z*axisNormalized.z*(1-c)+c; 

				m[ 3] = 
				m[ 7] = 
				m[11] = 
				m[12] = 
				m[13] = 
				m[14] = 0.0f;
				m[15] = 1.0f;

				return *this;
			}
		inline TMatrix4D<A> &SetRotateEuler(A x, A y, A z)
			{
				A cx = cos(x), sx = sin(x);
				A cy = cos(y), sy = sin(y);
				A cz = cos(z), sz = sin(z);

				m[ 0] = cz*cx-cy*sx*sy;
				m[ 1] = cz*sx+cy*cx*sy;
				m[ 2] = sz*sy;

				m[ 4] = -sz*cx-cy*sx*cz;
				m[ 5] = -sz*sx+cy*cx*cz;
				m[ 6] = cz*sy;

				m[ 8] = sy*sx;
				m[ 9] = -sy*cx;
				m[10] = cy;

				m[ 3] = m[ 7] = m[11] = m[12] = m[13] = m[14] = (A)0;
				m[15] = (A)1;

				return *this;
			}
		inline TMatrix4D<A> &SetRotateYawPitchRoll(A x, A y, A z)
			{
				A cx = cos(x), sx = sin(x);
				A cy = cos(y), sy = sin(y);
				A cz = cos(z), sz = sin(z);

				m[ 0] = cy*cz;
				m[ 1] = -cz*sx*sy+cx*sz;
				m[ 2] = cx*cz*sy+sx*sz;

				m[ 4] = -cy*sz;
				m[ 5] = cx*cz+sx*sy*sz;
				m[ 6] = cz*sx-cx*sy*sz;

				m[ 8] = -sy;
				m[ 9] = -cy*sx;
				m[10] = cx*cy;
/*				m[ 0] = cy*cx;
				m[ 1] = cy*sx;
				m[ 2] = -sy;

				m[ 4] = sz*sy*cx-cz*sx;
				m[ 5] = sz*sy*sx+cz*cx;
				m[ 6] = cy*cz;

				m[ 8] = cz*sy*cx+sz*sx;
				m[ 9] = cz*sy*sx-sz*cx;
				m[10] = cy*cz;*/

				m[ 3] = m[ 7] = m[11] = m[12] = m[13] = m[14] = (A)0;
				m[15] = (A)1;

				return *this;
			}

		inline TMatrix4D<A> &Translate(TVector3D<A> vec)
			{
				return Translate(vec.x,vec.y,vec.z);
			}
		inline TMatrix4D<A> &Translate(TVector4D<A> vec)
			{
				return Translate(vec.x,vec.y,vec.z);
			}
		inline TMatrix4D<A> &Translate(A x, A y, A z)
			{
				m[12]+=m[ 0]*x+m[ 4]*y+m[ 8]*z;
				m[13]+=m[ 1]*x+m[ 5]*y+m[ 9]*z;
				m[14]+=m[ 2]*x+m[ 6]*y+m[10]*z;
				m[15]+=m[ 3]*x+m[ 7]*y+m[11]*z;
				return *this;
			}
		inline TMatrix4D<A> &Scale(A x, A y, A z, A w = (A)1)
			{
				m[ 0] *= x;	m[ 4] *= y;	m[ 8] *= z;	m[12] *= w;
				m[ 1] *= x;	m[ 5] *= y;	m[ 9] *= z;	m[13] *= w;
				m[ 2] *= x;	m[ 6] *= y;	m[10] *= z;	m[14] *= w;
				m[ 3] *= x;	m[ 7] *= y;	m[11] *= z;	m[15] *= w;
				return *this;
			}
		inline TMatrix4D<A> &SetProjectionInf(SCALAR fovy,SCALAR aspect, SCALAR zNear)
			{
				SCALAR zFar = 1000.0f;
				Clear();
				SCALAR bottom = -(zNear*cos(fovy));
				SCALAR top = -bottom;
				SCALAR left = bottom * aspect;
				SCALAR right = top * aspect;

				m[a11] = (2.0f*zNear)/(right-left);
				m[a13] = (right+left)/(right-left);
				m[a22] = (2.0f*zNear)/(top-bottom);
				m[a23] = (top+bottom)/(top-bottom);
	/*			m[a33] = -(zFar+zNear)/(zFar-zNear);
				m[a34] = -(2.0f*zFar*zNear)/(zFar-zNear);*/
				m[a33] = -1;
				m[a34] = -2.0f*zNear;

				m[a43] = -1.0f;
				m[a44] = 0;
	//			glFrustum
				return *this;
			}
		inline TMatrix4D<A> &Transpose()
			{
				Transpose_M44f(m,m);
				return *this;
			}
	};

	template <typename A> class TLine
	{
	public:
		TVector3D<A> p, dir;

		TLine() {};
		TLine(TVector3D<A> _p, TVector3D<A> _dir) : p(_p), dir(_dir) {}

		TLine<A> &operator = (TLine<A> &obj)
			{
				p = obj.p;
				dir = obj.dir;
				retrun *this;
			}
	};

	template <typename A> class TPlane
	{
	public:
		TVector3D<A> n;
		A D;

		TPlane() {};
		TPlane(TPlane<A> &_p) 
			{
				n = _p.n;
				D = _p.D;
			}
		TPlane(TVector3D<A> &_n, A _D) 
			{
				A len=_n.abs();
				n = _n / len;
				D = _D / len;
				
			};
		TPlane(TVector3D<A> &_n, TVector3D<A> &_p)
			{
				n = _n;
				n.Normal();
				D= -(n.x*_p.x+n.y*_p.y+n.z*_p.z);
			};
		TPlane(TVector3D<A> p1, TVector3D<A> p2,TVector3D<A> p3)
			{
				n = (p1-p3)*(p2-p1);
				n.Normal();
				D= -(n.x*p1.x+n.y*p1.y+n.z*p1.z);
			};
		inline void Normalize()
			{
				A len = n.abs();
				D/=len;
				n/=len;
			}

		inline bool Visible(TVector3D<A> &p)
			{
				return Solve(p)>0;
			};

		inline A Solve(TVector3D<A> &_p)
			{
				return n.x*_p.x+n.y*_p.y+n.z*_p.z+D;
			};
		inline bool Intersect(TLine<A> &line, TVector3D<A> *res)
			{
				return Intersect(line.p,line.dir,res);
			}
		inline bool Intersect(TVector3D<A> &p,TVector3D<A> &dir, TVector3D<A> *res)
			{
				A div = (n.x*dir.x+n.y*dir.y+n.z*dir.z);
				if (div==(A)0.0) return FALSE;
				*res =  p-dir*(Solve(p)/div);
				return TRUE;
			}
		inline A GetAlpha(TPlane<A> &p)
			{
				A res = n.ScalarMultiply(p.n);
				res = res>(A)1?(A)1: (res<(A)-1?(A)-1:res);
				return (A)acos(res);
			}
	};

	template <typename A,typename TYP> class TQuadricBezierPatch
	{
	public:
		TYP controlPoints[3];

		inline TQuadricBezierPatch() {};
		inline void Tesselate(TYP &out, A t)
			{
				register A t1 = (1-t);
				register A b0 = t1*t1;
				register A b1 = 2*t*t1;
				register A b2 = t*t;	

				out = controlPoints[0]*b0+controlPoints[1]*b1+controlPoints[2]*b2;
			}
	};

	template <typename A,typename TYP> class TBiQuadricBezierPatch
	{
	public:
		TQuadricBezierPatch<A,TYP> upatch[3];

		inline TBiQuadricBezierPatch() {};

		inline void Tesselate(TYP &out, A u, A v)
			{
				TQuadricBezierPatch<A,TYP> ptemp;
				upatch[0].Tesselate(ptemp.controlPoints[0],u);
				upatch[1].Tesselate(ptemp.controlPoints[1],u);
				upatch[2].Tesselate(ptemp.controlPoints[2],u);

				ptemp.Tesselate(out,v);
			}
		inline void TesselateUpatch(TQuadricBezierPatch <A,TYP> &out, A u)
			{
				upatch[0].Tesselate(out.controlPoints[0],u);
				upatch[1].Tesselate(out.controlPoints[1],u);
				upatch[2].Tesselate(out.controlPoints[2],u);
			}
	};

	template <typename A,typename TYP> class TCubicBezierPatch	{
	public:
		TYP controlPoints[4];
		
		inline TCubicBezierPatch() {};
		inline void Tesselate(TYP &out, A t)
		{	
			register A t1 = 1-t;
			register A b0 = t1*t1*t1;
			register A b1 = 3*t*t1*t1;
			register A b2 = 3*t*t*t1;
			register A b3 = t*t*t;

			out =	controlPoints[0]*b0+
					controlPoints[1]*b1+
					controlPoints[2]*b2+
					controlPoints[3]*b3;
		}

	};

	inline float MMfabs(float _val)
	{
		if (_val>=0) return _val;
		else return -_val;
	}



	class CQuaternion 
	{
	private:
		float x, y, z, w;
	public:
		CQuaternion() { x = y = z = 0.0f;   w = 1.0f;  }
		CQuaternion(float X, float Y, float Z, float W) 
		{ 
			x = X;		y = Y;		z = Z;		w = W;
		}
		void CreateMatrix(float *pMatrix);
		void CreateFromMatrix(float *pMatrix, int rowColumnCount);
		CQuaternion Slerp(CQuaternion &q1, CQuaternion &q2, float t);
	};

	







}

#endif 

