#include "morftools.h"
#include "../../morfbase.h"

#include <stdarg.h>
#include <stdio.h>
#include <wtypes.h>
#include <deque>
#include <string>

#include "config.h"

#include "trace.h"


namespace GENERAL
{
	#ifdef DEBUG_TO_CONSOLE
	std::deque<std::string> debugconsole;

	class TConsoleMutex
	{
	private:
		HANDLE h;
	public:
		TConsoleMutex()
		{
			if ((h = CreateMutex(NULL,FALSE,NULL))==NULL)
			{
				throw exception("Cannot create trace synchronization object");
			}
		};
		virtual ~TConsoleMutex()
		{
			CloseHandle(h);
		};
		inline void Lock()
		{
			if (WaitForSingleObject(h,2000)!=WAIT_OBJECT_0)
				throw exception("Could not lock console synchronization object");
		}

		inline void Unlock()
		{
			if (ReleaseMutex(h)==0)
				throw exception("Could not unlock console synchronization object");
		}


	};
	TConsoleMutex *mutex=NULL;// = new TConsoleMutex();
#endif
		
	int dprintf(const char * format, ...)
	{
		int size;
		char *buffer;

		va_list args;
	    va_start(args, format);
			size = _vscprintf(format,args);
			buffer = (char*)malloc(size+1);
			vsprintf(buffer,format,args);
		va_end(args);

		#ifdef _DEBUG
		OutputDebugString(buffer);
		#endif

		#ifdef DEBUG_TO_FILE
		static FILE *f = NULL;
		
		if (f==NULL) f=fopen(DEBUG_TO_FILE,"w");
		if (f!=NULL)
		{
			fprintf(f,"%s",buffer);
			fflush(f);
		}
		#endif


		#ifdef DEBUG_TO_CONSOLE
		if (!mutex)
			mutex=new TConsoleMutex();
		mutex->Lock();
		debugconsole.push_back(buffer);
		mutex->Unlock();
		#endif

		free(buffer);

		return size;
	}

	FILE *morfOpenFile(char *filename,char *mode)
	{
		FILE *f=NULL;
		std::string fname;

		CONFIG::CConfig::TStringList::iterator it;

		for (it = CONFIG::g_config.ResPath.begin(); it != CONFIG::g_config.ResPath.end(); it++)
		{
			fname = *it;
			fname.append(filename);

			f = fopen(fname.data(),mode);

			if (f!=NULL)
			{
/*				FILE *fp=fopen("filesused.txt","a");
				fprintf(fp,"%s\n",fname.data());
				fclose(fp);*/
				return f;
			}
		}

		return NULL;
	}

	char *morfFindFile(char *filename)
	{
		static char result[1024];
		result[0]='\0';
		FILE *f;
		std::string fname;

		CONFIG::CConfig::TStringList::iterator it;

		for (it = CONFIG::g_config.ResPath.begin(); it != CONFIG::g_config.ResPath.end(); it++)
		{
			fname = *it;
			fname.append(filename);

			f = fopen(fname.data(),"r");

			if (f!=NULL)
			{
				fclose(f);
				sprintf(result,fname.data());
				return result;
			}
		}

		return 0;
	}


	bool IsInString(std::string strString, std::string strSubString)
	{
		if(strString.length() <= 0 || strSubString.length() <= 0) return false;

		unsigned int index = (unsigned int) strString.find(strSubString);

		if(index >= 0 && index < strString.length())
			return true;

		return false;
	}
}
