#include "pyscripting.h"
#include "morftools.h"
#include "morftypes.h"
#include "morfmath.h"
#include "../../morfbase.h"


namespace GENERAL
{
	struct TPyModuleInfo
	{
		PyObject *dict;
		PyObject *module;

		TPyModuleInfo() : dict(NULL),module(NULL) {};
		TPyModuleInfo(const TPyModuleInfo &mi) : dict(mi.dict), module(mi.module) {};
	};
	std::map< std::string, TPyModuleInfo> modules;

/*****************************************************************************************
PyGetError
------------------------------------------------------------------------------------------
What it does: 	Check for Python Errors and Print it
------------------------------------------------------------------------------------------
Parameter:		none
Return Code:	none
------------------------------------------------------------------------------------------
History:		
Date			Editor				Comment
19.05.05		ME					testing
------------------------------------------------------------------------------------------
ToDo:			PyErr_Print prints to stderr dprintf should be called
*****************************************************************************************/
	void PyGetError()
	{
		PyObject *err = PyErr_Occurred();
		if (err!=NULL)
		{
			PyErr_Print();
		}
	}
/* END: PyGetError
/****************************************************************************************/


/*****************************************************************************************
LoadModule
------------------------------------------------------------------------------------------
What it does: 	Loads a Module and stores into the local Modulemap
------------------------------------------------------------------------------------------
Parameter:		muduleName			char*			name of the module to load
Return Code:	bool								true if no error false if module not loaded
------------------------------------------------------------------------------------------
History:		
Date			Editor				Comment
19.05.05		ME					final
------------------------------------------------------------------------------------------
ToDo:			
*****************************************************************************************/
	bool LoadModule(char *moduleName)
	{
		TPyModuleInfo mi;
		PyObject *pModuleName = PyString_FromString(moduleName);
		mi.module=PyImport_Import(pModuleName);
		Py_DECREF(pModuleName);

		if (mi.module)
		{
			mi.dict = PyModule_GetDict(mi.module);
			modules[moduleName] = mi;
			return true;
		}

		PyGetError();
		return false;
	}
/* END: LoadModule
/****************************************************************************************/

	TPyModuleInfo *GetPyModule(char *moduleName)
	{
		std::map< std::string, TPyModuleInfo>::iterator it;
		it = modules.find(moduleName);
		if (it==modules.end())
		{
			if (!LoadModule(moduleName))
				return NULL;

			it = modules.find(moduleName);
		}

		return &(it->second);
	}

	bool InitPython()
	{
		Py_Initialize();
		return true;
	}

	void DonePython()
	{
		modules.clear();
		Py_Finalize();
	}


	PyObject *CallScript(char *module, char *name, char *params, ...)
	{
		std::map< std::string, TPyModuleInfo>::iterator it;
		it = modules.find(module);
		if (it==modules.end())
		{
			if (!LoadModule(module))
			{
				return NULL;
			}

			it = modules.find(module);
		}

		TPyModuleInfo &mi=it->second;


/*		int i = PyDict_Size(mi.dict);
		PyObject *pKeys = PyDict_Keys(mi.dict);

		while (i--)
		{
			PyObject *item = PyList_GetItem(pKeys,i);

			dprintf("%s\n",PyString_AsString(item));

			Py_DECREF(item);
		}
		Py_DECREF(pKeys);*/

		PyObject *pFunc = PyDict_GetItemString(mi.dict, name);
		if (pFunc && PyCallable_Check(pFunc)) 
		{
			PyObject *pArgs = PyTuple_New((int)strlen(params));

			va_list args;
			va_start(args, params);


			char *c;
			int i=0;
			for (c=params; (*c)!='\0'; c++, i++)
			{
				PyObject *param = NULL;

				switch (*c)
				{
				case 'g':
					{
						DOMAIN::CGameObject *go = (va_arg(args,DOMAIN::CGameObject*));
						param = PyCObject_FromVoidPtr(go,NULL);
					}
					break;
				case 'v':
					{
						CVector3f &vec = *(va_arg(args,CVector3f*));
						param = Vector3fToPython(vec);
					}
					break;
				case 'f':
					{
						float f = (float)va_arg(args,double);
						param = PyFloat_FromDouble(f);
					}
					break;
				case 'i':
					param = PyInt_FromLong(va_arg(args,int));
					break;
				case 's':
					param = PyString_FromString(va_arg(args,const char*));
					break;
				}

				if (param!=NULL)
				{
					 PyTuple_SetItem(pArgs, i, param);
				}
			}

			va_end(args);

		
			PyObject *pValue = PyObject_CallObject(pFunc, pArgs);
			if (pValue)
			{
				Py_DECREF(pValue);
			} else {
				PyErr_Print();
			}
			Py_DECREF(pArgs);
			return pValue;
		} else 
			return NULL;
	}

	PyObject *Vector3fToPython(CVector3f &vec)
	{
		TPyModuleInfo *mi = GetPyModule("testmodule");
		PyObject *pmcd = PyDict_GetItemString(mi->dict, "CVector3f");
		PyObject *pmc = PyInstance_NewRaw(pmcd,NULL);
		PyObject_SetAttrString(pmc,"x",PyFloat_FromDouble((double)vec.x));
		PyObject_SetAttrString(pmc,"y",PyFloat_FromDouble((double)vec.y));
		PyObject_SetAttrString(pmc,"z",PyFloat_FromDouble((double)vec.z));
		return pmc;
	}

	CVector3f PythonToVector3f(PyObject *vec)
	{
		CVector3f res;
		res.x=(float)PyFloat_AS_DOUBLE(PyObject_GetAttrString(vec,"x"));
		res.y=(float)PyFloat_AS_DOUBLE(PyObject_GetAttrString(vec,"y"));
		res.z=(float)PyFloat_AS_DOUBLE(PyObject_GetAttrString(vec,"z"));
		return res;
	}
}