/** \file serialize.cpp
*	\brief implementations of CSerializeableObject, CMemBuffer and the global functions related to serializing.  
*/

//#include <winbase>
#include "serialize.h"
#include "../../network/base/clientinfo.h"


#include "trace.h"

/*TClientInfoList *NETWORK::TClientInfoList::GetRef() 
{
	refCnt++; 
	return this;
};
void ReleaseRef() 
{
	if (refCnt==1) delete this; 
	refCnt--;
};*/

#ifdef _DEBUG
	GENERAL::CException::~CException() 
	{ 
		free(_what); 
	}
	const char *GENERAL::CException::what()
	{
		return _what;
	}
#endif

namespace GENERAL { IMPLEMENT_SERIALIZEABLE(CSerializeableObject) };

GENERAL::CSerializeableObject::~CSerializeableObject()
{
	if (sendToClients) sendToClients->ReleaseRef();
}

/*void GENERAL::CSerializeableObject::Destroy(CSerializeableObject *so) 
{
	ref--; 
	if (ref==0) 
		delete so;
}*/

/** Abstract constructor of the class CSerializeableObject.
*/
GENERAL::CSerializeableObject::CSerializeableObject() : sendToClients(NULL)
{
	key=GET_KEY(CSerializeableObject);
}
/** Return the size of the Object in serialized form.
*
*/
long GENERAL::CSerializeableObject::GetSerializedSize() 
{ 
	return sizeof(key); 
}

/** Serialize an Object.
*
*	\Note Overwrite this method for your own serializeable Objects!
*
*	To serialize an object do:
*
*	-# serialize the key 				   
*	-# call BaseObject::serialize().
*	-# write your own data to the buffer.
*	-# return CBuffer::GetSize();
*
*	\param buffer referenze to the CBuffer to serialize to.
*/
long GENERAL::CSerializeableObject::Serialize(GENERAL::CBuffer &buffer) 
{ 
	buffer.Write(&key,sizeof(key));
	return buffer.GetPos(); 
}

/** Retrieve the data from a buffer.
*
*	\Note Overwrite this method for your own serializeable Objects!
*
*	To deserialize an object do:
*
*	-# deserialize the key key = buffer.read(sizeof(key));
*	-# call BaseObject::Deserialize();
*	-# read your own data.
*	-# return buffer.GetSize();
*
*	\param buffer referenze to the buffer to deserialize from.
*/
long GENERAL::CSerializeableObject::DeSerialize(GENERAL::CBuffer &buffer) 
{ 
	key=*(long*)buffer.Read(sizeof(key));
	return buffer.GetPos(); 
}

/** Print debug information to a file
*
*	\param f file to print to.
*	\param newline append a linefeed 
*/
void GENERAL::CSerializeableObject::Print(FILE *f,bool newline) 
{
	fprintf(stdout,"%d:%s, ",key,GetClassName());
	if (newline) fprintf(f,"\n");
}

/** declare the global class map: class -> key.
*
*	This map is used to find the key of a registered class.
*/
GENERAL::TSerializeableClassList serializeableClassList;

/** declare the global class map: key -> class.
*
*	This map is used to find the corresponding class to a given key.
*/
GENERAL::TSerializeableClassListKeys serializeableClassListKeys;
GENERAL::TSerializeableClassListNames serializeableClassListNames;

GENERAL::TClassHirarchyTree classHirarchyTree;
GENERAL::TSerializeableClassKeyHirarchy serializeableClassKeyHirarchy;

/** register a class to the global class map
*
*	Creates a pair of the \a classcreator function of class and a userdefined \a key for that class 
*
*	\note Classes are always referenced by their creator functions.
*
*	\param classcreator the funtion that creates on object of this class
*	\param key the key to register for the class
*/
void GENERAL::RegisterSerializeableClass(char *classname, GENERAL::TClassCreator classcreator,long key)
{
	if (serializeableClassList.find(key)!=serializeableClassList.end())
	{
		throw exception("Key already registered!");
	}
	serializeableClassList[key]=classcreator;
	serializeableClassListKeys[classcreator]=key;
	serializeableClassListNames[key]=classname;
}

void GENERAL::RegisterSerializeableClassHirarchy(long superkey,long inheritedkey)
{
	GENERAL::TClassHirarchyTree::iterator it;
//GENERAL::TClassHirarchyTree classHirachyTree;
//GENERAL::TSerializeableClassKeyHirarchy serializeableClassKeyHirarchy;
	if (superkey==0)
	{
		it = classHirarchyTree.insert(classHirarchyTree.begin(),inheritedkey);
	} else {
		it = classHirarchyTree.find(superkey);
		if (it==classHirarchyTree.end())
			throw exception("Super class not found!");

		it = classHirarchyTree.insert(it,inheritedkey);
	}

	serializeableClassKeyHirarchy[inheritedkey] = (*it);
}

/** retrieves the key of a class from the global class map
*
*	\note Classes are always referenced by their creator functions.
*
*	\param classcreator the creatorfunction of the class you want to know the key of.
*/
long GENERAL::GetSerializeableObjectKey(GENERAL::TClassCreator classcreator)
{
	GENERAL::TSerializeableClassListKeys::iterator it;
	it = serializeableClassListKeys.find(classcreator);
	if (it==serializeableClassListKeys.end())
		return -1;
//		throw new exception("Not a Serializeable object or not registered");
	return (*it).second;;
}

bool GENERAL::IsOfType(TClassCreator classcreator, long key)
{
	TSerializeableClassKeyHirarchy::iterator it = serializeableClassKeyHirarchy.find(key);
	if (it==serializeableClassKeyHirarchy.end())
		return false;

	long classkey = GetSerializeableObjectKey(classcreator);

	TClassHirarchyNode *node = it->second;
	while (node!=NULL)
	{
		if (classkey==node->data)
			return true;

		node = node->parent;
	}

	return false;
}

/** Deserialize an object from a buffer.
*
*	Extracts the information of the buffer and creates an object of the type specified in the buffer.
*
*	-# Extracts a key from the buffer.
*	-# Determines the class associated with that buffer
*	-# Creates an Object of thata class
*	-# Calls CSerializeableObject::DeSerialize() on that object
*	-# Returns the newly created Object.
*
*	\param data	the CBuffer object to create the new object from.
*/
GENERAL::sptr<GENERAL::CSerializeableObject> GENERAL::Deserialize(GENERAL::CBuffer &data, void *constructdata)
{
	GENERAL::sptr<GENERAL::CSerializeableObject> newobject;

	GENERAL::TSerializeableClassList::iterator it;
	long key=*((long*)data.Peek(sizeof(long)));

	it = serializeableClassList.find(key);
	if (it==serializeableClassList.end())
		THROW("Invalid Object-Key");

	newobject=it->second(constructdata);
	if (newobject!=NULL)
	{
		newobject->DeSerialize(data);
	} else {
		THROW("new failed");
	}
	
	return newobject;
}

void GENERAL::CSerializeableObject::SetSendToClientList(NETWORK::PClientInfoList _sendToClients) 
{
	if (sendToClients) sendToClients->ReleaseRef();
	if (_sendToClients) _sendToClients->GetRef();
	sendToClients=_sendToClients;
};


GENERAL::CBuffer *GENERAL::Serialize(sptr<GENERAL::CSerializeableObject> o)
{
	GENERAL::CBuffer *buf=new CMemBuffer(o->GetSerializedSize());
	o->Serialize(*buf);
	return buf;	
}

void GENERAL::DumpClassHirachy(int depth,TClassHirarchyNode *root)
{
	if (root==NULL)
	{
		root=*classHirarchyTree.begin();
	}

	char line[1024];
	int i;
	for (i=0;i<depth*4;i++)
	{
		line[i]=' ';
	}
	line[i]='\0';
	sprintf(line+i,"[%d]%s\n",root->data,serializeableClassListNames[root->data].data());
//	OutputDebugString(line);
	TClassHirarchyNode::iterator it;
	for (it=root->begin();it!=root->end();it++)
	{
		GENERAL::DumpClassHirachy(depth+1,*it);
	}

}

namespace GENERAL
{
	CBuffer::~CBuffer() {};

	/** Creates a new CMemBuffer object.
	*
	*	\param _size the size of the buffer to create.
	*/
	CMemBuffer::CMemBuffer(int _size) : pos(0), size(_size), attached(false) 
	{
		data = (char*)malloc(_size);
	}

	/** Creates a new CMemBuffer object from existing data.
	*
	*	If created using this method the data will not be destroyed when destroying the CMemBuffer object.
	*
	*	\param _data pointer to the date to use.
	*	\param _size the size of the buffer to create.
	*/
	CMemBuffer::CMemBuffer(void *_data, int _size) : 
		pos(0), 
		data((char*)_data), 
		size(_size), 
		attached(true) 
	{
	}

	/** Destroys the object.
	*
	*	If the object was created from existing data, the data will not be destroyed
	*/
	CMemBuffer::~CMemBuffer()
	{
		if (!attached) free(data);
	}

	/** Returns the current position in the buffer.
	*
	*	\throws exception "No more data available!"  If the actual position in the buffer plus the length of the requested date exceeds the total length of the buffer
	*	
	*	\param len the length of the data you want to access.
	*/
	void *CMemBuffer::Peek(int len)
	{
		if (pos+len>size)
			THROW("No more data available!");
		return data+pos;
	}

	/** Read data from the buffer.
	*
	*	Performs a CMemBuffer::Peek to check wheter the request is valid, and returns a pointer to the data if sucessful.
	*	After a sucessful read the pos variable will be increased by len.
	*
	*	\param len the length of the desired data.
	*/
	void *CMemBuffer::Read(int len)
	{
		void *res = Peek(len);
		pos+=len;
		return res;
	}

	/** Write data to the buffer.
	*
	*	Performs a CMemBuffer::Peek to check wheter the request is valid, and writes the data if sucessful
	*	After a sucessful write the pos variable will be increased by len.
	*
	*	\param _data pointer to the data you want to write into the buffer.
	*	\param len the length of the data to write into the buffer.
	*/
	void *CMemBuffer::Write(int len)
	{
		void *wdata=Peek(len);
		pos+=len;
		return wdata;
	}

	void CMemBuffer::Write(void *_data, int len)
	{
		void *wdata=Peek(len);
		memcpy(wdata,_data,len);
		pos+=len;
	}

	/** Return the current postion in the buffer.
	*
	*	This method returns the current position in the buffer.
	*	
	*/
	int CMemBuffer::GetPos()
	{
		return pos;
	}
	/** Return a pointer to the data.
	*/
	void *CMemBuffer::GetData()
	{
		return (void*)data;
	}
}


