#ifndef SERIALIZE_H
#define SERIALIZE_H

#include <map>
#include <string>
#include <conio.h>

#include "tick.h"
#include "tree.h"
#include "smartpointer.h"
//#include "../../network/base/clientinfo.h"

namespace NETWORK
{
	class CClientInfo;
	class TClientInfoList;
	typedef TClientInfoList *PClientInfoList;
}

namespace GENERAL {
	/** remove this! */
	class CException : public exception
	{
	#ifdef _DEBUG
		private:
			char *_what;

		public:
			inline CException(char *what, char *file, int line) : exception() 
			{
				int strlen = _cprintf("%s(%d): %s",file,line,what);
				_what = (char*)malloc(sizeof(char)*(strlen+1));
				sprintf("%s(%d): %s",file,line,what);
			}
			virtual ~CException();
			virtual const char *what();

	#else
		public:
			inline CException(char *what) : exception(what) {}
	#endif
	};

	#ifdef THROW
	#undef THROW
	#endif

	#ifdef _DEBUG
		#define MORFTHROW(message,exceptionclass) throw exceptionclass(message,__FILE__,__LINE__)
	#else
		#define MORFTHROW(message,exceptionclass) throw exceptionclass(message)
	#endif

	#define THROW(message) MORFTHROW(message,GENERAL::CException)

	/** Declares a Prototype for the Object creator function and defines the GetRef function
	*	\param classname Ezeugt den Funktionsheader static classname::CreateObject()
	*/
	#define DECLARE_SERIALIZEABLE(classname) \
		static ::GENERAL::sptr<::GENERAL::CSerializeableObject> CreateObject(void *); \
		virtual char *GetClassName() { return #classname; }

	/** Dieses Makro definert die Funktion CreateObject.
	*	\param classname Ezeugt Funktion classname::CreateObject()
	*/
	#define IMPLEMENT_SERIALIZEABLE(classname) \
		::GENERAL::sptr<GENERAL::CSerializeableObject> classname::CreateObject(void *data) \
		{ return ::GENERAL::sptr<GENERAL::CSerializeableObject>(new classname()); }	

	/** Dieses Makro erzeugt einen Eintrag in der Klassenliste.
	*	\param classname Human-readable name der Klasse fr die ein Key zugewiesen wird.
	*	\param key eindeutige ID fr diese Klasse.
	*/
	/** Liefert den key einer Klasse.
	*	\param classname Klasse deren Key geliefert werden soll.
	*/
	#define GET_KEY(classname) \
		::GENERAL::GetSerializeableObjectKey(classname::CreateObject)

	#define ISOFTYPE(object,classname) \
		(::GENERAL::IsOfType(classname::CreateObject,object->GetClassKey()))

	#define REGISTER_SERIALIZEABLE_BASECLASS(classname, keyname) \
		::GENERAL::RegisterSerializeableClass(#classname,classname::CreateObject, keyname); \
		::GENERAL::RegisterSerializeableClassHirarchy(0, keyname);

	#define REGISTER_SERIALIZEABLE(classname, keyname, superclass) \
		::GENERAL::RegisterSerializeableClass(#classname,classname::CreateObject, keyname); \
		::GENERAL::RegisterSerializeableClassHirarchy(GET_KEY(superclass), keyname);

//	#define Release(serializeableobject) \
//		if ((serializeableobject)->ReleaseRef()==0) delete (serializeableobject)
		

	/** Abstract class for reading and writing on data-blocks.
	*/
	class CBuffer
	{
	public:
		virtual ~CBuffer();

		virtual void *Peek(int len)=0;
		virtual void *Read(int len)=0;
		virtual void Write(void *data, int len)=0;
		virtual void *Write(int len)=0;
		virtual int GetPos()=0;
		virtual void *GetData()=0;
	};

	/** Implementation of the abstract class CBuffer.
	*/
	class CMemBuffer : public CBuffer
	{
	protected:
		char *data;		///< pointer to the begin of the data.
		int size;		///< size of the data.
		int pos;		///< actual position in the data.
		bool attached;  ///< true if the buffer was biult on extisting data.
	public:
		CMemBuffer(int _size);
		CMemBuffer(void *_data, int _size);
		virtual ~CMemBuffer();

		virtual void *Peek(int len);
		virtual void *Read(int len);
		virtual void Write(void *_data, int len);
		virtual void *Write(int len);
		virtual int GetPos();
		virtual void *GetData();
	};

	/** Basisklasse fr alle Serializeable Objekte.
	*/
	class CSerializeableObject {
	public:
		typedef struct _TSerializedData
		{
			long key;
		} TSerializedData;

	private:
		TGameTick lastPush;
		long key; ///<contains the classidentifier key
	protected:
		/// List of clients iterested in this Object
		NETWORK::PClientInfoList sendToClients;

		/// increases the reference counter and returns a pointer to the object
	public:
		DECLARE_SERIALIZEABLE(CSerializeableObject)

		CSerializeableObject();
		virtual ~CSerializeableObject();

		virtual long GetSerializedSize();
		virtual long Serialize(CBuffer &buffer);
		virtual long DeSerialize(CBuffer &buffer);
		virtual void Print(FILE *f,bool newline=true);

		inline NETWORK::PClientInfoList GetSendToClientList() {return sendToClients;};
		void SetSendToClientList(NETWORK::PClientInfoList _sendToClients);
		/** Decreases the reference counter and destroys th object only if ref==0
		*	\note do not use delete thoug to saftyness, use this method enstead
		*	\todo make threadsafe
		*/
//		virtual void Destroy(CSerializeableObject *so);// {ref--; if (ref==0) delete this;};

		inline long GetClassKey() {return key;};
		inline void SetClassKey(long _key) {key = _key;};
		inline TGameTick GetLastPush() {return lastPush;};
		inline void SetLastPush(TGameTick _lastPush) {lastPush = _lastPush;};
	};

	/** Typ der Creator-Funktion fr ein Serializeable Objekt.
	*/
//	class CDomain;
	typedef ::GENERAL::sptr<CSerializeableObject> (*TClassCreator)(void*);

	/** Type of the class list.
	*	key -> class
	*/
	typedef std::map<long, TClassCreator> TSerializeableClassList;

	/** Type of the class list.
	*	key -> class
	*/
	typedef std::map<long, std::string> TSerializeableClassListNames;

	/** Type of the class list.
	*	class -> key
	*/
	typedef std::map<TClassCreator, long> TSerializeableClassListKeys;

	/** Type of Classhirachy Tree
	*/
	typedef std::tree<long> TClassHirarchyTree;

	/** Type of Classhirachy Treenode
	*/
	typedef std::treenode<long> TClassHirarchyNode;
	
	/** Type of the map between class keys and Hirarchynodes
	*/
	typedef std::map<long, TClassHirarchyNode* > TSerializeableClassKeyHirarchy;


	/** Registers a class to the class list.
	*	\param classcreator	the creator of the class to be inserted.
	*	\warning dont call this method! use the REGISTER_SERIALIZEABLE makro.
	*/
	void RegisterSerializeableClass(char *classname, TClassCreator classcreator,long key);
	
	/** Registers a class to the Classhirarchy
		\param superkey key of the baseclass
		\param inheritedkey key of the childclass
	*/
	void RegisterSerializeableClassHirarchy(long superkey,long inheritedkey);

	/** Retrieves the key of a class from the class list.
	*	\param classcreator	the creator of the class to be inserted.
	*	\param key global key used to uniquely identify the class
	*	\warning dont call this method! use the GET_KEY makro.
	*/
	long GetSerializeableObjectKey(TClassCreator classcreator);

	bool IsOfType(TClassCreator classcreator, long key);
	void DumpClassHirachy(int depth=0,TClassHirarchyNode *root=NULL);
	
	/** Creates and deserializes an object from a databuffer object.
	*	\param data Reference to the dataobject.
	*/
	sptr<CSerializeableObject> Deserialize(CBuffer &data, void *constructdata);
	CBuffer *Serialize(sptr<CSerializeableObject> o);
}

#endif