#ifdef _DEBUG

#include <windows.h>

#include <map>

//#include "morftools.h"

class CTrace;
CTrace *trace=NULL;
static int allocCount=0;


class CTrace
{
private:
	struct TMemData
	{
		void *p;
		int size;
		char *file;
		int line;
		int allocNum;

		TMemData() : p(0),size(-1),file((char*)0),line(-1) {}

		TMemData &operator = (TMemData &md)
		{
			p=md.p;
			size=md.size;
			file=md.file;
			line=md.line;
			allocNum=md.allocNum;

			return *this;
		}

		TMemData(TMemData *md)
		{
			*this = *md;
		}

		TMemData(void *_p, int _size, char* _file, int _line, int _allocNum)
		{
			p = _p;
			size = _size;
			file = _file;
			line = _line;
			allocNum = _allocNum;
		}

		void Dump()
		{
			char l[1024];
			int i,n;
			i =sprintf(l,  "%s(%d) : #%d ",file,line,allocNum);
			i+=sprintf(l+i,"%d Bytes at 0x%08x not freed. data[",size,(long long)p);
			n=0;
			while (n<(size<16?size:16))
			{
				i+=sprintf(l+i,"%02hx ",((unsigned char*)p)[n]);
				n++;
			}
			if (size>n)
			{
				i+=sprintf(l+i,"... ]\n");
			} else
			if (n>0)
			{
				i--;
				i+=sprintf(l+i,"]\n");
			}
			OutputDebugString(l);
		}
	};
	std::map<void*,TMemData> data;
//	std::map<void*,TMemData> deleted;

	HANDLE hLock;
public:

	void Add(void *p, size_t s, char *file, int line)
	{
		if (WaitForSingleObject(hLock,2000)!=WAIT_OBJECT_0)
			throw exception("Could not lock trace synchronization object");

		data[p]=TMemData(p,(int)s,file,line,allocCount);

		if (ReleaseMutex(hLock)==0)
			throw exception("Could not unlock trace synchronization object");
		allocCount++;
	}

	void Remove(void *p)
	{
		if (WaitForSingleObject(hLock,2000)!=WAIT_OBJECT_0)
			throw exception("Could not lock trace synchronization object");

		std::map<void*,TMemData>::iterator it;
		it = data.find(p);
		if (it!=data.end())
		{
//			deleted[p]=it->second;
			data.erase(p);
		} else {
//			it = deleted.find(p);
//			if (it!=deleted.end())
//				GENERAL::dprintf("Already deleted %s, %i\n",it->second.file,it->second.line);
		}

		if (ReleaseMutex(hLock)==0)
			throw exception("Could not unlock trace synchronization object");
	}

	CTrace()
	{
		if ((hLock = CreateMutex(NULL,FALSE,NULL))==NULL)
		{
			throw exception("Cannot create trace synchronization object");
		}

		data.clear();
	}

	virtual ~CTrace()
	{
		size_t notFreed=0;
		trace=NULL;
		std::map<void*,TMemData>::iterator it;
		if (data.empty())
		{
			OutputDebugString("\n##########################################\n");
			OutputDebugString("# No memory leaks detected!              #\n");
			OutputDebugString("##########################################\n\n");
		} else {
			char line[1024];
			OutputDebugString("\n##########################################\n");
			sprintf(line,     "# %5d memory leaks detected!           #\n",data.size());
			OutputDebugString(line);
			OutputDebugString("##########################################\n");
			for (it=data.begin();it!=data.end();it++)
			{
				it->second.Dump();
				notFreed+=it->second.size;
			}
			sprintf(line,     "all in all %d Bytes are not freed\n\n",notFreed);
			OutputDebugString(line);
		}
		data.clear();
//		deleted.clear();
		CloseHandle(hLock);
	}
};

void InitTrace()
{
	trace = new CTrace();
}

void DoneTrace()
{
	delete trace;
}

void AddTrace(void *p, size_t s, const char *file, int line)
{
/*	if ((allocCount>=38691) && (allocCount<=38691+10))
	{
		int blu=12;
	}*/
	if (trace) trace->Add(p,s,(char*)file,line);
}

void DeleteTrace(void *p)
{
	if (trace) trace->Remove(p);
}

char *tracestrdup(const char *str, const char *file, int line)
{
	char *res = strdup(str);
	AddTrace(res,strlen(str)+1,file,line);
	return res;
}


void *tracemalloc(size_t s, const char *file, int line)
{
	void *p = malloc(s);
	AddTrace(p,s,file,line);
	return p;
}

void *tracerealloc(void *p, size_t s, const char *file, int line)
{
	DeleteTrace(p);
	p = realloc(p,s);
	AddTrace(p,s,file,line);
	return p;
}

void tracefree(void *p)
{
	static long long delcnt=0;
	delcnt++;
	DeleteTrace(p);
	free(p);
}


/*
void *operator new (size_t s,const char *file,int line)
{
	return tracemalloc(s);
}

void operator delete (void *p,const char *file,int line)
{
	tracefree(p);
}

void operator delete (void *p)
{
	tracefree(p);
}
*/
/*
void operator delete (void *p,const char *file,int line)
{
	free(p);
}*/

#endif //_DEBUG
