#ifndef TRACE_H
#define TRACE_H

#include <stdlib.h>

#ifdef _DEBUG

	void InitTrace();
	void DoneTrace();

	void AddTrace(void *p, size_t s, const char *file, int line);
	void DeleteTrace(void *p);


	char *tracestrdup(const char *str, const char *file, int line);
	void *tracemalloc(size_t s, const char *file, int line);
	void *tracerealloc(void *p, size_t s, const char *file, int line);
	void tracefree(void *p);

	inline void * __cdecl operator new(unsigned int size, const char *file, int line)
	{
		return (void *)tracemalloc(size,file,line);
	};

	inline void __cdecl operator delete(void *p, const char *file, int line)
	{
		delete p;
	}

	inline void __cdecl operator delete(void *p)
	{
		tracefree(p);
	};

	inline void * __cdecl operator new [] (unsigned int size, const char *file, int line)
	{
		return (void *)tracemalloc(size,file,line);
	};

	inline void __cdecl operator delete [] (void *p, const char *file, int line)
	{
		delete [] p;
	}

	inline void __cdecl operator delete [] (void *p)
	{
		tracefree(p);
	};

	#define DEBUG_NEW new(__FILE__, __LINE__)
	#define new DEBUG_NEW


	#undef malloc
	#undef free
	#undef strdup
	#undef realloc
	#define malloc(X) tracemalloc(X,__FILE__, __LINE__)
	#define realloc(P,X) tracerealloc(P,X,__FILE__, __LINE__)
	#define free(X) tracefree(X)
	#define strdup(X) tracestrdup(X,__FILE__, __LINE__)

#else  //_DEBUG

	inline void InitTrace() {}
	inline void DoneTrace() {}

#endif  //_DEBUG

#ifdef SAFE_DELETE
#undef SAFE_DELETE
#endif
#ifdef SAFE_DELETE_AR
#undef SAFE_DELETE_AR
#endif
#ifdef SAFE_FREE
#undef SAFE_FREE
#endif
#define SAFE_DELETE(x) { if (x!=NULL) delete x; x=NULL; }
#define SAFE_DELETE_AR(x) { if (x!=NULL) delete [] x; x=NULL; }
#define SAFE_FREE(x) {if (x!=NULL) free(x); x=NULL;}


#endif
