#ifndef CLIENTINFO_H
#define CLIENTINFO_H

#include "../../general/base/serialize.h"
#include "../../general/base/smartpointer.h"

namespace NETWORK
{
	/// BaseClass for ClientInformation
	class CClientInfo : public GENERAL::CSerializeableObject 
	{
	private:
		bool isValid;
		void *worldData;
	public:
		inline CClientInfo() : GENERAL::CSerializeableObject(), isValid(false), worldData(NULL)
		{
			SetClassKey(GET_KEY(CClientInfo));
		}
		virtual ~CClientInfo();

		DECLARE_SERIALIZEABLE(CClientInfo);

		inline bool IsValid() {return isValid;}
		inline void SetInvalid() {isValid=false;}
		inline void SetValid() {isValid=true;}
		inline void SetWorldData(void *data) {worldData=data;}
		inline void *GetWorldData() { return worldData;}

		virtual char *GetClientInfoString();
		virtual void Shutdown();
		virtual PClientInfoList GetClientInfo();
	};

	class TClientInfoList : public std::list<GENERAL::sptr<CClientInfo> > 
	{
	private:
		int refCnt;
	public:
		inline TClientInfoList() : std::list<GENERAL::sptr<CClientInfo> >(),refCnt(1) {};
		virtual ~TClientInfoList();
		TClientInfoList *GetRef() {refCnt++; return this;};
		void ReleaseRef() {if (refCnt==1) delete this; else refCnt--;};
	};

	typedef TClientInfoList *PClientInfoList;	

	class CNetworkEvent : public GENERAL::CSerializeableObject 
	{
	public:
		DECLARE_SERIALIZEABLE(CNetworkEvent);
		typedef enum 
			{
				UNKNOWN = 0,
				NEWCONNECTION = 1,
				CONNECTIONRESET = 2,  // connection closed by foreign host
				CONNECTIONSHUTDOWN = 3, // connection shutdown
				CRITICALERROR = 4,
			} TEventTyp;

	private:
		TEventTyp typ;
		GENERAL::sptr<CClientInfo> ci;

	public:
		inline CNetworkEvent() : GENERAL::CSerializeableObject(), typ(UNKNOWN), ci(NULL) 
		{
			SetClassKey(GET_KEY(CNetworkEvent));
		}
		inline CNetworkEvent(TEventTyp _typ, GENERAL::sptr<CClientInfo> &_ci) : GENERAL::CSerializeableObject(), typ(_typ), ci(_ci)
		{
			SetClassKey(GET_KEY(CNetworkEvent));
		}
		inline GENERAL::sptr<CClientInfo> &GetClientInfo() {return ci;}
		inline TEventTyp GetTyp() {return typ;}
	};
}

#endif