#ifndef DOMAIN_CHARACTER_OBJECT_H
#define DOMAIN_CHARACTER_OBJECT_H

#include <morfbase.h>
#include "md3object.h"

#include "../../game/implementation/playerstats.h"

namespace DOMAIN
{
	class CCharacterObject : public CMD3Object
	{

	private: 
		float newdir;
		GENERAL::TGameTick until;

	protected:

		sptr<CGraphObject> inventory;
		sptr<CWorldObject> useTarget;
		TGameTick nextUse;
		int health;

		sptr<GAME::CPlayerStats> stats;

	public:
		DECLARE_SERIALIZEABLE(CCharacterObject);

		inline CCharacterObject(CDomain *_dom) 
			: CMD3Object(_dom),
			  inventory(new CGraphObject(domain, 1, 0)),
			  nextUse(-1),
			  health(100)/// @todo id of inventorytype should be dynamic!!
		{ 
			SetClassKey(GET_KEY(CCharacterObject));
			inventory->SetContainerType(CONFIG::g_config.ObjectStatus["INCONTAINER"]);

			//boundingBox = GENERAL::TBox(GENERAL::CVector3f(-33.953125,-12.062500,-23.093750), GENERAL::CVector3f(-5.750000,6.796875,8.187500));
			boundingBox = GENERAL::TBox(GENERAL::CVector3f(-13.218750,-12.484375,-23.109375), GENERAL::CVector3f(15.328125,14.093750,14.640625));
			collisionMode = COLLISIONMODE_BOX;

			stats = new GAME::CPlayerStats(this, 0,-1,0,100,100);
			newdir = 0;
			until = 0;
			stats->SetKiType(1000);

		};

		inline CCharacterObject(
			CDomain *_domain,
			int _odid,
			long _id,
			CLevel *_lvl,
			GENERAL::CVector3f _pos,
			GENERAL::CVector3f _rot,
			GENERAL::CVector3f _scale,
			float _mass,
			float _radius,
			GENERAL::TBox &_boundingBox,
			GENERAL::CVector3f _velocity)
				: CMD3Object(_domain, _odid, _id, _lvl, _pos, _rot, _scale,_mass,_radius, _boundingBox, _velocity),
				  inventory(new CGraphObject(domain, 1, 0)),/// @todo id of inventorytype should be dynamic!!
				  nextUse(-1),
				  health(100)
						{
							SetClassKey(GET_KEY(CCharacterObject));
							inventory->SetContainerType(CONFIG::g_config.ObjectStatus["INCONTAINER"]);

							//boundingBox = GENERAL::TBox(GENERAL::CVector3f(-33.953125,-12.062500,-23.093750), GENERAL::CVector3f(-5.750000,6.796875,8.187500));
							boundingBox = GENERAL::TBox(GENERAL::CVector3f(-13.218750,-12.484375,-23.109375), GENERAL::CVector3f(15.328125,14.093750,14.640625));

							collisionMode = COLLISIONMODE_BOX;
							newdir = 0;
							until = 0;

								stats = new GAME::CPlayerStats(this, 0,-1,0,100,100);
								stats->SetKiType(1000);

						};

		virtual ~CCharacterObject();
		virtual bool Load();

		virtual long GetSerializedSize();
		virtual long Serialize(GENERAL::CBuffer &buffer);
		virtual long DeSerialize(GENERAL::CBuffer &buffer);
		virtual void Print(FILE *f,bool newline=true);

		virtual int AddLink(sptr<DOMAIN::CMD3Object> &_obj);		//  wenn link ein CMD3Object --> dessen feStatus auf FESTATUS_TAGGED setzen
		virtual int RemoveLink(int _tag, int _id);  //  --"-- FESTATUS_NORMAL


		virtual void TimeStep(GENERAL::TGameTick dt); 

		virtual void HandleDomainObjectActionRequest(sptr<CDomainCommand> &cmd);		
		virtual void HandleDomainObjectActionReply(sptr<CDomainCommand> &cmd);		

		virtual void GetActionList(tActionList &_actions, int witch=-1);
		virtual int ExecuteAction(int _actionId,  CGraphObject *_caller, long lParam = 0, void *pParam = 0);

		inline TGameTick &GetNextUse() { return nextUse;}
		inline void SetNextUse(TGameTick _nextUse) { nextUse = _nextUse;}
		inline void IncNextUse(TGameTick _time) { nextUse += _time;}

		virtual int ActionTake(sptr<CGraphObject> &_caller, bool isRequest);
		virtual int ActionDrop(sptr<CGraphObject> &_caller, bool isRequest);
		virtual	int ActionEquip(sptr<CGraphObject> &_caller, bool isRequest);
		virtual	int ActionUse(sptr<CGraphObject> &_caller, bool isRequest);
		virtual int ActionOpenInventory(sptr<CGraphObject> &_caller, bool isRequest);
		virtual int ActionChatReply(CGraphObject *_caller, long _whoId, char *_msg);
		virtual void ActionHit(int damage);
		virtual ECollisionAction Collided(CVector3f pos, CVector3f normal);


		inline sptr<CGraphObject> &GetInventory() { return inventory; };

		inline sptr<GAME::CPlayerStats> &GetStats() { return stats; };

		inline void SetHealth(int h) { health=h; };
		inline int GetHealth() { return health; };

		
	

		
	};

}

#endif
