#include <morfbase.h>

#include "chatdomain.h"
#include "logindata.h"

#include "characterobject.h"

#include <string>
#include <trace.h>

#include "../../game/objectdescriptor.h"

namespace DOMAIN
{

	CChatDomain::CChatDomain(bool _isserver) 
		: CDomain(),
		  name(NULL),
  		  isserver(_isserver),
		  shouldTerminate(false)
	{
		if (isserver)
		{
			name=strdup("server");
			userPassMap["martin"]="6maerz78";
			userPassMap["don"]="James1";
			userPassMap["c1"]="c1";
			userPassMap["c2"]="c2";
			userPassMap["c3"]="c3";
			userPassMap["c4"]="c4";
		} else 
			name = strdup("<not yet set>");
	}

	CChatDomain::~CChatDomain()
	{
		handleNetwork();
		SAFE_FREE(name);
		handleNetwork();
	}

	void CChatDomain::Init()
	{
		CDomain::Init();
		if (!objMan) objMan = new CSimpleObjectManager(this);
	}

	bool CChatDomain::ShouldTerminate()
	{
		return shouldTerminate;
	}

	void CChatDomain::Terminate()
	{
		shouldTerminate=true;
	}

	void CChatDomain::pushNetworkEvent(sptr<GENERAL::CSerializeableObject> o)
	{
		networkEvents.push_back(o);
	}

	void CChatDomain::pushIncomingNetwork(GENERAL::CBuffer *o,NETWORK::PClientInfoList _from) 
	{
		incomingObjects.push_back(CChatDomain::TIncomingObject(o,_from->GetRef()));
	}
	void CChatDomain::updateObject(GENERAL::CBuffer &buffer, sptr<CGameObject> go, NETWORK::PClientInfoList from)
	{
		go->DeSerialize(buffer);
	}

	void CChatDomain::handleNetwork()
	{
		GENERAL::CBuffer *buffer;
		NETWORK::PClientInfoList from;

		while (!networkEvents.empty())
		{
			sptr<GENERAL::CSerializeableObject> so;
			so = *networkEvents.begin(); 
			networkEvents.pop_front();

			handleNetworkObject(so);
			networkEvents.clear();
		}

		//if (a TypeOf CSerializeableObject) {};

		CGameObject::TSerializedData *sd;

		while (!incomingObjects.empty())
		{
			sptr<GENERAL::CSerializeableObject> so;
			buffer = incomingObjects.front().buffer;
			from = incomingObjects.front().from;
			incomingObjects.pop_front();

			sd=(CGameObject::TSerializedData*)buffer->Peek(sizeof(long)*2);
			if (GENERAL::IsOfType(CGameObject::CreateObject,sd->key))
			// check for incoming data is a CGameObject
			{
				// check whether it has an global id
				if (sd->id>0)
				{
					sptr<CGameObject> go = objMan->GetRef(sd->id);
					// check if data type matches with own object
					if ((go!=NULL) && (go->GetClassKey()==sd->key))
					{
//						updateObject(*buffer,go,from);
						delete buffer;
						continue;
					}
				}
			}

			so = GENERAL::Deserialize(*buffer, this);
			if (ISOFTYPE(so,DOMAIN::CGameObject))
			{
				((DOMAIN::CGameObject*)&so)->SetDeathReckoning();
			}
			if (ISOFTYPE(so,DOMAIN::CCharacterObject))
			{
				if (((sptr<DOMAIN::CCharacterObject>)so)->GetID()>5999)
				{
					((sptr<DOMAIN::CCharacterObject>)so)->GetStats()->SetKiType(1000);

				}
				
			}


			delete buffer;
			so->SetSendToClientList(from);

			handleNetworkObject(so);
		}
	}

	void CChatDomain::handleNetworkObject(sptr<GENERAL::CSerializeableObject> &so)
	{
		NETWORK::TClientInfoList::iterator it;

		if (so->GetClassKey()==GET_KEY(NETWORK::CNetworkEvent))
		{
			sptr<NETWORK::CNetworkEvent> ne=so;

			switch (ne->GetTyp())
			{
			case NETWORK::CNetworkEvent::NEWCONNECTION:
/*				if (isserver)
					connectedToNotAuth.push_back(ne->GetClientInfo());
				else
				{
					connectedTo.push_back(ne->GetClientInfo());

					std::string newline="application:Connected to ";
//					newline.append(ne->GetClientInfo()->GetClientInfoString());

					localso = new CLoginData("martin","6maerz78");
					localso->SetSendToClientList(ne->GetClientInfo()->GetClientInfo());
					APPLICATION::CApplication::GetApplication()->GetNetwork()->PushObject(localso,ne->GetClientInfo()->GetClientInfo());
					localso->ReleaseRef();
				}*/
				break;

				// connection closed by foreign host
			case NETWORK::CNetworkEvent::CONNECTIONRESET:
			case NETWORK::CNetworkEvent::CONNECTIONSHUTDOWN:
//				connectedTo.remove(ne->GetClientInfo());
//				connectedToNotAuth.remove(ne->GetClientInfo());
//1				ne->GetClientInfo()->ReleaseRef();

				if ((!isserver) && (ne->GetTyp()==NETWORK::CNetworkEvent::CONNECTIONRESET))
					shouldTerminate=true;
				break;
			}

//1			ne->ReleaseRef();
			return;
		}
		
//1		so->ReleaseRef();
	}


	void CChatDomain::execute()
	{
		handleNetwork();
	}

	bool CChatDomain::AuthUser(char *name, char *password)
	{
		std::string storedpass=userPassMap[name];
		return (storedpass==std::string(password));
	}

	CDomain::PConsoleData CChatDomain::GetConsoleData()
	{
		return &consoleData;
	}
	
	CDomain::PObjectList CChatDomain::GetVisibleObjects()
	{
		return NULL;
	}

	void CChatDomain::SetName(char *_name) 
	{ 
		SAFE_FREE(name); 
		name=strdup(_name); 
	};

} // namespace


