#include <morfbase.h>
#include "domainquakebsp.h"
#include <trace.h>

namespace DOMAIN
{
	IMPLEMENT_SERIALIZEABLE_DOMAIN(CQuakeBsp);

	CQuakeBsp::CQuakeBsp(CDomain *_domain) : CLevel(_domain)
	{
		m_numOfNodes	 = 0;
		m_numOfLeafs	 = 0;
		m_numOfLeafFaces = 0;
		m_numOfPlanes	 = 0;
		m_numOfTextures	 = 0;

		m_pNodes		 = NULL;
		m_pLeafs		 = NULL;
		m_pPlanes		 = NULL;
		m_pLeafFaces	 = NULL;
		m_pTextures		 = NULL;
		memset(&m_clusters, 0, sizeof(GENERAL::tBSPVisData));

		m_szFname=NULL;

		SetClassKey(GET_KEY(CQuakeBsp));

		loaded=false;
	}

	CQuakeBsp::CQuakeBsp(CDomain *_domain,int _odid,long _id, char *_szFname) : CLevel(_domain,_odid,_id), m_szFname(strdup(_szFname))
	{
		SetClassKey(GET_KEY(CQuakeBsp));

		m_numOfNodes	 = 0;
		m_numOfLeafs	 = 0;
		m_numOfLeafFaces = 0;
		m_numOfPlanes	 = 0;
		m_numOfTextures	 = 0;

		m_pNodes		 = NULL;
		m_pLeafs		 = NULL;
		m_pPlanes		 = NULL;
		m_pLeafFaces	 = NULL;
		m_pTextures		 = NULL;



		memset(&m_clusters, 0, sizeof(GENERAL::tBSPVisData));

		t.traceRatio		= 0;
		t.traceType			= 0;
		//t.traceRadius		= 0;

		t.collided	= false;
		t.grounded	= false;
		t.tryStep	= false;

		t.traceMins = GENERAL::CVector3f(0, 0, 0);
		t.traceMaxs = GENERAL::CVector3f(0, 0, 0);
		t.extents   = GENERAL::CVector3f(0, 0, 0);

		t.collisionNormal = GENERAL::CVector3f(0, 0, 0);        

		loaded=false;

		
		//Load();


	}

	CQuakeBsp::~CQuakeBsp()
	{
		for (int i=0; i<m_clusters.numOfClusters; i++)
		{
	//		SAFE_DELETE(m_pClusterChilds[i]);
		}
		delete [] m_pClusterChilds;
		delete [] m_pClusterVisibleLights;
		free(m_pTextures);
		delete [] m_pNodes;
		delete [] m_pLeafs;
		delete [] m_pPlanes;
		delete [] m_pLeafFaces;
		delete [] m_pBrushes;
		delete [] m_pBrushSides;
		delete [] m_pLeafBrushes;
		delete [] m_clusters.pBitsets;
		SAFE_FREE(m_szFname);
	}

	long CQuakeBsp::GetSerializedSize()
	{
		return CWorldObject::GetSerializedSize()+
				sizeof(long)+
				(int)strlen(m_szFname);
	}

	long CQuakeBsp::Serialize(GENERAL::CBuffer &buffer)
	{
		CWorldObject::Serialize(buffer);

		long len=(long)strlen(m_szFname);
		buffer.Write(&len,sizeof(len));
		buffer.Write(m_szFname,len);
		
		return buffer.GetPos();
	}

	long CQuakeBsp::DeSerialize(GENERAL::CBuffer &buffer)
	{
		CWorldObject::DeSerialize(buffer);

		if (m_szFname!=NULL) free(m_szFname);
		long len=*(long*)buffer.Read(sizeof(len));
		m_szFname=(char *)malloc(sizeof(char)*(len+1)); m_szFname[len]='\0';
		memcpy(m_szFname,buffer.Read(len),len);

		return buffer.GetPos();
	}

	void CQuakeBsp::Print(FILE *f,bool newline)
	{
	}

	void CQuakeBsp::Assign(CGameObject *o)
	{
	}

	void CQuakeBsp::TimeStep(GENERAL::TGameTick dt) 
	{
	}

	bool CQuakeBsp::Load()
	{
		GENERAL::CQuakeBSPFile qbsp;
		if (!qbsp.Open(m_szFname))
			return false;

		m_pTextures = qbsp.GetTextures(m_numOfTextures);
		m_pBrushes = qbsp.GetBrushes(m_numOfBrushes);
		m_pBrushSides = qbsp.GetBrushSides(m_numOfBrushSides);
		m_pNodes = qbsp.GetNodes(m_numOfNodes);
		m_pLeafs=qbsp.GetLeafs(m_numOfLeafs);
		m_pLeafFaces = qbsp.GetLeafFaces(m_numOfLeafFaces);
		m_pPlanes = qbsp.GetPlanes(m_numOfPlanes);
		m_pLeafBrushes =  qbsp.GetLeafBrushes(m_numOfLeafBrushes);
		qbsp.GetVisData(m_clusters);

		m_pClusterChilds = new sptr<DOMAIN::CGraphObject>[m_clusters.numOfClusters];
		for (int i=0; i<m_clusters.numOfClusters; i++)
		{
			m_pClusterChilds[i] = sptr<DOMAIN::CGraphObject>(new CGraphObject(domain));
			AddChild(m_pClusterChilds[i]);
			m_pClusterChilds[i]->SetfeData(
				sptr<FRONTEND::CRenderAble>(
					new FRONTEND::CRenderAble(NULL, &m_pClusterChilds[i])));
		}

		m_pClusterVisibleLights = new CGraphObject::tLightsType[m_clusters.numOfClusters];

		loaded=true;
		UpdateClusterChilds();

		GENERAL::CQuakeBSPFile::TEntities ents;
		qbsp.GetEntities(ents);

		std::pair<GENERAL::CQuakeBSPFile::TEntities::iterator,GENERAL::CQuakeBSPFile::TEntities::iterator> p;
		GENERAL::CQuakeBSPFile::TEntities::iterator it;
		p = ents.equal_range("light");

		for (it = p.first; it!=p.second; it++)
		{
			sptr<CLight> light = sptr<CLight>(new CLight(domain,0,0,this));

			GENERAL::CQuakeBSPFile::TEntity &ent = it->second;
			GENERAL::CQuakeBSPFile::TEntity::iterator eit;

			eit = ent.find("origin");
			if (eit!=ent.end())
			{
				GENERAL::CVector3f pos;
				// convert coords due to overall rotation of the bsp
				sscanf(eit->second.data(), "%f %f %f", &pos.x, &pos.z, &pos.y);
				pos.z=-pos.z;
				light->SetPos(pos);
			}
			eit = ent.find("_color");
			if (eit!=ent.end())
			{
				GENERAL::CVector4f color;
				sscanf(eit->second.data(), "%f %f %f", &color.x, &color.y, &color.z);
				light->SetDiffuseColor(color);
			}
			eit = ent.find("light");
			if (eit!=ent.end())
			{
				float att;
				sscanf(eit->second.data(), "%f", &att);
				light->SetLinearAtt(1/(att*10));
				light->SetConstantAtt(0.0f);
				light->SetQuadAtt(0.0f);

			}


			//$$ APPLICATION::CApplication::GetApplication()->GetObjectMan()->Add(light);
			domain->GetObjMan()->Add(light);

			AddChild(light);
		}

		if (APPLICATION::CApplication::GetApplication()->GetFE())
			SetfeData(APPLICATION::CApplication::GetApplication()->GetFE()->CreateObject(this,0));

		return CLevel::Load();
	}

	void CQuakeBsp::UpdateClusterChilds()
	{
		tChildsType::iterator it;
		tChildsType dummy;

		for (it = childs.begin(); it!=childs.end();it++)
		{
			if (GET_KEY(CGraphObject)!=(*it)->GetClassKey())
			{
				int cl = m_pLeafs[FindLeaf(((CWorldObject*)&(*it))->GetPos())].cluster;
				if (cl>=0)
				{
					m_pClusterChilds[cl]->AddChild(*it);

					if (ISOFTYPE((*it),DOMAIN::CLevelObject))
					{
						((DOMAIN::CLevelObject*)&(*it))->GetLevelInfo()=cl;
					}
				}

				(*it)->SetParent(this);
			}

		}
/*		for (it = childs.begin(); it!=childs.end();it++)
		{
			if (GET_KEY(CGraphObject)==(*it)->GetClassKey())
			{
				dummy.push_back(*it);
				continue;
			}
			int cl = m_pLeafs[FindLeaf(((CWorldObject*)&(*it))->GetPos())].cluster;
			if (cl>=0)
				m_pClusterChilds[cl]->AddChild(*it);
		}
		childs.clear();
		childs = dummy;*/
	}

	
	//
	///
	//////
	////////////////
	/////////////////////////////////////
	// COLLISION DETECTION ////////////////////////////////////////////////////

	int CQuakeBsp::GetCluster(GENERAL::CVector3f &pos)
	{
		return m_pLeafs[FindLeaf(pos)].cluster;
	}

	int CQuakeBsp::FindLeaf(GENERAL::CVector3f &pos)
	{
		GENERAL::tBSPNode *nodes=m_pNodes;

		int current=0;
		float distance;

		while (current>=0)
		{
			GENERAL::tBSPPlane &plane=m_pPlanes[nodes[current].plane];
			distance = pos.ScalarMultiply(plane.vNormal)-plane.d;

			if (distance>=0)
			{
				// i am in front of the plane
				current=nodes[current].front;
			} else {
				// i am in back of the plane
				current=nodes[current].back;
			}
		}

		return ~current;
	}

	/////////////////////////////////// TRY TO STEP \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
	/////
	/////	This checks a bunch of different heights to see if we can step up
	/////
	/////////////////////////////////// TRY TO STEP \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*

/*	GENERAL::CVector3f CQuakeBsp::TryToStep(GENERAL::CVector3f _start, GENERAL::CVector3f _end)
	{
		for(float height = 1.0f; height <= t.stepHeight; height++)
		{
			t.collided = false;
			t.tryStep = false;

			GENERAL::CVector3f vStepStart = GENERAL::CVector3f(_start.x, _start.y + height, _start.z);
			GENERAL::CVector3f vStepEnd   = GENERAL::CVector3f(_end.x, _start.y + height, _end.z);
					
			GENERAL::CVector3f vStepPosition = Trace(NULL,vStepStart, vStepEnd);

			if(!t.collided)
			{
				GENERAL::dprintf("[ INFO ] stepped!\n");
				return vStepPosition;		
			}
		}
		return _start;
	}*/

	void CQuakeBsp::TryToStep(GENERAL::CVector3f _end)
	{
		FitTo(GENERAL::CVector3f(0.0,t.stepHeight,0.0f));
		_end.y+=t.stepHeight;
		TracePrivate(t.outPos,_end);
		FitTo(GENERAL::CVector3f(0.0,-t.stepHeight,0.0f));
		return;
	}
		
	/////////////////////////////////// TRACE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
	/////
	/////	This takes a start and end position (general) to test against the BSP brushes
	/////
	/////////////////////////////////// TRACE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*


	GENERAL::CVector3f CQuakeBsp::Trace(DOMAIN::CPhysicalLevelObject *_obj,GENERAL::CVector3f _start, GENERAL::CVector3f _end)
	{
		t.traceRatio = 1.0f;
		t.collided = false;
		t.tryStep = false;
		t.trySlide = true;
		t.grounded=false;

		t.stepHeight = _obj->GetStepHeight();
		t.tryStep = (t.stepHeight>0);
		
		m_obj = _obj;

		switch (_obj->GetCollisionMode())
		{
		case COLLISIONMODE_BOX:
			{
				t.traceMins = _obj->GetBoundingBox().minV;
				t.traceMaxs = _obj->GetBoundingBox().maxV;

				t.extents = GENERAL::CVector3f(-t.traceMins.x > t.traceMaxs.x ? -t.traceMins.x : t.traceMaxs.x,
												-t.traceMins.y > t.traceMaxs.y ? -t.traceMins.y : t.traceMaxs.y,
												-t.traceMins.z > t.traceMaxs.z ? -t.traceMins.z : t.traceMaxs.z);

				t.traceType = TYPE_BOX;

			}  break;

		case COLLISIONMODE_SPHERE:
			{
				t.traceRadius = _obj->GetBoundingRadius();
				t.traceType = TYPE_SPHERE;
			}  break;
		case COLLISIONMODE_POINT: t.traceType = TYPE_RAY; break;
		
		case COLLISIONMODE_NONE: return _end; break;

		}

		TracePrivate(_start, _end);

		if(t.collided && t.tryStep)
		{
			TryToStep(_end);
		}
			

		/** @ todo: grounded -> physicallevelobject */

//		if (t.collisionNormal.y > 0.2f || t.grounded)
		if ((t.collisionNormal.y > 0.2f) && (t.collided))
		{
			t.grounded = true;
			CVector3f vel = _obj->GetVelocity();
			vel.y=0.0f;
			_obj->SetVelocity(vel);
		}
		else
			t.grounded = false;

		return t.outPos;
	}

	void CQuakeBsp::FitTo(GENERAL::CVector3f _direction)
	{
		t.trySlide=false;
		GENERAL::CVector3f newPos = t.outPos+_direction;
		TracePrivate(t.outPos,newPos);
	}

	

	void CQuakeBsp::TracePrivate(GENERAL::CVector3f _start, GENERAL::CVector3f _end)
	{
		t.traceRatio = 1.0f;

		t.inStart = _start;
		t.inEnd = _end;

		CheckNode(0, 0.0f, 1.0f, _start, _end);

		if(t.traceRatio == 1.0f)
		{
			t.outPos = _end;
		}
		else	// collision
		{
			t.collided=true;
			t.outPos = _start + ((_end - _start) * t.traceRatio);
			ECollisionAction action = m_obj->Collided(t.outPos,t.collisionNormal);

			if (action==COLLISIONACTION_SLIDE)
			{
				GENERAL::CVector3f vMove = _end - t.outPos;
				float distance = vMove.ScalarMultiply(t.collisionNormal);
				GENERAL::CVector3f _endPosition = _end - t.collisionNormal*distance;
				TracePrivate(t.outPos, _endPosition);
			}
		}
	}


	/////////////////////////////////// CHECK NODE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
	/////
	/////	This traverses the BSP to find the brushes closest to our position
	/////
	/////////////////////////////////// CHECK NODE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*

	void CQuakeBsp::CheckNode(int nodeIndex, float startRatio, float endRatio, GENERAL::CVector3f _start, GENERAL::CVector3f _end)
	{
		if(nodeIndex < 0)
		{
			GENERAL::tBSPLeaf *pLeaf = &m_pLeafs[-(nodeIndex + 1)];

			for(int i = 0; i < pLeaf->numOfLeafBrushes; i++)
			{
				GENERAL::tBSPBrush *pBrush = &m_pBrushes[m_pLeafBrushes[pLeaf->leafBrush + i]];

				if	( (pBrush->numOfBrushSides > 0) && 
					  (m_pTextures[pBrush->textureID].textureType & 0x00010001))
				{
					CheckBrush(pBrush);
				}
			}
			return;
		}

		GENERAL::tBSPNode *pNode = &m_pNodes[nodeIndex];
		GENERAL::tBSPPlane *pPlane = &m_pPlanes[pNode->plane];
		
		float startDistance = _start.ScalarMultiply(pPlane->vNormal) - pPlane->d;
		float endDistance = _end.ScalarMultiply(pPlane->vNormal) - pPlane->d;
		float offset = 0.0f;

		if (t.traceType == TYPE_SPHERE) 
			offset = t.traceRadius;

		if (t.traceType == TYPE_BOX)
		{	
			offset = (float)(GENERAL::MMfabs( t.extents.x * pPlane->vNormal.x ) +
							 GENERAL::MMfabs( t.extents.y * pPlane->vNormal.y ) +
							 GENERAL::MMfabs( t.extents.z * pPlane->vNormal.z ) );
		}

/**		Error in www.gametutorials.com (Quake 3 BSP Collision Part3)
		gametutorials passes startDistance & andDistance to the recursion instead of 
		startRatio and endRatio

		CheckNode(pNode->front, startDistance, endDistance, _start, _end);		
		CheckNode(pNode->back, startDistance, endDistance, _start, _end);
**/

		if(startDistance >= offset && endDistance >= offset)
		{
			CheckNode(pNode->front, startRatio, endRatio,_start,_end);
		}
		else if(startDistance < -offset && endDistance < -offset)
		{
			CheckNode(pNode->back, startRatio, endRatio,_start,_end);
		}	
		else
		{
			float Ratio1 = 1.0f, Ratio2 = 0.0f, middleRatio = 0.0f;
			GENERAL::CVector3f vMiddle;

			int side = pNode->front;

			if(startDistance < endDistance)
			{
				side = pNode->back;

				float inverseDistance = 1.0f / (startDistance - endDistance);
				Ratio1 = (startDistance - offset - BSPEPSILON) * inverseDistance;
				Ratio2 = (startDistance + offset + BSPEPSILON) * inverseDistance;
			}
			else if(startDistance > endDistance)
			{
				float inverseDistance = 1.0f / (startDistance - endDistance);
				Ratio1 = (startDistance + offset + BSPEPSILON) * inverseDistance;
				Ratio2 = (startDistance - offset - BSPEPSILON) * inverseDistance;
			}

			if (Ratio1 < 0.0f) Ratio1 = 0.0f;
			else if (Ratio1 > 1.0f) Ratio1 = 1.0f;

			if (Ratio2 < 0.0f) Ratio2 = 0.0f;
			else if (Ratio2 > 1.0f) Ratio2 = 1.0f;

			middleRatio = startRatio + ((endRatio - startRatio) * Ratio1);
			vMiddle = _start + ((_end - _start) * Ratio1);

			CheckNode(side, startRatio, middleRatio, _start, vMiddle);

			middleRatio = startRatio + ((endRatio - startRatio) * Ratio2);
			vMiddle = _start + ((_end - _start) * Ratio2);

			if(side == pNode->back)
				CheckNode(pNode->front, middleRatio, endRatio, vMiddle, _end);
			else
				CheckNode(pNode->back, middleRatio, endRatio, vMiddle, _end);
		}
	}


	/////////////////////////////////// CHECK BRUSH \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
	/////
	/////	This checks our movement vector against all the planes of the brush
	/////
	/////////////////////////////////// CHECK BRUSH \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*

	void CQuakeBsp::CheckBrush(GENERAL::tBSPBrush *pBrush)
	{
		float startRatio = -1.0f;		
		float endRatio = 1.0f;			
		bool startsOut = false;			

		for(int i = 0; i < pBrush->numOfBrushSides; i++)
		{
			GENERAL::tBSPBrushSide *pBrushSide = &m_pBrushSides[pBrush->brushSide + i];
			GENERAL::tBSPPlane *pPlane = &m_pPlanes[pBrushSide->plane];

			float offset = 0.0f;
			
			if(t.traceType == TYPE_SPHERE) offset = t.traceRadius;
			
			float startDistance = t.inStart.ScalarMultiply(pPlane->vNormal) - (pPlane->d + offset);
			float endDistance = t.inEnd.ScalarMultiply(pPlane->vNormal) - (pPlane->d + offset);

			GENERAL::CVector3f vOffset = GENERAL::CVector3f(0, 0, 0);

			if(t.traceType == TYPE_BOX)
			{
				vOffset.x = (pPlane->vNormal.x < 0)	? t.traceMaxs.x : t.traceMins.x;
				vOffset.y = (pPlane->vNormal.y < 0)	? t.traceMaxs.y : t.traceMins.y;
				vOffset.z = (pPlane->vNormal.z < 0)	? t.traceMaxs.z : t.traceMins.z;

				startDistance = (t.inStart + vOffset).ScalarMultiply(pPlane->vNormal) - pPlane->d;
				endDistance   = (t.inEnd + vOffset).ScalarMultiply(pPlane->vNormal) - pPlane->d;
			}

			if(startDistance > 0)	startsOut = true;

			if(startDistance > 0 && endDistance > 0)
				return;

			if(startDistance <= 0 && endDistance <= 0)
				continue;

			if(startDistance > endDistance)
			{
				float Ratio1 = (startDistance - BSPEPSILON) / (startDistance - endDistance);

				if(Ratio1 > startRatio)
				{
					startRatio = Ratio1;
					t.collided = true;		// Let us know we collided!

					t.collisionNormal = pPlane->vNormal;

/*					if ((t.inStart.x != t.inEnd.x || t.inStart.z != t.inEnd.z) && pPlane->vNormal.y != 1)
					{
						t.tryStep = true;
					}*/
					
					if(t.collisionNormal.y >= 0.2f)
						t.grounded = true;
				}
			}
			else
			{
				float Ratio = (startDistance + BSPEPSILON) / (startDistance - endDistance);

				if(Ratio < endRatio)
					endRatio = Ratio;
			}
		}

		if(startsOut == false)
		{
			return;
		}

		if(startRatio < endRatio)
		{
			if(startRatio > -1 && startRatio < t.traceRatio)
			{
				if(startRatio < 0)
					startRatio = 0;

				t.traceRatio = startRatio;
			}
		}
	}

	bool CQuakeBsp::IsClusterVisible(int current, int test)
	{
		if (current<0) return true;
		GENERAL::tBSPVisData &clusters = m_clusters;
		GENERAL::UBYTE vis = clusters.pBitsets[current*clusters.bytesPerCluster+test/8];
//		return (vis>>(test%8))&0xFE;
#pragma warning( disable : 4800 )
		return ((bool)(vis & (1<<((GENERAL::UBYTE)(test)&7))));
#pragma warning( default : 4800 )
	}

	void CQuakeBsp::AddChild(sptr<CGraphObject> _child)
	{
		/*if ((GET_KEY(CGraphObject)==_child->GetClassKey()) || (!loaded))
		{
			CGraphObject::AddChild(_child);
			return;
		}*/

		if (_child->GetID()==27)
		{
			int blu=1;
		}

		CGraphObject::AddChild(_child);

		if (!loaded) 
		{
			if (ISOFTYPE(_child,DOMAIN::CLevelObject))
			{
				((CLevelObject*)&_child)->GetLevelInfo() = -1;
			}
		} 
		else 
		{
			int cl = m_pLeafs[FindLeaf(((CWorldObject*)&_child)->GetPos())].cluster;
			if (cl!=-1)
			{
				m_pClusterChilds[cl]->AddChild(_child);
			} else {
				GENERAL::dprintf("[ WARN ] CQuakeBsp::AddChild: Child(%d) is outside - ignored\n",_child->GetID());
			}

			if (ISOFTYPE(_child,DOMAIN::CLevelObject))
			{
				((CLevelObject*)&_child)->GetLevelInfo() = cl;
			}
		}

		_child->SetParent(this);
	}

	void CQuakeBsp::RemoveChild(sptr<CGraphObject> _child)
	{
/*		if ((GET_KEY(CGraphObject)==_child->GetClassKey()) || (!loaded))
		{
			CGraphObject::RemoveChild(_child);
			return;
		}*/
		CGraphObject::RemoveChild(_child);
		if (ISOFTYPE(_child,DOMAIN::CLevelObject))
		{
			int cl = ((CLevelObject*)&_child)->GetLevelInfo();
			if (cl!=-1)
				m_pClusterChilds[cl]->RemoveChild(_child);

			((CLevelObject*)&_child)->GetLevelInfo() = -1;
		} 
		else 
		{
			int cl = m_pLeafs[FindLeaf(((CWorldObject*)&_child)->GetPos())].cluster;
			m_pClusterChilds[cl]->RemoveChild(_child);
		}
	}

	void CQuakeBsp::HandoverChild(sptr<CGraphObject> _target, sptr<CGraphObject> _child)
	{
	}

	void CQuakeBsp::LevelObjectMoved(CLevelObject *_obj)
	{
		static int i=0;
		i++;
		sptr<CLevelObject> lo = (sptr<CLevelObject>)domain->GetObjMan()->GetPtr(_obj->GetID());

		int oldCluster = _obj->GetLevelInfo();
		int newCluster = m_pLeafs[FindLeaf(_obj->GetPos())].cluster;

		if (oldCluster!=newCluster)
		{
			if (oldCluster!=-1)
				m_pClusterChilds[oldCluster]->RemoveChild(lo);

			if (newCluster!=-1)
				m_pClusterChilds[newCluster]->AddChild(lo);

			_obj->GetLevelInfo() = newCluster;
			_obj->SetParent(this);
		}

	}


	
	void CQuakeBsp::RegisterLight(sptr<CLight> _light)
	{
		GENERAL::CVector3f lpos = _light->GetPos();
		int cluster = m_pLeafs[FindLeaf(lpos)].cluster;

		for (int i=0; i<m_clusters.numOfClusters; i++)
		{

			if (IsClusterVisible(cluster,i))
			{
				DOMAIN::CGraphObject::tLightsType *lightlist = &m_pClusterVisibleLights[i];
				lightlist->push_back(_light);
                

				
			}
		}
	}

	void CQuakeBsp::UnRegisterLight(sptr<CLight> _light)
	{
		int cluster = m_pLeafs[FindLeaf(_light->GetPos())].cluster;

		for (int i=0; i<m_clusters.numOfClusters; i++)
		{
			if (IsClusterVisible(cluster,i))
			{
				CGraphObject::tLightsType::iterator it;

				for (it=m_pClusterVisibleLights[i].begin(); it!=m_pClusterVisibleLights[i].end(); it++)
				{
					if (*it==_light) break;
				}
				if (it!=m_pClusterVisibleLights[i].end())
					m_pClusterVisibleLights[i].erase(it);
			}
		}
	}

}
