#ifndef DOMAINOBJECT_QUAKE_BSP_H
#define DOMAINOBJECT_QUAKE_BSP_H

#include "../../general/fileformats/quakebspfile.h"
#include <morfbase.h>

namespace DOMAIN
{
	const float kGravity =  9.8f;
	const float kJumpAcceleration = 4;

	class CQuakeBsp : public CLevel
	{
	public:

		struct tTrace
		{
			int		traceType;
			float	traceRatio;
			float	traceRadius;
			float	stepHeight;
			float	outFraction;

			bool	collided;
			bool	grounded;
			bool	tryStep;
			bool	trySlide;
			bool	outStartsOut;
			bool	outAllSolid;

			GENERAL::CVector3f traceMins;
			GENERAL::CVector3f traceMaxs;
			
			GENERAL::CVector3f extents;	
			GENERAL::CVector3f collisionNormal;

			GENERAL::CVector3f inStart;
			GENERAL::CVector3f inEnd;
			GENERAL::CVector3f outPos;
			
		};

	public:
		int m_numOfNodes;
		int m_numOfLeafs;
		int m_numOfLeafFaces;
		int m_numOfPlanes;
		int m_numOfBrushes;
		int m_numOfBrushSides;
		int m_numOfLeafBrushes;
		int m_numOfTextures;		

		GENERAL::tBSPNode				*m_pNodes;
		GENERAL::tBSPLeaf				*m_pLeafs;
		GENERAL::tBSPPlane				*m_pPlanes;
		int								*m_pLeafFaces;
		GENERAL::tBSPVisData			 m_clusters;	
		GENERAL::tBSPTexture			*m_pTextures;
		GENERAL::tBSPBrush				*m_pBrushes;
		GENERAL::tBSPBrushSide			*m_pBrushSides;
		int								*m_pLeafBrushes;
	
		sptr<DOMAIN::CGraphObject>			*m_pClusterChilds;
		DOMAIN::CGraphObject::tLightsType	*m_pClusterVisibleLights;




		///// unsere

		tTrace	t;

		char *m_szFname;

		CPhysicalLevelObject *m_obj;


		DECLARE_SERIALIZEABLE(CQuakeBsp);

		CQuakeBsp(CDomain *_domain);
		CQuakeBsp(CDomain *_domain,int _odid,long _id, char *_szFname);
		virtual ~CQuakeBsp();

		virtual long GetSerializedSize();
		virtual long Serialize(GENERAL::CBuffer &buffer);
		virtual long DeSerialize(GENERAL::CBuffer &buffer);
		virtual void Print(FILE *f,bool newline=true);

		virtual void Assign(CGameObject *o);
		virtual void TimeStep(GENERAL::TGameTick dt); 

		virtual bool IsClusterVisible(int current, int test);

		virtual bool Load();

		int FindLeaf(GENERAL::CVector3f &pos);
		int GetCluster(GENERAL::CVector3f &pos);

		GENERAL::CVector3f Trace(DOMAIN::CPhysicalLevelObject *_obj,GENERAL::CVector3f _start, GENERAL::CVector3f _end);
		

		
		void FitTo(GENERAL::CVector3f _direction);
		void TryToStep(GENERAL::CVector3f _end);

		void TracePrivate(GENERAL::CVector3f _etart, GENERAL::CVector3f _end);
		void CheckNode(int _nodeIndex, float _startRatio, float _endRatio, GENERAL::CVector3f _start, GENERAL::CVector3f _end);
		void CheckBrush(GENERAL::tBSPBrush *_brush);

		virtual void AddChild(sptr<CGraphObject> _child);
		virtual void RemoveChild(sptr<CGraphObject> _child);
		virtual void HandoverChild(sptr<CGraphObject> _target, sptr<CGraphObject> _child);
		virtual void RegisterLight(sptr<CLight> _light);
		virtual void UnRegisterLight(sptr<CLight> _light);
		virtual void UpdateClusterChilds();
		virtual void LevelObjectMoved(CLevelObject *_obj);

	};



}
#endif