#ifndef FILEFORMAT_MD3_H
#define FILEFORMAT_MD3_H

#include <stdlib.h>
#include <stdio.h>
#include <string>
#include <fstream>
#include <vector>

#include <morfbase.h>

namespace GENERAL
{

	//#define MAX_TEXTURES 100

	class CVector3 
	{
	public:
		float x, y, z;
	};

	class CVector2 
	{
	public:
		float x, y;
	};


	struct tMD3FaceTex
	{
		int vertIndex[3];
		int coordIndex[3];
	};

	struct tMD3MaterialInfo
	{
		char strName[255];
		char strFile[255];

		//GENERAL::TColor3ub color;

		unsigned char  color[3];	

		int textureId;

		float uTile, vTile;
		float uOffset, vOffset;
	};

	struct tMD3Object
	{
		int numOfVerts;
		int numOfFaces;
		int numTexVertex;
		int materialId;

		bool hasTexture;
		char strName[255];

		CVector3  *pVerts;		
		GENERAL::CVector3f  *pNormals;	
		CVector2  *pTexVerts;	

		//GENERAL::CVector3f   *pVerts;
		//GENERAL::CVector3f   *pNormals;
		//GENERAL::TMorfPointf *pTexVerts;

		tMD3FaceTex *pFaces;
	};

	struct tMD3AnimationInfo
	{
		char strName[255];
		int startFrame;
		int endFrame;
		int loopingFrames;
		int framesPerSecond;

	};

	struct tMD3Frame
	{
		float	mins[3];					
		float	maxs[3];					
		float	origin[3];				
		float	boundingRadius;						
		char	name[16];				
	};

	struct tMD3Model
	{
		int numOfObjects;
		int numOfMaterials;
		int numOfAnimations;
		int numOfFrames;

		std::vector<tMD3MaterialInfo> pMaterials;
		std::vector<tMD3Object> pObject;
		std::vector<tMD3AnimationInfo> pAnimations; 
		
		tMD3Frame *pFrames; 
					
		int currentAnim;					
		int currentFrame;					
		int nextFrame;						
		float t;							
		float lastTime;		

		int numOfTags;						
		struct tMD3Tag *pTags;			

		tMD3Model **pLinks;		
	};

	/************************************************************************* 
	* MD3 Zeug
	**************************************************************************
	*
	*
	*
	*/

	#define kLower	0
	#define kUpper	1
	#define kHead	2
	#define kWeapon	3

	typedef enum 
	{
		BOTH_DEATH1 = 0,		
		BOTH_DEAD1,				
		BOTH_DEATH2,			
		BOTH_DEAD2,				
		BOTH_DEATH3,			
		BOTH_DEAD3,				

		TORSO_GESTURE,			
		TORSO_ATTACK,			
		TORSO_ATTACK2,			
		TORSO_DROP,				
		TORSO_RAISE,			
		TORSO_STAND,			
		TORSO_STAND2,			

		LEGS_WALKCR,			
		LEGS_WALK,				
		LEGS_RUN,				
		LEGS_BACK,				
		LEGS_SWIM,				
		LEGS_JUMP,				
		LEGS_LAND,				
		LEGS_JUMPB,				
		LEGS_LANDB,				
		LEGS_IDLE,				
		LEGS_IDLECR,			
		LEGS_TURN,				

		MAX_ANIMATIONS			
	} eMD3Animations;


	
	typedef enum 
	{
		MD3STATE_IDLE = 0,
		MD3STATE_WALK,
		MD3STATE_DIE,
		MD3STATE_DEAD,
		
		MD3_MAX_STATES			
	} eMD3States;

	typedef enum 
	{
		MD3COMMAND_NOC = -1,
		MD3COMMAND_FINISHED,
		MD3COMMAND_STOP,
		MD3COMMAND_WALK,
		MD3COMMAND_DIE,
		MD3COMMAND_RESURRECT
				
	} eMD3StateCommands;


	extern int g_MD3TorsoStateAnimation[4];
	extern int g_MD3LegsStateAnimation[4];



				
	






	struct tMD3Header
	{ 
		char	fileID[4];					
		int		version;					
		char	strFile[68];				
		int		numFrames;					
		int		numTags;					
		int		numMeshes;					
		int		numMaxSkins;				
		int		headerSize;					
		int		tagStart;					
		int		tagEnd;						
		int		fileSize;					
	};

	struct tMD3MeshInfo
	{
		char	meshID[4];					
		char	strName[68];				
		int		numMeshFrames;				
		int		numSkins;					
		int     numVertices;				
		int		numTriangles;				
		int		triStart;					
		int		headerSize;					
		int     uvStart;					
		int		vertexStart;				
		int		meshSize;					
	};

	struct tMD3Tag
	{
		char		strName[64];			
		CVector3	vPosition;				
		float		rotation[3][3];			
	};

/***
**	http://linux.ucla.edu/~phaethon/q3/formats/md3format.html

Datatype	name/purpose	Description

VEC3		MIN_BOUNDS		First corner of the bounding box.
VEC3		MAX_BOUNDS		Second corner of the bounding box.
VEC3		LOCAL_ORIGIN	Local origin, usually (0, 0, 0).
F32			RADIUS			Radius of bounding sphere.
U8 * 16		NAME			Name of Frame. ASCII character string, NUL-terminated (C-style).

*/



	struct tMD3Triangle
	{
		signed short vertex[3];			
		unsigned char normal[2];			
	};

	struct tMD3Face
	{
		int vertexIndices[3];				
	};

	struct tMD3TexCoord
	{
	float textureCoord[2];
	};

	struct tMD3Skin 
	{
		char strName[68];
	};

	class CLoadMD3
	{
	public:

		CLoadMD3();								

		bool ImportMD3(tMD3Model *pModel, char *strFileName);
		bool LoadSkin(tMD3Model *pModel, char* strSkin);
		bool LoadShader(tMD3Model *pModel, char* strShader);

	private:

		void ReadMD3Data(tMD3Model *pModel);
		void ConvertDataStructures(tMD3Model *pModel, tMD3MeshInfo meshHeader);
		void CleanUp();
		
		FILE *m_FilePointer;

		tMD3Header				m_Header;			

		tMD3Skin				*m_pSkins;			
		tMD3TexCoord			*m_pTexCoords;		
		tMD3Face				*m_pTriangles;		
		tMD3Triangle			*m_pVertices;		
		tMD3Frame				*m_pFrames;			
	};
}

#endif