#include "md3model.h"

namespace MD3MODEL
{

	CLoadMD3Link::CLoadMD3Link()
	{
		memset(&m_Header, 0, sizeof(tMD3Header));
		m_pShaders=NULL;
		m_pTexCoords=NULL;
		m_pTriangles=NULL;
		m_pFrames=NULL;
	}

	bool CLoadMD3Link::ReadMD3LinkData(tMD3Link *pLink, char *strFileName)
	{
		int i = 0;

		FILE *fp;

		memset(pLink,0,sizeof(tMD3Link));


		char withext[512];

		sprintf(withext,"%s.md3",strFileName);

		fp = GENERAL::morfOpenFile(withext, "rb");

		if(!fp) 
		{
			GENERAL::dprintf("[FAILED] LoadMD3Link: Unable to find the file: %s.md3!\n", strFileName);
			return false;
		}
		
		fread(&m_Header, 1, sizeof(tMD3Header), fp);

		char *ID = m_Header.fileID;

		if((ID[0] != 'I' || ID[1] != 'D' || ID[2] != 'P' || ID[3] != '3') || m_Header.version != 15)
		{
			GENERAL::dprintf("[FAILED] LoadMD3Link: Invalid file format (Version not 15): %s!\n", strFileName);
			return false;
		}

		m_pFrames = new tMD3Frame [m_Header.numFrames];
		fread(m_pFrames, sizeof(tMD3Frame), m_Header.numFrames, fp);

		pLink->numOfFrames = m_Header.numFrames;
		pLink->pFrames = m_pFrames;
		
		tMD3TagLoader *pTagsLoader;
		pLink->pTags = new tMD3Tag [m_Header.numFrames * m_Header.numTags];
		pTagsLoader = new tMD3TagLoader [m_Header.numFrames * m_Header.numTags];

		fread(pTagsLoader, sizeof(tMD3TagLoader), m_Header.numFrames * m_Header.numTags, fp);
		for (i=0; i<m_Header.numFrames * m_Header.numTags; i++)
		{
			pLink->pTags[i].tagid = CONFIG::g_config.TagTypes[pTagsLoader[i].strName];
			pLink->pTags[i].rotation = pTagsLoader[i].rotation;
			pLink->pTags[i].vPosition = pTagsLoader[i].vPosition;
		}
		delete [] pTagsLoader;


		/*		
		for (int n=0; n<m_Header.numFrames * m_Header.numTags; n++)
		{
			GENERAL::dprintf("%s %f,%f,%f\n", pLink->pTags[n].strName, pLink->pTags[n].vPosition.x, pLink->pTags[n].vPosition.y,pLink->pTags[n].vPosition.z);
		}*/


		pLink->numOfTags = m_Header.numTags;
		
		//pLink->pLinks = (tMD3Model **) malloc(sizeof(tMD3Model) * m_Header.numTags);
		//for (i = 0; i < m_Header.numTags; i++)
		//	pLink->pLinks[i] = NULL;

		

		long meshOffset = m_Header.ofsMeshes;

		//fseek(fp, m_Header.ofsMeshes, SEEK_SET);

		tMD3MeshInfo meshHeader;

		for (i = 0; i < m_Header.numMeshes; i++)
		{
			
			fseek(fp, meshOffset, SEEK_SET);
			fread(&meshHeader, sizeof(tMD3MeshInfo), 1, fp);
			
			m_pShaders   = new tMD3Shader [meshHeader.numShader];
			m_pTexCoords = new tMD3TexCoord [meshHeader.numVertices];
			m_pTriangles = new tMD3Triangle [meshHeader.numTriangles];
			m_pVertices  = new tMD3Vertex [meshHeader.numVertices * meshHeader.numMeshFrames];

			fseek(fp, meshOffset + meshHeader.ofsShader, SEEK_SET);
			fread(m_pShaders, sizeof(tMD3Shader), meshHeader.numShader, fp);

			for (int n=0; n<meshHeader.numShader; n++)
			{
				char *res;
				strlwr(m_pShaders[n].strName);
				if ((res=strstr(m_pShaders[n].strName,".tga"))!=NULL) 
					*res='\0';
				if ((res=strstr(m_pShaders[n].strName,".jpg"))!=NULL) 
					*res='\0';
			}
			
			fseek(fp, meshOffset + meshHeader.ofsTriangles, SEEK_SET);
			fread(m_pTriangles, sizeof(tMD3Triangle), meshHeader.numTriangles, fp);

			fseek(fp, meshOffset + meshHeader.ofsTexCoord, SEEK_SET);
			fread(m_pTexCoords, sizeof(tMD3TexCoord), meshHeader.numVertices, fp);

			fseek(fp, meshOffset + meshHeader.ofsVertices, SEEK_SET);
			fread(m_pVertices, sizeof(tMD3Vertex), meshHeader.numMeshFrames * meshHeader.numVertices, fp);

			ConvertDataStructures(pLink, meshHeader);

			delete [] m_pShaders;    
			delete [] m_pTexCoords;
			delete [] m_pTriangles;
			delete [] m_pVertices;   

			meshOffset += meshHeader.meshSize;
			
		}
		fclose(fp);

			tMD3Frame fr;

			GENERAL::dprintf("***************************\n");
			GENERAL::dprintf("*\n");
			GENERAL::dprintf("* File: %s\n",strFileName);
			GENERAL::dprintf("*\n");
			fr = pLink->pFrames[0];
			GENERAL::dprintf("**** FIRST BOUNDINGBOX ****\n");
			GENERAL::dprintf("*\n");
			GENERAL::dprintf("* BoundingboxMins = %f,%f,%f\n", fr.mins[0],fr.mins[1],fr.mins[2]);
			GENERAL::dprintf("* BoundingboxMaxs = %f,%f,%f\n", fr.maxs[0],fr.maxs[1],fr.maxs[2]);
			GENERAL::dprintf("*\n");


        /*** read skins
		*
		* @todo: alle skins in ein file rein mit <skinname>...</skinname>
		*
		*/

		sprintf(withext,"%s.skin",strFileName);
		fp = GENERAL::morfOpenFile(withext, "r");

		if(!fp) 
		{
			GENERAL::dprintf("[ WARN ] LoadMD3Link: Unable to find the file: %s.skin!\n", strFileName);
		}
		else
		{

			char line[512];

			while (fgets(line,512,fp)!=NULL)
			{
				char *token;
				token = strtok(line,", \t");
				if (token==NULL)
				{
					GENERAL::dprintf("[ WARN ] LoadMD3Link: invalid line in: %s.skin! Skipped!\n", strFileName);
					continue;	
				}
				for (int i=0; i<pLink->numOfMeshes; i++)
				{
					if (stricmp(pLink->pMeshes[i].strName,token)==0)
					{
						token=strtok(NULL,", \t\n\r");
						if (token==NULL)
						{
							GENERAL::dprintf("[ WARN ] LoadMD3Link: invalid line in: %s.skin! Skipped!\n", strFileName);
						} else 
						{
							char *res = strstr(token,".tga");
							if (res!=NULL) *res='\0';
							strcpy(pLink->pMeshes[i].pShaders[0].strName, token);
						}
					}
				}
			}

			fclose(fp);
		}

		/*** read animations ***/

		pLink->fsmName[0]='\0';
		sprintf(withext,"%s.anim",strFileName);
		fp = GENERAL::morfOpenFile(withext, "r");

		if(!fp) 
		{
			GENERAL::dprintf("[ WARN ] LoadMD3Link: Unable to find the file: %s.anim!\n", strFileName);
			return true;
		}

		char line[512];
		int animCnt=0;
		int legsOffset=0;
		bool calculateLegsOffset=false;
		bool isLegAnim;
		bool isFirstLegsAnim;

		while (fgets(line,512,fp)!=NULL)
		{
			char *comment;
			if ((comment = strstr(line,"//"))!=NULL)
			{
				*comment = '\0';
			}

			char *token;

			token = strtok(line," \t\n");
			if (token==NULL)
				continue;	
  			if ( (*token=='\n') || (*token=='/') )
				continue;

			if (stricmp(token,"#fsm")==0)
			{
				token = strtok(NULL," \t\n");
				if (token==NULL)
				{
					GENERAL::dprintf("[ WARN ] LoadMD3Link: invalid line in: %s.anim! Skipped!\n", strFileName);
					continue;	
				}
				strcpy(pLink->fsmName,token);
			} else if (stricmp(token,"#animations")==0)
			{
				break;
			}
		}

		while (fgets(line,512,fp)!=NULL)
		{
			calculateLegsOffset=(strstr(line,"BOTH_DEAD3")!=NULL);
			isLegAnim=(strstr(line,"LEGS")!=NULL);
			isFirstLegsAnim = (strstr(line,"LEGS_WALKCR")!=NULL);

			char *token,*eptr;
			token = strtok(line," \t");
			if (token==NULL)
			{
				GENERAL::dprintf("[ WARN ] LoadMD3Link: invalid line in: %s.anim! Skipped!\n", strFileName);
				continue;	
			}

  			if ( (*token=='\n') || (*token=='/') )
				continue;

			

			tMD3AnimationInfo anim;

			anim.startFrame = strtol(token,&eptr,0);
			if ( (eptr==token) ||
					( (*eptr!='\0') && (*eptr!='\n') )
					)
			{
				GENERAL::dprintf("[ WARN ] LoadMD3Link: invalid line in: %s.anim! Skipped!\n", strFileName);
				continue;	
			}
			if (isLegAnim)
			{
				if (isFirstLegsAnim)
				{
					legsOffset=anim.startFrame-legsOffset;
				}

				anim.startFrame-=legsOffset;
			}

			token = strtok(NULL," \t");
			if (token==NULL)
			{
				GENERAL::dprintf("[ WARN ] LoadMD3Link: invalid line in: %s.anim! Skipped!\n", strFileName);
				continue;	
			}
			anim.numFrames = strtol(token,&eptr,0);
			if ( (eptr==token) ||
					( (*eptr!='\0') && (*eptr!='\n') )
					)
			{
				GENERAL::dprintf("[ WARN ] LoadMD3Link: invalid line in: %s.anim! Skipped!\n", strFileName);
				continue;	
			}

			token = strtok(NULL," \t");
			if (token==NULL)
			{
				GENERAL::dprintf("[ WARN ] LoadMD3Link: invalid line in: %s.anim! Skipped!\n", strFileName);
				continue;	
			}
			anim.loopingFrames = strtol(token,&eptr,0);
			if ( (eptr==token) ||
					( (*eptr!='\0') && (*eptr!='\n') )
					)
			{
				GENERAL::dprintf("[ WARN ] LoadMD3Link: invalid line in: %s.anim! Skipped!\n", strFileName);
				continue;	
			}
			if (anim.loopingFrames==0)
				anim.loopingFrames=1;

			token = strtok(NULL," \t");
			if (token==NULL)
			{
				GENERAL::dprintf("[ WARN ] LoadMD3Link: invalid line in: %s.anim! Skipped!\n", strFileName);
				continue;	
			}
			anim.framesPerSecond = strtol(token,&eptr,0);
			if ( (eptr==token) ||
					( (*eptr!='\0') && (*eptr!='\n') )
					)
			{
				GENERAL::dprintf("[ WARN ] LoadMD3Link: invalid line in: %s.anim! Skipped!\n", strFileName);
				continue;	
			}

			token = strtok(NULL," \t");
			token = strtok(NULL," \t");
			if (token==NULL)
			{
				GENERAL::dprintf("[ WARN ] LoadMD3Link: invalid line in: %s.anim! Skipped!\n", strFileName);
				continue;	
			}
			char *res = 0;
			if (res = strstr(token,"_")) token = res+1;
            
			anim.id = -1;

			for (std::map<std::string, long>::iterator it=CONFIG::g_config.AnimationIds.begin(); it!=CONFIG::g_config.AnimationIds.end(); it++)
			{
				if (stricmp(it->first.data(),token)) 
				{
					anim.id=it->second;
					break;
				}
			}
                        

			pLink->pAnimations.push_back(anim);
			pLink->numOfAnimations++;
	
			if (calculateLegsOffset)
				legsOffset = anim.numFrames+anim.startFrame;
		}
		if (ferror(fp))
		{
			GENERAL::dprintf("[FAILED] LoadMD3Link: Error reading the file: %s.anim!\n", strFileName);
			return false;
		}


		if (pLink->numOfAnimations > 15)
		{
			fr = pLink->pFrames[LEGS_IDLE-7];
			GENERAL::dprintf("**** LEGIDLE BOUNDINGBOX ****\n");
			GENERAL::dprintf("*\n");
			GENERAL::dprintf("* BoundingboxMins = %f,%f,%f\n", fr.mins[0],fr.mins[1],fr.mins[2]);
			GENERAL::dprintf("* BoundingboxMaxs = %f,%f,%f\n", fr.maxs[0],fr.maxs[1],fr.maxs[2]);
			GENERAL::dprintf("*\n");
		}
		GENERAL::dprintf("*****************************\n");



		return true;

	}

	void CLoadMD3Link::ConvertDataStructures(tMD3Link *pLink, tMD3MeshInfo meshHeader)
	{
		int i = 0;

		pLink->numOfMeshes++;
			
		tMD3Mesh currentMesh = {0};

		strcpy(currentMesh.strName, meshHeader.strName);

		currentMesh.numOfVerts   = meshHeader.numVertices;
		currentMesh.numOfTriangles   = meshHeader.numTriangles;
		currentMesh.numShaders = meshHeader.numShader;

		currentMesh.pVerts     = new GENERAL::CVector3f [currentMesh.numOfVerts * meshHeader.numMeshFrames];
		currentMesh.pNormals   = new GENERAL::CVector3f [currentMesh.numOfVerts * meshHeader.numMeshFrames];
		currentMesh.pTexCoords = new GENERAL::CVector2f [currentMesh.numOfVerts];
		currentMesh.pTriangles = new tMD3Triangle [currentMesh.numOfTriangles];
		currentMesh.pShaders   = new tMD3Shader [meshHeader.numShader];

		for (i=0; i < meshHeader.numShader; i++)
		{
			memcpy(currentMesh.pShaders+i,m_pShaders+i,sizeof(tMD3Shader));
		}

		for (i=0; i < currentMesh.numOfVerts * meshHeader.numMeshFrames; i++)
		{
			currentMesh.pVerts[i].x =  m_pVertices[i].vertex[0] / 64.0f;
			currentMesh.pVerts[i].y =  m_pVertices[i].vertex[1] / 64.0f;
			currentMesh.pVerts[i].z =  m_pVertices[i].vertex[2] / 64.0f;

			float lat = ((float)m_pVertices[i].normal[1]) * (2.0f*PI) / 255.0f;
			float lng = ((float)m_pVertices[i].normal[0]) * (2.0f*PI) / 255.0f;
			currentMesh.pNormals[i].x = cosf(lat)*sinf(lng);
			currentMesh.pNormals[i].y = sinf(lat)*sinf(lng);
			currentMesh.pNormals[i].z = cosf(lng);
//			currentMesh.pNormals[i] = GENERAL::CVector3f(1,0,0);
		}

		for (i=0; i < currentMesh.numOfVerts; i++)
		{
			currentMesh.pTexCoords[i].x =  m_pTexCoords[i].textureCoord.x;
			currentMesh.pTexCoords[i].y =  m_pTexCoords[i].textureCoord.y;
		}

		for(i=0; i < currentMesh.numOfTriangles; i++)
		{
			currentMesh.pTriangles[i].vertexIndices[0] = m_pTriangles[i].vertexIndices[0];
			currentMesh.pTriangles[i].vertexIndices[1] = m_pTriangles[i].vertexIndices[1];
			currentMesh.pTriangles[i].vertexIndices[2] = m_pTriangles[i].vertexIndices[2];
		}
		
		pLink->pMeshes.push_back(currentMesh);
	}




}