#ifndef TCPNETWORK_H
#define TCPNETWORK_H

#include <morfbase.h>
#include <windows.h>
#include <winsock.h>
#include "TCPsocket.h"

namespace NETWORK
{
	class CTCPNetwork : public CNetwork
	{
	private:
		typedef struct _TOutgoingObject 
		{
			inline _TOutgoingObject(sptr<GENERAL::CSerializeableObject> _so, PClientInfoList _destination)
				: so(_so),destination(_destination) {};
			inline _TOutgoingObject(const _TOutgoingObject &oo)
				: so(oo.so),destination(oo.destination) {};

			sptr<GENERAL::CSerializeableObject> so;
			PClientInfoList destination;
		} TOutgoingObject;

		typedef std::list<TOutgoingObject> TObjectList;

	protected:
		TClientInfoList connectedTo;

		HANDLE hNetworkLock;

		HANDLE hSenderThread;
		HANDLE hOutgoingSemaphore;
		bool bTerminateSender;
		TObjectList outgoingObject;

		CTCPServerSocket *serverSocket;


	public:
		inline CTCPNetwork(DOMAIN::CDomain *_domain, bool listen) : 
				CNetwork(_domain),
				bTerminateSender(false),
				serverSocket(NULL)
			{
				Init(listen);
			};

		/** virtual Destructor are nesecary
		*/
		virtual ~CTCPNetwork();

		/** Initializes the Networkclass
		*	\param listen Networkclass should act as server
		*/
		virtual void Init(bool listen);
		/** Deinitializes the Network and shutdown all connections
		*/
		virtual void Done();

		/** Connect to a the server
		*/
		virtual void Connect();
/*		virtual void Disconnect(char *ip,long port) = 0;*/

		/** Push a Object to the Outgoing list
		*	\param o the object to be pushed
		*
		*	o will be serialized, dipatched and sent to all clients witch need
		*	the Object.
		*/
		virtual void PushObject(sptr<GENERAL::CSerializeableObject> o, PClientInfoList destination);

		/** Normaly Networkclasses are multi threaded
		*	Call this function to prevent the networkclasse to access to any gamedata
		*/
		virtual void Lock(DWORD time=INFINITE);

		/** Normaly Networkclasses are multi threaded
		*	Call this function to reallow the access to the gamedata
		*/
		virtual void Unlock();

	private:
		static void staticSenderThread(CTCPNetwork *self);

	protected:
		virtual void SenderThread();

		friend class CTCPServerSocket;
		friend class CTCPClientSocket;
	};

} // namespace NETWORK

#endif