#include "TCPsocket.h"
#include "TCPnetwork.h"
#include <trace.h>

namespace NETWORK
{
/**********************************************************************************/
/* CTCPClientInfo                                                                 */
/**********************************************************************************/
	CTCPClientInfo::~CTCPClientInfo()
	{
		delete GetSocket();
	}

	void CTCPClientInfo::Shutdown()
	{
		GetSocket()->Shutdown();
	}

	char *CTCPClientInfo::GetClientInfoString()
	{
		return inet_ntoa(((sockaddr_in*)(clientsocket->GetpAddr()))->sin_addr);
	}

/**********************************************************************************/
/* CTCPSocket                                                                     */
/**********************************************************************************/
	CTCPSocket::~CTCPSocket()
	{
		Shutdown();
	}

	void CTCPSocket::Shutdown()
	{
		if (socketfd!=0) closesocket(socketfd);
		socketfd=0;
	}

	int CTCPSocket::DoRecv(char *data,int len,int flags)
	{
		int res,pos=0;
		while (pos<len)
		{
			res=recv(socketfd,data+pos,len-pos,flags);
			if (res==SOCKET_ERROR)
				NETTHROW("Could not reveive data");
			pos+=res;
		}
		return (pos);
	}

	int CTCPSocket::DoSend(char *data,int len,int flags)
	{
		int res,pos=0;
		while (pos<len)
		{
			res=send(socketfd,data+pos,len-pos,flags);
			if (pos==SOCKET_ERROR)
				NETTHROW("Could not send data");
			pos+=res;
		}
		return (pos);
	}


/**********************************************************************************/
/* CTCPClientSocket                                                               */
/**********************************************************************************/
	CTCPClientSocket::~CTCPClientSocket()
	{
		Shutdown();
		if (listenThread!=(HANDLE)-1L)
		{
			if (WaitForSingleObject(listenThread,2000)==WAIT_TIMEOUT)
			{
				NETTHROW("Listen Thread did not finish within 2 seconds -> releasing Resources anyway");
			}
			listenThread=(HANDLE)-1L;
		}
		clientInfoList->ReleaseRef();
	}

	void CTCPClientSocket::ConnectTo(char *ip, int port)
	{
		if ( (socketfd = socket(AF_INET,SOCK_STREAM,0)) < 0)
			NETTHROW("Could not get socket filedescriptor");
		int res = setsockopt(socketfd,IPPROTO_TCP,TCP_NODELAY,"true",5);

		ZeroMemory(&addr,sizeof(addr));
		addr.sin_family=AF_INET;
		addr.sin_addr.s_addr=inet_addr(ip);
		addr.sin_port=htons(port);

		res = connect(socketfd,(struct sockaddr*)&addr,sizeof(addr));
		if (res==SOCKET_ERROR)
		{
			NETTHROW("Could not connect to server");
		}

		clientInfo->SetValid();
		StartListenThread();
	}

	void CTCPClientSocket::StartListenThread()
	{
		listenThread = (HANDLE)_beginthread((void (*)(void *))staticListenThread,0,(void*)this);
		if (listenThread == (HANDLE)-1L)
		{
			NETTHROW("Could not Start new ListenThread");
		}
	}

	void CTCPClientSocket::staticListenThread(CTCPClientSocket *self)
	{
		self->ListenThread();
	}

	void CTCPClientSocket::ListenThread()
	{
		WORD len;
		char *data;

		sptr<CNetworkEvent> ne = sptr<CNetworkEvent>(new CNetworkEvent(CNetworkEvent::NEWCONNECTION,(GENERAL::sptr<CClientInfo>)clientInfo));
		ne->SetSendToClientList(&GetClientInfoList());
		network->Lock();
			network->domain->pushNetworkEvent(ne);
			ne = sptr<CNetworkEvent>(); 
		network->Unlock();

		try 
		{
			while (true)
			{
				data=NULL;
				try 
				{
					DoRecv((char*)&len,sizeof(len),0);
					GENERAL::CMemBuffer *buffer=new GENERAL::CMemBuffer(len);
					data = (char*)buffer->Peek(len);
					DoRecv(data,len,0);
	//				so = GENERAL::Deserialize(buffer);
	//				so->SetSendToClientList(&GetClientInfoList());

					network->Lock();
					network->domain->pushIncomingNetwork(buffer,&GetClientInfoList());
					network->Unlock();
				} 
				catch (networkexception & e)
				{
					int err;
					switch (err=WSAGetLastError())
					{
					case WSAECONNRESET:
						network->Lock();
						clientInfo->SetInvalid();
						network->domain->pushNetworkEvent(sptr<CNetworkEvent>(new CNetworkEvent(CNetworkEvent::TEventTyp::CONNECTIONRESET,(GENERAL::sptr<CClientInfo>)GetClientInfo())));
						ne = sptr<CNetworkEvent>(); 
						network->connectedTo.remove(clientInfo);
						clientInfo = sptr<CTCPClientInfo>();
						clientInfoList->clear();
						network->Unlock();
						_endthread();
						break;
					case WSAECONNABORTED:
					case WSAENOTSOCK:
						{
							sptr<CNetworkEvent> ne = sptr<CNetworkEvent>(new CNetworkEvent(CNetworkEvent::TEventTyp::CONNECTIONSHUTDOWN,(GENERAL::sptr<CClientInfo>)GetClientInfo()));
							network->Lock();
							clientInfo->SetInvalid();
							network->domain->pushNetworkEvent(ne);
							ne = sptr<CNetworkEvent>(); 
							network->connectedTo.remove(clientInfo);
							clientInfo = sptr<CTCPClientInfo>();
							clientInfoList->clear();
							network->Unlock();
							_endthread();
							break;
						}
					default:
						throw e;
					}
				}

				//if (data) free(data);
			}
		} 
		catch (exception &)
		{
			network->Lock();
			network->domain->pushNetworkEvent(sptr<CNetworkEvent>(new CNetworkEvent(CNetworkEvent::TEventTyp::CRITICALERROR,(GENERAL::sptr<CClientInfo>)GetClientInfo())));
			clientInfo = sptr<CTCPClientInfo>();
			clientInfoList->clear();
			network->Unlock();
			_endthread();
		}

		clientInfo = sptr<CTCPClientInfo>();
		clientInfoList->clear();
		_endthread();
	}

/**********************************************************************************/
/* CTCPServerSocket                                                               */
/**********************************************************************************/
	CTCPServerSocket::~CTCPServerSocket()
	{
		Shutdown();
		if (WaitForSingleObject(listenThread,INFINITE)==WAIT_TIMEOUT)
		{
			NETTHROW("Listen Thread did not finish within 2 seconds -> releasing Resources anyway");
		}
	}

	void CTCPServerSocket::Listen(char *ip, int port, int backlog)
	{
		if ( (socketfd = socket(AF_INET,SOCK_STREAM,0)) < 0)
			NETTHROW("Could not get socket filedescriptor");

		ZeroMemory(&addr,sizeof(addr));
		addr.sin_family=AF_INET;
		addr.sin_addr.s_addr=inet_addr(ip);
		addr.sin_port=htons(port);

		int res = setsockopt(socketfd,IPPROTO_TCP,TCP_NODELAY,"true",5);

		if (bind(socketfd,(struct sockaddr*)&addr,sizeof(addr))<0)
			NETTHROW("Could not bind socket");

		if (listen(socketfd,backlog)!=0)
			NETTHROW("Could not listen on socket");
	
		StartListenThread();
	}
	void CTCPServerSocket::StartListenThread()
	{
		listenThread = (HANDLE)_beginthread((void (*)(void *))staticListenThread,0,(void*)this);
		if (listenThread == (HANDLE)-1L)
		{
			NETTHROW("Could not Start new ListenThread");
		}
	}

	void CTCPServerSocket::staticListenThread(CTCPServerSocket *self)
	{
		self->ListenThread();
	}

	void CTCPServerSocket::ListenThread()
	{
		while (true)
		{
			int addrlen = sizeof(addr);
			CTCPClientSocket *cs = new CTCPClientSocket(network);

			SOCKET newsock = accept(socketfd,cs->GetpAddr(),&addrlen);
			int res = setsockopt(newsock,IPPROTO_TCP,TCP_NODELAY,"true",5);

			if (newsock==INVALID_SOCKET)
			{
//1				cs->GetClientInfo()->ReleaseRef();
//				cs->GetClientInfo() = sptr<CTCPClientInfo>();
//				cs->GetClientInfoList().clear();
//				SAFE_DELETE(cs);
				_endthread();
			} else {
				cs->SetSocketfd(newsock);
				network->Lock();
				network->connectedTo.push_back(cs->GetClientInfo());
				network->Unlock();
				cs->StartListenThread();
			}
		}
	}


} // namespace NETWORK
